#include <stdlib.h>
#include <string.h>
#include "fitsio.h"
#include "fitsdebug.h"
#include "utilities.h"

#define MAXLEN 100

/* The data type of each column in a binary table needs to */
/* be known to cfitsio (in order to calculate the correct  */
/* number of bits for one row of data). This routine makes */
/* this typecode transition.                               */
/*                                                         */
/*  (taken from p. 151 of the cfitsio manual)              */
/*  Datatype        typecode      Mnemonic (cfitsio)       */
/*  bit, X                 1      TBIT                     */
/*  byte, B               11      TBYTE                    */
/*  logical, L            14      TLOGICAL                 */
/*  ASCII character, A    16      TSTRING                  */
/*  short integer, I      21      TSHORT                   */
/*  integer, J            41      TINT32BIT                */
/*  long long integer, K  81      TLONGLONG                */
/*  real, E               42      TFLOAT                   */
/*  double precision, D   82      TDOUBLE                  */
/*  complex, C            83      TCOMPLEX                 */
/*  double complex, M    163      TDOUBLECOMPLEX           */
/*                                                         */
/*  - The TLONGLONG is an experimental column type and is  */
/*    not recognized in the official FITS Standard document*/
/*  - Information in fitsio.h and testprog.c advices us to */
/*    assign the following realtionships                   */
/*                                                         */
/*     char[xx]      -> xxA                                */
/*     unsigned char -> B (and X)                          */
/*     short         -> I                                  */
/*     int           -> J                                  */
/*     long          -> K                                  */
/*     float         -> E                                  */
/*     double        -> D                                  */
/*                                                         */
/* This routine takes as input a string of the form "int"  */
/* and returns the appropriate Datatype code               */
/*                                                         */

char *fd_get_typecode_binary(typecodeS *result, char *type, int *status) {
  
  /* possible input:                                       */
  /*  'charxxx(yyy)'                                       */
  /*  'short(yyy)'                                         */
  /*  'long(yyy)'                                          */
  /*  'float(yyy)'                                         */
  /*  'double(yyy)'                                        */
  /*  'bitxxx(yyy)'                                        */
  /*  'bytexxx(yyy)'                                       */
  /*  'logical(yyy)'                                       */
  /*  'complex(yyy)'                                       */
  /*  'doublecomplex(yyy)'                                 */
  
  int nchar  = 4,
    nshort   = 5,
    nlong    = 4,
    nfloat   = 5,
    ndouble  = 6,
    nbit     = 3,
    nbyte    = 4,
    nint     = 3,
    nlogical = 7,
    ncomplex = 7,
    ndcomplex= 13;

  int bracket_start, bracket_end;

  char *tmp, *tmp2, tmp3[] = "";
  int len = strlen(type);
  
  result->type = '\0';
  result->int_type = 0;
  result->size = 1;
  result->precision = 0;
  result->vla_length = 0;
  result->vla = 0;
  
  /* allocate correct amount of memory */
  result->raw = mem_malloc_safe((len + 1)*sizeof(char));
  strcpy(result->raw,type);

  tmp = (char *) mem_malloc_safe((len + 1)*sizeof(char));
  tmp2 = (char *) mem_malloc_safe((len + 1)*sizeof(char));

  /* 1: determine if we are dealing with variable length   */
  /*    arrays                                             */
  strcpy(tmp,type);
  bracket_start = str_position(tmp,'[');
  bracket_end = str_position(tmp,']');
  
  if (bracket_start != -1) {
    result->vla = 1;
    if (bracket_end - bracket_start > 1) {
      str_substring(tmp2,tmp,bracket_start+1,bracket_end-1);
      result->vla_length = atoi(tmp2);
    }
    tmp[bracket_start] = '\0';
  }
  
  
  if (strncmp(tmp,"char",nchar) == 0) {
    result->type = 'A';
    result->int_type = TSTRING;
    if (strlen(tmp) > nchar) result->size = atoi(tmp+nchar);
  } else if (strncmp(tmp,"short",nshort) == 0) {
    result->type = 'I';
    result->int_type = TSHORT;
    if (strlen(tmp) > nshort) result->size = atoi(tmp+nshort);
  } else if (strncmp(tmp,"int",nint) == 0) {
    result->type = 'J';
    result->int_type = TINT;
    if (strlen(tmp) > nint) result->size = atoi(tmp+nint);
  } else if (strncmp(tmp,"long",nlong) == 0) {
    result->type= 'J';  /* this should be K but cfitsio crashed in this case */
    result->int_type = TLONG;
    if (strlen(tmp) > nlong) result->size = atoi(tmp+nlong);
  } else if (strncmp(tmp,"float",nfloat) == 0) {
    result->type = 'E';
    result->int_type = TFLOAT;
    if (strlen(tmp) > nfloat) result->size = atoi(tmp+nfloat);
  } else if (strncmp(tmp,"double",ndouble) == 0) {
    result->type = 'D';
    result->int_type = TDOUBLE;
    if (strlen(tmp) > ndouble) result->size = atoi(tmp+ndouble);
  } else if (strncmp(tmp,"bit",nbit) == 0) {
    result->type = 'X';
    result->int_type = TBIT;
    if (strlen(tmp) > nbit) result->size = atoi(tmp+nbit);
  } else if (strncmp(tmp,"byte",nbyte) == 0) {
    result->type = 'B';
    result->int_type = TBYTE;
    if (strlen(tmp) > nbyte) result->size = atoi(tmp+nbyte);
  } else if (strncmp(tmp,"logical",nlogical) == 0) {
    result->type = 'L';
    result->int_type = TLOGICAL;
    if (strlen(tmp) > nbyte) result->size = atoi(tmp+nbyte);
  } else if (strncmp(tmp,"complex",ncomplex) == 0) {
    result->type = 'C';
    result->int_type = TCOMPLEX;
    if (strlen(tmp) > nbyte) result->size = atoi(tmp+nbyte);
  } else if (strncmp(tmp,"doublecomplex",ndcomplex) == 0) {
    result->type = 'M';
    result->int_type = TDBLCOMPLEX;
    if (strlen(tmp) > nbyte) result->size = atoi(tmp+nbyte);
    /*
      The following lines allow one to use the ascii fits table 
      typecode specifications for a binary fits table.
     */
  } else if (strncmp(tmp,"A",1) == 0) {
    result->type = 'A';
    result->int_type = TSTRING;
    result->size = atoi(tmp+1);
  } else if (strncmp(tmp,"I",1) == 0) {
    result->type = 'J';
    result->int_type = TINT;
  } else if (strncmp(tmp,"F",1) == 0) {
    result->type= 'E';
    result->int_type = TFLOAT;
  } else if (strncmp(tmp,"E",1) == 0) {
    result->type = 'E';
    result->int_type = TFLOAT;
  } else if (strncmp(tmp,"D",1) == 0) {
    result->type = 'D';
    result->int_type = TDOUBLE;
    /*
      End of ascii code for binary tables.
     */
  } else {
    msg_message(2,"Error (fd, fd_get_typecode_binary) : The type declaration not recognized.\n");
    *status = FD_STAT_NO_TYPE;
  }


  /* Build up TTYPE string */
  if (result->vla) {
    if (result->vla_length == 0) {
      sprintf(tmp3,"%d%c%c",1,'P',result->type);
    } else {
      sprintf(tmp3,"%d%c%c%c%d%c",1,'P',result->type,'(',result->vla_length,')');
    }
  } else {
    sprintf(tmp3,"%d%c",result->size,result->type);
  }

  result->type_code = mem_malloc_safe((strlen(tmp3)+1)*sizeof(char));
  strcpy(result->type_code,tmp3);

  return result->type_code;

}

char *fd_get_typecode_ascii(typecodeS *result, char *type, int *status) {
  
  /* possible input:                                       */
  /*  'Aw'         (character)                             */
  /*  'Iw'         (integer)                               */
  /*  'Fw.d'       (single precision float)                */
  /*  'Ew.d'       (single precision exponential notation) */
  /*  'Dw.d'       (double precision exponential notation) */
  
  char *tmp, sub[MAXSTRLEN];
  int p1 = 1, p2;
  int len = strlen(type);
  
  result->type = '\0';
  result->int_type = 0;
  result->size = 1;
  result->vla = 0;
  result->vla_length = 0;
  
  /* allocate correct amount of memory */
  result->raw = mem_malloc_safe((len + 1)*sizeof(char));
  strcpy(result->raw,type);

  tmp = (char *) mem_malloc_safe((len + 1)*sizeof(char));
  strcpy(tmp,type);
  
  if (strncmp(tmp,"A",1) == 0) {
    result->type = 'A';
    result->int_type = TSTRING;
  } else if (strncmp(tmp,"I",1) == 0) {
    result->type = 'K';
    result->int_type = TINT;
  } else if (strncmp(tmp,"F",1) == 0) {
    result->type= 'E';
    result->int_type = TFLOAT;
  } else if (strncmp(tmp,"E",1) == 0) {
    result->type = 'E';
    result->int_type = TFLOAT;
  } else if (strncmp(tmp,"D",1) == 0) {
    result->type = 'D';
    result->int_type = TDOUBLE;
  } else {
    msg_message(2,"Error (fd, fd_get_typecode_ascii) : The type declaration not recognized.\n");
    *status = FD_STAT_NO_TYPE;
  }

  p2 = str_position(tmp,'.');
  if (p2 != -1) {
    result->size = atoi(str_substring(sub,tmp,p1,p2-1));
    result->precision = atoi(tmp+p2+1);
  } else {
    result->size = atoi(tmp+1);
  }

  result->type_code = mem_malloc_safe((strlen(type)+1)*sizeof(char));
  strcpy(result->type_code,type);


  free(tmp);

  return result->type_code;

}


int fd_get_type_from_typecode(typecodeS typecode) {

  return typecode.int_type;

}

int fd_get_size_from_typecode(typecodeS typecode) {

  return typecode.size;

}

long fd_get_nelem_from_typecode(typecodeS typecode) {

  return typecode.vla_length;

}


int fd_set_typecode_binary(typecodeS *typecode, char* tform, int *status) {

  int p1, p2;
  char sub[MAXSTRLEN] = "";
  char raw_type[MAXSTRLEN] = "";
  char raw_vla[MAXSTRLEN] = "";


  /* 1: set typecode */
  typecode->type_code = 
    mem_malloc_safe((strlen(tform)+1)*sizeof(char));
  strcpy(typecode->type_code, tform);

  /* 2: set vla */
  if (str_position(tform,'P') != -1) {
    typecode->vla = 1; 
    p1 = str_position(tform,'(');
    p2 = str_position(tform,')');
    if (p2-p1 != 1 && p2 > p1) {
      typecode->vla_length = atoi(str_substring(sub,tform,p1+1,p2-1)); 
      sprintf(raw_vla,"[%d]",typecode->vla_length);
    } else {
      typecode->vla_length = 0;
      sprintf(raw_vla,"[]");
    }
  } else {
    typecode->vla = 0;
    typecode->vla_length = 0;
  }

  typecode->size = 1;
  /* set type and int_type */
  if ((p1 = str_position(tform,'A')) != -1) {
    typecode->type = 'A';
    typecode->int_type = TSTRING;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "char");
    else
      sprintf(raw_type, "char%d", typecode->size);
  } else if ((p1 = str_position(tform,'I')) != -1) {
    typecode->type = 'I';
    typecode->int_type = TSHORT;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "short");
    else
      sprintf(raw_type, "short%d", typecode->size);
  } else if ((p1 = str_position(tform,'J')) != -1) {
    typecode->type= 'J';
    typecode->int_type = TINT;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "int");
    else
      sprintf(raw_type, "int%d", typecode->size);
  } else if ((p1 = str_position(tform,'K')) != -1) {
    typecode->type= 'K';
    typecode->int_type = TLONG;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "long");
    else
      sprintf(raw_type, "long%d", typecode->size);
  } else if ((p1 = str_position(tform,'E')) != -1) {
    typecode->type = 'E';
    typecode->int_type = TFLOAT;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "float");
    else
      sprintf(raw_type, "float%d", typecode->size);
  } else if ((p1 = str_position(tform,'D')) != -1) {
    typecode->type = 'D';
    typecode->int_type = TDOUBLE;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "double");
    else
      sprintf(raw_type, "double%d", typecode->size);
  } else if ((p1 = str_position(tform,'X')) != -1) {
    typecode->type = 'X';
    typecode->int_type = TBIT;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "bit");
    else
      sprintf(raw_type, "bit%d", typecode->size);
  } else if ((p1 = str_position(tform,'B')) != -1) {
    typecode->type = 'B';
    typecode->int_type = TBYTE;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "byte");
    else
      sprintf(raw_type, "byte%d", typecode->size);
  } else if ((p1 = str_position(tform,'L')) != -1) {
    typecode->type = 'L';
    typecode->int_type = TLOGICAL;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "logical");
    else
      sprintf(raw_type, "logical%d", typecode->size);
  } else if ((p1 = str_position(tform,'C')) != -1) {
    typecode->type = 'C';
    typecode->int_type = TCOMPLEX;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "complex");
    else
      sprintf(raw_type, "complex%d", typecode->size);
  } else if ((p1 = str_position(tform,'M')) != -1) {
    typecode->type = 'M';
    typecode->int_type = TDBLCOMPLEX;
    unless(typecode->vla,
	   typecode->size = atoi(str_substring(sub,tform,0,p1-1)););
    if (typecode->size == 1) 
      sprintf(raw_type, "dblcomplex");
    else
      sprintf(raw_type, "dblcomplex%d", typecode->size);
  } else if ((p1 = str_position(tform,'K')) != -1) {
    msg_message(2,"Error (fd, fd_set_typecode_binary) : The type declaration \"int\" is not allowed.\n       Please define integers as either \"short\" or \"long\".\n");
    *status = FD_STAT_NO_INT;
  } else {
    msg_message(2,"Error (fd, fd_set_typecode_binary) : TFORM declaration not recognized.\n");
    *status = FD_STAT_NO_TYPE;
  }

  strcat(raw_type,raw_vla);
  typecode->raw = mem_malloc_safe((strlen(raw_type)+1)*sizeof(char));
  strcpy(typecode->raw,raw_type);

    
  return 0;

}

int fd_set_typecode_ascii(typecodeS *typecode, char* tform, int *status) {

  /* possible input:                                       */
  /*  'Aw'         (character)                             */
  /*  'Iw'         (integer)                               */
  /*  'Fw.d'       (single precision float)                */
  /*  'Ew.d'       (single precision exponential notation) */
  /*  'Dw.d'       (double precision exponential notation) */

  int p1 = 1, p2;
  char sub[MAXSTRLEN] = "";
  char tmp[MAXSTRLEN] = "";
  int len = strlen(tform)+1;

  /* 2: set vla */
  typecode->vla = 0;
  typecode->vla_length = 0;
  typecode->type = '\0';
  typecode->int_type = 0;
  typecode->size = 1;
  
  /* allocate correct amount of memory */
  typecode->raw = mem_malloc_safe((len + 1)*sizeof(char));
  strcpy(typecode->raw,tform);

  strcpy(tmp,tform);
  
  if (strncmp(tmp,"A",1) == 0) {
    typecode->type = 'A';
    typecode->int_type = TSTRING;
  } else if (strncmp(tmp,"I",1) == 0) {
    typecode->type = 'K';
    typecode->int_type = TINT;
  } else if (strncmp(tmp,"F",1) == 0) {
    typecode->type= 'E';
    typecode->int_type = TFLOAT;
  } else if (strncmp(tmp,"E",1) == 0) {
    typecode->type = 'E';
    typecode->int_type = TFLOAT;
  } else if (strncmp(tmp,"D",1) == 0) {
    typecode->type = 'D';
    typecode->int_type = TDOUBLE;
  } else {
    msg_message(2,"Error (fd, fd_get_typecode_ascii) : The type declaration not recognized.\n");
    *status = FD_STAT_NO_TYPE;
  }

  p2 = str_position(tmp,'.');
  if (p2 != -1) {
    typecode->size = atoi(str_substring(sub,tmp,p1,p2-1));
    typecode->precision = atoi(tmp+p2+1);
  } else {
    typecode->size = atoi(tmp+1);
  }

  typecode->type_code = mem_malloc_safe(len*sizeof(char));
  strcpy(typecode->type_code,tform);


  return 0;

}


void fd_print_typecode(typecodeS typecode) {


  printf("Raw        : %s\n",typecode.raw);
  printf("Typecode   : %s\n",typecode.type_code);
  printf("Type       : %c\n",typecode.type);
  printf("Int_Type   : %d\n",typecode.int_type);
  printf("Size       : %d\n",typecode.size);
  printf("Precision  : %d\n",typecode.precision);
  printf("VLA        : %d\n",typecode.vla);
  printf("VLA_length : %d\n",typecode.vla_length);


}
