#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fitsio.h>

#include "messages.h"

#define MAXSTRLEN 128
#define BITPIX 8
#define MAXNCOLUMN 6

void fits_printerror(char *process, int status);

/***************************************************************************/
/* This routine takes a file header string and opens one of three possible */
/* files, depending upon the logical value of file_type.  An appropriate   */
/* file header is written, including a dummy primary HDU and a first HDU   */
/* that will contain the information listing temperatures, densities, etc  */
/* that are in this file.  A fits file pointer is returned.                */
/* Filetype = 0 : Line output                                              */
/* Filetype = 1 : Complete binned continuum output                         */
/* Filetype = 2 : New per element compressed continuum output              */
/* Filetype = 3 : Complete compressed per ion continuum output             */
/***************************************************************************/

void fits_spec_init(char *filehead, int filetype, int Nrows, fitsfile **fptr) {
  
  int status=0;
  char *ttype[MAXNCOLUMN], *tunit[MAXNCOLUMN], *tform[MAXNCOLUMN];
  int tfields=0, i;
  long naxis2;
  static char *ext[4]={"_line.fits","_cont.fits","_coco.fits","_comp.fits"};
  char filename[MAXSTRLEN];

  if ((filetype<0)||(filetype>3)) {
    errmess("fits_spec_init","Illegal file type %d: 0-3 legal.",
	    filetype);
  }

  strcpy(filename, filehead);
  strcat(filename, ext[filetype]);
  if (fits_create_file(fptr, filename, &status)) {
    fits_printerror("creating emission file", status);
  }
  
  if (fits_write_imghdr(*fptr, BITPIX, 0, 0, &status)) {
    fits_printerror("writing image header", status);
  }
  
  /* Additional keywords describing file. */  
  if (fits_write_key(*fptr, TSTRING, "CONTENT", "SPECTRUM",
		     "Emission Line Project output", &status)) {
    fits_printerror("writing CONTENT keyword", status);
  }
  
  if (fits_write_key(*fptr, TSTRING, "FILENAME", "original",
		     "Parent File ", &status)) {
    fits_printerror("writing FILENAME keyword", status);
  }
  
  if (fits_write_key(*fptr, TSTRING, "ORIGIN", "APEC",
		     "origin of FITS file", &status)) {
    fits_printerror("writing ORIGIN keyword",status);
  }
  
  /* 
     if (fits_write_key(*fptr, TSTRING, "EXTNAME", "EMISSIVITY",
     "Spectral emission data", &status)) {
     fits_printerror("writing EXTNAME keyword",status);
     }
     
     if (fits_write_key(*fptr, TSTRING, "HDUNAME", hduname,
     "Spectral emission data", &status)) {
     fits_printerror("writing HDUNAME keyword",status);
     }
     
     if (fits_write_key(*fptr, TSTRING, "HDUCLASS", "ASC",
     "ASC data", &status)) {
     fits_printerror("writing HDUCLASS keyword",status);
     }
     
     if (fits_write_key(*fptr, TSTRING, "HDUCLAS1", "EMISSIVITY",
     "Line and continuum emissivity", &status)) {
     fits_printerror("writing HDUCLAS1 keyword", status);
     }
     
     switch (filetype)
     case (0):
     if (fits_write_key(*fptr, TSTRING, "HDUCLAS2", "LINE",
     "Line emission data", &status)) {
     fits_printerror("writing HDUCLAS2 keyword",status);
     }
     case (1):
     if (fits_write_key(*fptr, TSTRING, "HDUCLAS2", "CONTINUUM",
     "Continuum emission data", &status)) {
     fits_printerror("writing HDUCLAS2 keyword",status);
     }
     }
  */
  
  if (fits_write_key(*fptr, TSTRING, "HDUVERS1", "1.0.0",
		     "version of format", &status)) {
    fits_printerror("writing HDUVERS1 keyword",status);
  }
  
  /* Primary header is now written.  Time to create first HDU */

  for(i=0;i<MAXNCOLUMN;i++) {
    ttype[i] = (char *) malloc(MAXSTRLEN*sizeof(char));
    tunit[i] = (char *) malloc(MAXSTRLEN*sizeof(char));
    tform[i] = (char *) malloc(MAXSTRLEN*sizeof(char));
  }

  switch(filetype) {
  case(0):
    tfields = 4;
    sprintf(ttype[0],"kT");
    sprintf(tform[0],"1E");
    sprintf(tunit[0],"keV");
    sprintf(ttype[1],"EDensity");
    sprintf(tform[1],"1E");
    sprintf(tunit[1],"cm**-3");
    sprintf(ttype[2],"Nelement");
    sprintf(tform[2],"1J");
    sprintf(tunit[2]," ");
    sprintf(ttype[3],"Nline");
    sprintf(tform[3],"1J");
    sprintf(tunit[3]," ");
    break;
  case (1):
    tfields = 4;
    sprintf(ttype[0],"kT");
    sprintf(tform[0],"1E");
    sprintf(tunit[0],"keV");
    sprintf(ttype[1],"EDensity");
    sprintf(tform[1],"1E");
    sprintf(tunit[1],"cm**-3");
    sprintf(ttype[2],"NElement");
    sprintf(tform[2],"1J");
    sprintf(tunit[2]," ");
    sprintf(ttype[3],"NCont");
    sprintf(tform[3],"1J");
    sprintf(tunit[3]," ");
    break;
  case (2):
    tfields = 5;
    sprintf(ttype[0],"kT");
    sprintf(tform[0],"1E");
    sprintf(tunit[0],"keV");
    sprintf(ttype[1],"EDensity");
    sprintf(tform[1],"1E");
    sprintf(tunit[1],"cm**-3");
    sprintf(ttype[2],"NElement");
    sprintf(tform[2],"1J");
    sprintf(tunit[2]," ");
    sprintf(ttype[3],"NCont");
    sprintf(tform[3],"1J");
    sprintf(tunit[3]," ");
    sprintf(ttype[4],"NPseudo");
    sprintf(tform[4],"1J");
    sprintf(tunit[4]," ");
    break;
  case (3):
    tfields = 5;
    sprintf(ttype[0],"kT");
    sprintf(tform[0],"1E");
    sprintf(tunit[0],"keV");
    sprintf(ttype[1],"EDensity");
    sprintf(tform[1],"1E");
    sprintf(tunit[1],"cm**-3");
    sprintf(ttype[2],"NElement");
    sprintf(tform[2],"1J");
    sprintf(tunit[2]," ");
    sprintf(ttype[3],"NCont");
    sprintf(tform[3],"1J");
    sprintf(tunit[3]," ");
    sprintf(ttype[4],"NPseudo");
    sprintf(tform[4],"1J");
    sprintf(tunit[4]," ");
    break;
  default:
    errmess("fits_spec_init","Illegal filetype value %d",filetype);
  }

  /* Now actually create the new HDU  */
  naxis2 = (long) Nrows;
  if (fits_create_tbl(*fptr, BINARY_TBL, naxis2, tfields, ttype, tform, tunit,
		      "PARAMETERS", &status)) {
    errmess("fits_spec_init","Error creating binary HDU table");
  }

  /* And write some header keywords */

  if (fits_write_key(*fptr, TSTRING, "HDUCLASS", "Proposed OGIP",
		     "Proposed OGIP standard ", &status)) {
    fits_printerror("writing HDUCLASS keyword", status);
  }
  
  switch (filetype) {
  case (0):
    if (fits_write_key(*fptr, TSTRING, "HDUCLAS1", "LINE MODEL",
		       "line emission spectra model", &status)) {
      fits_printerror("writing HDUCLAS1 keyword", status);
    }
    break;
  case (1):
    if (fits_write_key(*fptr, TSTRING, "HDUCLAS1", "CONTINUUM MODEL",
		       "model continuum spectra ", &status)) {
      fits_printerror("writing HDUCLAS1 keyword", status);
    }
    break;
  case (2):
    if (fits_write_key(*fptr, TSTRING, "HDUCLAS1", "COMP CONT MODEL",
		       "compressed continua spectra", 
		       &status)) {
      fits_printerror("writing HDUCLAS1 keyword", status);
    }
    break;
  case (3):
    if (fits_write_key(*fptr, TSTRING, "HDUCLAS1", "COMP CONT MODEL",
		       "per ion compressed continua spectra output", 
		       &status)) {
      fits_printerror("writing HDUCLAS1 keyword", status);
    }
    break;
  default:
    errmess("fits_spec_init","Illegal filetype value %d",filetype);
  }

  if (fits_write_key(*fptr, TSTRING, "HDUCLAS2", "PARAMETERS",
		     "extension containing parameter info", &status)) {
    fits_printerror("writing HDUCLAS2 keyword", status);
  }

  if (fits_write_key(*fptr, TSTRING, "HDUVERS1", "1.0.0",
		     "version of format", &status)) {
    fits_printerror("writing HDUVERS1 keyword",status);
  }

  /* And now we're done; time to free up some memory. */

  for(i=0;i<MAXNCOLUMN;i++) {
    free(ttype[i]);
    free(tunit[i]);
    free(tform[i]);
  }

}

void f77_fits_spec_init_(char *filehead, int *filetype, fitsfile **fptr,
			    int *Nrows, int namlen) {

  fits_spec_init(filehead, *filetype, *Nrows, fptr);
  
}
