#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <pfile.h>

#include "messages.h"
#include "apec.h"
#include "slang.h"
#include "apec_slang.h"
#include "apec_funcdefs.h"

static unsigned char _initialized;

static ElementP *Atomic_params;
static struct DEBUG_FILES debug;
static struct PARAMETERS params;
static struct BIN_SPEC spectrum;
static struct ION_BAL_TABLE ionbal;
static int changed_atoms,changed_ions,changed_bin,changed_abund,changed_ionbal;

int strip_strcmp(const char *s1, const char *s2);
void create_spectral_bins(struct PARAMETERS *params,struct BIN_SPEC *spectrum);
void get_atomic_data(struct PARAMETERS *params, ElementP **Atomic_params);
void get_abund(struct PARAMETERS *params);
void fits_open(struct PARAMETERS *params, ElementP *Atomic_data, 
	       struct BIN_SPEC *spectrum);
void init_brems_hummer(char *filemap_file);
void init_brems_relativistic(char *filemap_file);
void init_sigma_hydrogenic_arr(void);
void calc_maxwell_rates(struct COLLTRNS_DATA *collision, double dE, double T,
			int Z, double degl, double degu, 
			double *exc_rate, double *dex_rate);
void open_debug_output(struct DEBUG_FILES *debug);
void read_ion_bal(char *ionbal_file, struct DEBUG_FILES *debug, 
		  struct ION_BAL_TABLE *ionbal);
void parse_atom(char *inc_atoms, char *inc_ions, struct atomstr **atoms);
void extract_atom_input(struct PARAMETERS *params, struct atomstr *atoms);
int create_parameter_file(char *filename);
void expand_env(char *filename);

void apec_params(void);
void apec_init(void);
void apec_run(void);
void apec_calc_upsilon(void);

void get_input_data(int argc, char *argv[], Param_File_Type *p_input,
		    struct PARAMETERS *params, 
		    struct DEBUG_FILES *debug);

int init_apec_module(void);

/* Need to define :  get_coll, get_xxx */
static SLang_Intrin_Fun_Type APEC_Intrinsics [] = 
  {
    MAKE_INTRINSIC_0("apec_params", apec_params, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("apec_init", apec_init, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("apec_run", apec_run, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("apec_calc_upsilon", apec_calc_upsilon, SLANG_VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
  };

int init_apec_module() {

  Param_File_Type *p;

  if (_initialized) return 0;  /* Routine has already run. */

  _initialized = 1;
  if (SLadd_intrin_fun_table (APEC_Intrinsics, NULL) == -1) {
    exit (1);
  }

  /* Now set up some APEC initialization values .*/

  /* Read apec.par to get initial parameters. If it doesn't exist, create it */

  if (NULL == (p = pf_open_parameter_file("apec.par", "r"))) {
    SLang_vmessage("apec: creating new apec.par parameter file.\n");
    if (create_parameter_file("apec.par")) { /* So make one... */
      SLang_vmessage("apec: error creating new apec.par parameter file.\n");
      exit(1);
    }
    if (NULL == (p = pf_open_parameter_file("apec.par", "r"))) { 
      SLang_vmessage("apec: error reading new apec.par parameter file.\n");
      exit(1);
    }
  }

  /* And now get the input data. */
  get_input_data(0, NULL, p, &params, &debug);

  /* Set up the truly never-changing stuff.  Need to use the filemap,
     which means we need to expand any variables first */
  expand_env(params.filemap_file);

  /* Initialize Hummer Bremsstrahlung data points */
  init_brems_hummer(params.filemap_file);

  /* Initialize Relativistic Bremsstrahlung data points */
  init_brems_relativistic(params.filemap_file);

  /* Initialize Hydrogenic Cross Section calculation */
  init_sigma_hydrogenic_arr();

  /* Need to load all data in, so assume all values are "changed" */
  changed_atoms=TRUE;
  changed_ions=TRUE;
  changed_bin=TRUE;
  changed_abund=TRUE;
  changed_ionbal=TRUE;

  return(0);
}


void apec_params(void) {  

  /*  If called with no inputs, returns the current default (or most
      recently set) input values as a structure.  If called with a
      structure, uses it to reset input parameters.  */
  /*  Feature: checks certain values, and sets 'flags' to tell the 
      apec_init routine to reload.  The relevant variables are:
      atoms, ions, binning, abundances, ionbal */

  struct atomstr *atoms=NULL;

  int i;
  APEC_Param_Type apec_param;
  SLang_Array_Type *Tarr, *Narr;

  char *small, *full, hummer[MAXSTRLEN];
  char kellogg[MAXSTRLEN], relativistic[MAXSTRLEN], none[MAXSTRLEN];
  
  small = (char *) malloc(10*sizeof(char));
  full  = (char *) malloc(10*sizeof(char));
  strcpy(small, "small");
  strcpy(full, "full");
  strcpy(hummer, "Hummer");
  strcpy(kellogg, "Kellogg");
  strcpy(relativistic, "Relativistic");
  strcpy(none, "None");
  
  if (SLang_Num_Function_Args > 1 ) {
    SLang_verror (SL_USAGE_ERROR, "Usage: apec_init([inputstructure])");
    return;
  }

  if (SLang_Num_Function_Args == 0) {
    /* Transfer output values to S-lang array.  Note that only what we
       consider to be "useful" values are transferred. */

    Tarr=SLang_create_array (SLANG_DOUBLE_TYPE, 0, NULL, &params.Ntemp, 1);
    Narr=SLang_create_array (SLANG_DOUBLE_TYPE, 0, NULL, &params.Ndens, 1);
    if ((Tarr == NULL)||(Narr == NULL)) {
      SLang_vmessage("Error creating SLang arrays.");
      return;
    }
    
    /* Now set the elements of the array */
    for (i = 0; i < params.Ntemp; i++) {
      SLang_set_array_element(Tarr, &i, &(params.temperature[i]));
    }
    for (i = 0; i < params.Ndens; i++) {
      SLang_set_array_element(Narr, &i, &(params.density[i]));
    }

    apec_param.modelname = params.modelname; 
    if (params.do_comp == FALSE) apec_param.outputtype = small;
    if (params.do_comp != FALSE) apec_param.outputtype = full;
    apec_param.filemap_file = params.filemap_file;
    apec_param.ionbal_file = params.ionbal_file;
    apec_param.abund_source = params.abund_source;
    apec_param.min_epsilon = params.min_epsilon;
    apec_param.T_array = Tarr;
    apec_param.N_array = Narr;
    apec_param.Nbin = params.Nbin;
    apec_param.BinMin = params.BinMin;
    apec_param.BinMax = params.BinMax;
    apec_param.inc_atoms = params.inc_atoms;
    apec_param.inc_ions = params.inc_ions;
    apec_param.do_rrc   = params.do_rrc;
    apec_param.do_sat   = params.do_sat;
    apec_param.do_lines = params.do_lines;  
    apec_param.do_twoph = params.do_twoph;
    apec_param.do_brems = params.do_brems;
    if (params.Brems_Type == HUMMER) apec_param.brems_type = hummer;
    if (params.Brems_Type == KELLOGG) apec_param.brems_type = kellogg;
    if (params.Brems_Type == RELATIVISTIC) 
      apec_param.brems_type = relativistic; 
    if (params.Brems_Type == BREMS_NONE) apec_param.brems_type = none; 
      
    if (SLang_push_cstruct ((VOID_STAR)&apec_param, APEC_Param_Layout)==-1){
      SLang_vmessage("Error pushing struct.");
    }
    SLang_free_array(Tarr);
    SLang_free_array(Narr);
    return;
  }

  if (SLang_Num_Function_Args == 1) { /* Set new values. */

    if (SLang_pop_cstruct((VOID_STAR)&apec_param, APEC_Param_Layout)==-1) {
      SLang_vmessage("Error getting input parameters.");
      return;
    }
    strcpy(params.modelname,apec_param.modelname);
    if (strcmp(params.filemap_file,apec_param.filemap_file)!=0) {
      changed_atoms=TRUE;
      changed_ions=TRUE;
      changed_abund=TRUE;
      changed_ionbal=TRUE;
      strcpy(params.filemap_file,apec_param.filemap_file);
    }
    if (strcmp(params.ionbal_file,apec_param.ionbal_file)!=0) {
      changed_ionbal=TRUE;
      strcpy(params.ionbal_file,apec_param.ionbal_file);

    }
    if (strcmp(params.ionbal_file,apec_param.ionbal_file)!=0) {
      changed_abund=TRUE;
      strcpy(params.abund_source,apec_param.abund_source);
    }
    params.min_epsilon = apec_param.min_epsilon;

    if ((strip_strcmp(apec_param.outputtype,"full")==0)||
	(strip_strcmp(apec_param.outputtype,"all")==0)) {
      params.do_line = TRUE;
      params.do_coco = TRUE;
      params.do_comp = TRUE;
      params.do_bincont = TRUE;
      params.do_xspectable = TRUE;
    }
    
    if (strip_strcmp(apec_param.outputtype,"small")==0) {
      params.do_line = TRUE;
      params.do_coco = TRUE;
      params.do_comp = FALSE;
      params.do_bincont = FALSE;
      params.do_xspectable = FALSE;
    }

    /* Some fancy work here to get the temperature/density arrays */
    params.Ntemp = (apec_param.T_array)->num_elements;
    if (params.temperature != NULL) free(params.temperature);
    params.temperature = (double *) malloc(params.Ntemp*sizeof(double));
    for (i=0;i<params.Ntemp;i++) 
      SLang_get_array_element(apec_param.T_array, &i,
			      &(params.temperature[i]));
    params.Ndens = (apec_param.N_array)->num_elements;
    if (params.density != NULL) free(params.density);
    params.density = (double *) malloc(params.Ndens*sizeof(double));
    for (i=0;i<params.Ndens;i++) 
      SLang_get_array_element(apec_param.N_array, &i, &(params.density[i]));

    if (params.Nbin != apec_param.Nbin ||
	params.BinMin != apec_param.BinMin ||
	params.BinMax != apec_param.BinMax) {
      changed_bin = TRUE;
      params.Nbin = apec_param.Nbin;
      params.BinMin = apec_param.BinMin;
      params.BinMax = apec_param.BinMax;      
    }

    if ((strcpy(params.inc_atoms,apec_param.inc_atoms)!=0)||
	(strcpy(params.inc_ions,apec_param.inc_ions)!=0)) {
      changed_atoms=TRUE;
      changed_ions=TRUE;
      parse_atom(apec_param.inc_atoms, apec_param.inc_ions, &atoms);
      extract_atom_input(&params, atoms);
    }

    params.do_rrc = apec_param.do_rrc;
    params.do_sat = apec_param.do_sat;      
    params.do_lines = apec_param.do_lines;  
    params.do_twoph = apec_param.do_twoph;  
    params.do_brems = apec_param.do_brems;
    
    if (strip_strcmp(apec_param.brems_type,"Hummer")==0) 
      {params.Brems_Type = HUMMER; }
    else if (strip_strcmp(apec_param.brems_type,"Kellogg")==0) 
      {params.Brems_Type = KELLOGG; }
    else if (strip_strcmp(apec_param.brems_type,"Relativistic")==0) 
      {params.Brems_Type = RELATIVISTIC; }
    else if (strip_strcmp(apec_param.brems_type,"None")==0) 
      {params.Brems_Type = BREMS_NONE; }

    SLang_free_cstruct ((VOID_STAR)&apec_param, APEC_Param_Layout);
  }
}

void apec_init(void) {

  /*  Similar to initialize.c */
  /*  Loads APED files and sets up parameters for a run, using the 
      most current set of input parameters OR the passed-in values. 
      Note that this should keep track of what has been read so that
      we don't unnecessarily re-read files.  */

  APEC_Param_Type apec_param;

  if (SLang_Num_Function_Args > 1 ) {
    SLang_verror (SL_USAGE_ERROR, "Usage: apec_load([inputstructure])");
    return;
  }

  if (SLang_Num_Function_Args == 1) { 
    apec_params(); /* Call apec_params to set up the params structure. */
    /* Pop result off the stack. */
    if (SLang_pop_cstruct((VOID_STAR)&apec_param, APEC_Param_Layout)==-1) {
      SLang_vmessage("Error getting input parameters.");
      return;
    }
  }
  
  /* Open all output files.  */

  /* The debug values are all 0, so we can redo it safely. */
  open_debug_output(&debug);
  
  /* Create spectral bins (and fill in some parameters) */
  if (changed_atoms||changed_ions||changed_bin) {
    create_spectral_bins(&params, &spectrum);
  }

  /* Read atomic data into structure state */
  if (changed_atoms||changed_ions) {
    get_atomic_data(&params, &Atomic_params);
  }

  /* Read atomic abundances */
  if (changed_abund) {
    get_abund(&params);
  }

  /* Read ionization balance */
  if (changed_ionbal) {
    if (params.ionbal_file[0] != '\0') {
      expand_env(params.ionbal_file);
      read_ion_bal(params.ionbal_file, &debug, &ionbal);
    }
  }

  /* Initialize FITS output files */
  fits_open(&params, Atomic_params, &spectrum); 

}

void apec_run(void) {

  /*  Runs apec code using set up input parameters.  If passed a 
      structure, calls apec_init first. */
  struct ELEMENT_ION_STATE *state;
  struct DEBUG_FILES debug;
  struct LEVEL_RATES *level_rates=NULL;
  struct TRANS_RATES trans_rates; 
  struct EMISSION_LINES *line=NULL, *ionrec_line=NULL;
  struct TWOPHOT_DATA twophot;
  int iEl, irmJ, iTemp, iDens, nlev;
  int Z, rmJ;
  double Te, Ne;
  double *Ion_pop=NULL;       /* These relate to the AR data */
  double *Lev_pop=NULL;       /* level population */
  double nfree;

  if (SLang_Num_Function_Args > 1 ) {
    SLang_verror (SL_USAGE_ERROR, "Usage: apec_run([inputstructure])");
    return;
  }

  if (SLang_Num_Function_Args == 1) { 
    apec_init(); /* Call apec_init to read & load new parameters */
  }

  /* And now run the code...*/
  for (iDens = 0; iDens < params.Ndens; iDens++) {
    Ne = params.density[iDens];

    for (iTemp=0; iTemp < params.Ntemp ; iTemp++) {
      Te = params.temperature[iTemp];
      
      SLang_vmessage("apec: T = %e, n = %e",Te, Ne);

      for (iEl = 0; iEl < params.N_Elements; iEl++) {

	state = Atomic_params[iEl];
	Z = params.list_Z[iEl];

	/* Print out the numerical values in structure arrays */
	/* print_table(&debug, Atomic_params[iEl]); */

	/* ********************************************************* */
	/*    Calculate ion population for ions of interest          */
	/* ********************************************************* */

	Ion_pop = (double *) malloc_safe((Z+1)*sizeof(double));
	level_rates = (struct LEVEL_RATES *)
	  malloc_safe((Z+1)*sizeof(struct LEVEL_RATES));
	
	if (params.ionbal_file[0] != '\0') { /* Load table-driven ion bal */
	  find_ion_bal(Z, Te, Ne, Ion_pop, &ionbal, &params);
	
	  for (rmJ = 1; rmJ < Z+1; rmJ++) {
	    
	    allocate_level_rates(state[rmJ-1].ion_inf->nlev,
				 &(level_rates[rmJ-1]));
	    
	    /* calc_ioniz_rate(Te, Ne, &debug, &params,&(state[rmJ-1]),
	       &(level_rates[rmJ-1].qioniz_total)); */
	    level_rates[rmJ-1].qioniz_total = 0.0;
	    if (rmJ-1 < Z ) {
	      assume_recomb_rates(Ion_pop[rmJ-1],Ion_pop[rmJ],
				  level_rates[rmJ-1].qioniz_total,
				  &(level_rates[rmJ-1].qrecomb_total),
				  &(level_rates[rmJ-1].ionc));
	    }
	  }
	} else { 
	  calc_ionbal(Te, Ne, state, &debug, level_rates, Ion_pop, 
		      &ionrec_line); 
	}

	/* ********************************************************* */
	/*          Calculate level population and emission          */
	/* ********************************************************* */

	for (irmJ=0; irmJ < params.list_NrmJ[iEl]; irmJ++)
	  alloc_spectrum(Z, params.list_rmJ[iEl][irmJ], &spectrum);

	for (irmJ=0; irmJ < params.list_NrmJ[iEl]; irmJ++) {

	  rmJ = params.list_rmJ[iEl][irmJ];
	  nlev=state[rmJ-1].ion_inf->nlev;
	  
	  if ((nlev > 0) && (Ion_pop[rmJ-1] > MIN_ION_POP)) {

	    if (params.do_lines) {
	      Lev_pop = (double *) malloc_safe(nlev*sizeof(double));
	      calc_trans_rates(Te, Te, Ne, &debug, &(state[rmJ-1]),
			       &trans_rates);
	    }

	    if (params.do_rrc) {
	      calc_rad_rec_cont(Te, &debug, Z, rmJ, &(state[rmJ]),
				&(state[rmJ-1]), Ion_pop[rmJ],
				level_rates[rmJ-1].qrecomb, &params, 
				&spectrum);
	    }
	    if (params.do_sat) {
	      calc_satellite(Te, Ne, &debug, &(state[rmJ]),
			     &(state[rmJ-1]),&params,Ion_pop[rmJ],
			     level_rates[rmJ-1].qsatel,&line);
	    }

	    /* Set up rate matrix and inversion. Return Lev_pop with */
	    /* total pop of each ion normalized to one. */
	    
	    if (params.do_lines) {
	      solve_rate_eqns(&debug,Ne,&trans_rates,&(level_rates[rmJ-1]),
			      Lev_pop);
	    }

	    
	    if (params.do_lines) { /* Calculate line emission. */
	      nfree = calc_free_particles(&params);
	      calc_emission(Te, Ne, &debug, &(state[rmJ-1]),&params,
			    Ion_pop[rmJ-1],Lev_pop,nfree,&twophot, &line);
	      
	      calc_two_phot(&debug, &(state[rmJ-1]), &twophot, &spectrum);

	      if ((state[rmJ-1].ion_inf)->nemiss != 0)
		calc_emission_features(&debug, &(state[rmJ-1]), 
				       &params, Te, Ion_pop[rmJ-1], &line, 
				       &spectrum);
	      
	      bin_lines(Z, rmJ, &params, line, &spectrum);
	      free(Lev_pop);
	    }
	  }

	  if ((params.do_brems)&&(Ion_pop[rmJ-1]>MIN_ION_POP)&&(rmJ>1)) {
	    bremsstrahlung(Z, rmJ, Te, Ion_pop, &params, &spectrum);
	  }
	}
	
	if (params.nei_output) 
	  calc_spectrum_per_ion(Z, Ion_pop, line, &spectrum);

	free(Ion_pop);
	dealloc_level_rates(Z, level_rates);
      }  /* Close the Element loop */

      fits_output(&params, Te, Ne, iTemp, iDens, &line, &spectrum);

      clear_spectrum(params.N_Elements, params.list_Z, &spectrum);
      clear_lines(&line);
      
    } /* Close the Te loop */
  }   /* Close the Ne loop */
  
  /* Close all output files */
  close_debug_output(&debug);
  fits_close(&params);

}

void apec_calc_upsilon(void) {

  struct ELEMENT_ION_STATE *state;
  int Z, rmJ, ll, ul;
  SLang_Array_Type *Tarr, *Upsilon;
  int iTemp, NTemp, ii, iEl, nexc;
  double Te;
  struct COLLTRNS_DATA *ptr;
  double degl;   /* statistical weight of lower level */
  double degu;   /* statistical weight of upper level */
  double delE;   /* energy difference */
  double El;     /* energy of lower level */
  double Eu;     /* energy of upper level */
  double *exc_rate, *dex_rate;

  if (SLang_Num_Function_Args != 5 ) {
    SLang_verror(SL_USAGE_ERROR,"Usage: apec_calc_upsilon(Z, rmJ, ll, ul, T)");
    return;
  }

  /* Should check types before popping, but work on that later */
  if ( (SLang_pop_array_of_type(&Tarr, SLANG_DOUBLE_TYPE) == -1 ) ||
       (SLang_pop_integer(&ul) == -1 ) ||
       (SLang_pop_integer(&ll) == -1 ) ||
       (SLang_pop_integer(&rmJ) == -1 ) ||
       (SLang_pop_integer(&Z) == -1 ) ) {
    SLang_vmessage("calc_upsilon: Error in datatypes.");
    return;
  }
 
  /* Add some checks to make sure Z, rmJ are OK. */

  NTemp = Tarr->num_elements;
  exc_rate = (double *) malloc(NTemp*sizeof(double));
  dex_rate = (double *) malloc(NTemp*sizeof(double));

  Upsilon=SLang_create_array(SLANG_DOUBLE_TYPE, 0, NULL, &NTemp, 1);

  for (iTemp=0;iTemp<NTemp;iTemp++) {
    SLang_get_array_element(Tarr, &iTemp, &Te);
    exc_rate[iTemp]=0.0;
    dex_rate[iTemp]=0.0;

    for (iEl=0;iEl < params.N_Elements; iEl++) { /* Find the right data */
      if (Z == params.list_Z[iEl]) {
	state = Atomic_params[iEl];
	nexc = state[rmJ].ion_inf->nexc;
	for (ii=0;ii<nexc;ii++) {
	  ptr = &(state[rmJ].colltrns_dat[ii]);
	  if ((ptr->lev_lower == ll) && (ptr->lev_upper == ul)) {
	    El = state[rmJ].lev_dat[ll-1].energy;
	    Eu = state[rmJ].lev_dat[ul-1].energy;
	    degl = state[rmJ].lev_dat[ll-1].deg;
	    degu = state[rmJ].lev_dat[ul-1].deg;
	    delE = Eu - El;  /* in eV */
	    calc_maxwell_rates(ptr, delE/1.e3, Te, Z, degl, degu, 
			       &(exc_rate[iTemp]), &(dex_rate[iTemp]));
	  }
	}
      }
    }
  }
  for (ii=0;ii<NTemp;ii++) {
    SLang_set_array_element(Upsilon, &ii, &(exc_rate[ii]));
  }
  (void) SLang_push_array(Upsilon, 0);
  SLang_free_array(Upsilon);
  free(exc_rate);
  free(dex_rate);
}
