#include <stdlib.h>
#include <stdio.h>

#include "messages.h"
#include "apec.h"

/*! \file bin_lines.c */

void clear_lines(struct EMISSION_LINES **line);

void compress_energy_spectrum(int Nbin, double *bin_edge,  double *emission, 
			      double error, char *type,
			      int *Nval, double **Eval, double **Fluxval);

/*! \brief Put linked line list emission into spectral bins 
 *  \param Z proton number of element 
 *  \param rmJ ion stage (neutral=1) of recombined ion
 *  \param min_epsilon Minimum value (in erg cm^3/s) to go in main spectrum 
 *  \param line linked list of line emissions
 *  \param spectrum Spectral output structure [optional; set to NULL to ignore]
 */

void bin_lines(int Z, int rmJ, double min_epsilon, 
	       struct EMISSION_LINES *line, struct BIN_SPEC *spectrum) {
  
  long iLine=0, iPos, iBin;
  double lambda, Plambda, E, epsilon;
  double *emission;
  struct COMPRESS_CONT *cs_ptr;

  if (spectrum == NULL) return;

  emission = (double *) malloc_safe(spectrum->Nbin*sizeof(double));
  for (iBin=0;iBin<spectrum->Nbin;iBin++) emission[iBin] = 0.0;

  while (line != NULL) {
    if ((line->Z==Z)&&(line->rmJ==rmJ)) {
      lambda = line->lambda;    /* Angstroms */
      Plambda = line->Plambda;  /* erg cm^3 s^-1 */
      epsilon = line->epsilon;  /* photons cm^3 / s */
      iLine++;
      if ((lambda > 0)&&(Plambda > 0)) {
	E       = HC_IN_KEV_A / lambda; /* keV */
	spectrum->total_emission[Z] += Plambda; /* ergs cm^3 / s */
	spectrum->tot_line_emis[Z] += Plambda; /* ergs cm^3 / s */
	switch(spectrum->spec_type) {
	case (BIN_LIN_WAVE):
	  iPos = (long) floor((lambda-spectrum->bin_min)/spectrum->delta_bin);
	  if ((iPos < 0) || (iPos >= spectrum->Nbin)) {
	    /* message("bin_lines","Line out of range: lambda = %e",lambda); */
	  } else {
	    (spectrum->bin_emission)[Z][iPos] += epsilon;
	    if (line->epsilon < min_epsilon) {
	      (spectrum->pseudo_emiss[Z])[iPos] += epsilon;
	      emission[iPos] += Plambda;
	    }
	  }
	  break;
	case (BIN_LOG_WAVE):
	  iPos = (long) floor((log10(lambda)-spectrum->bin_logmin)/
			      spectrum->delta_bin);
	  if ((iPos < 0) || (iPos >= spectrum->Nbin)) {
	    /* message("bin_lines","Line out of range: lambda = %e",lambda); */
	  } else {
	    (spectrum->bin_emission[Z])[iPos] += epsilon;
	    if (line->epsilon < min_epsilon) {
	      (spectrum->pseudo_emiss[Z])[iPos] += Plambda;
	      emission[iPos] += Plambda;
	    }
	  }
	  break;
	case (BIN_LIN_ENERGY):
	  iPos = (long) floor((E - spectrum->bin_min)/spectrum->delta_bin);
	  if ((iPos < 0) || (iPos >= spectrum->Nbin)) {
	    /* message("bin_lines","Line out of range: E = %e",E); */
	  } else {
	    (spectrum->bin_emission[Z])[iPos] += epsilon;
	    if (line->epsilon < min_epsilon) {
	      (spectrum->pseudo_emiss[Z])[iPos] += epsilon;
	      emission[iPos] += Plambda;
	      /* message("bin_lines","%d %e\n",iPos,Plambda); */
	    }
	  }
	  break;
	case (BIN_LOG_ENERGY):
	  iPos=(long)
	    floor((log10(E)-spectrum->bin_logmin)/spectrum->delta_bin);
	  if ((iPos < 0) || (iPos >= spectrum->Nbin)) {
	    /* message("bin_lines","Line out of range: E = %e",E); */
	  } else {
	    (spectrum->bin_emission[Z])[iPos] += epsilon;
	    if (line->epsilon < min_epsilon) {
	      (spectrum->pseudo_emiss[Z])[iPos] += epsilon;
	      emission[iPos] += Plambda;
	    }
	  }
	  break;
	case (BIN_ARBITRARY):
	  quitmess("bin_lines","Arbitrary binning not ready yet.");
	  break;
	default:
	  errmess("bin_lines","No spectral binning type set!");
	  break;
	}
      } else {
	if (Plambda < 0) { 
	  message("bin_lines","Bad emission in ion %d %d, l = %e, Pl = %e",
		  Z, rmJ, lambda, Plambda);
	}
	if (lambda < 0) {
	  message("bin_lines","Line %d has bad wavelength = %e",iLine,lambda);
	}
	if (Plambda == 0) {
	  /* message("bin_lines","Zero emission in ion %d %d, l = %e, Pl = %e",
	     Z, rmJ, lambda, Plambda); */
	}
      }
    }
    line = line->next;
  }
  /* Now, compress the emission results */
  cs_ptr = spectrum->cs;
  while ((cs_ptr != NULL) && ((cs_ptr->Z != Z) || (cs_ptr->rmJ != rmJ))) 
    cs_ptr = cs_ptr->next;
  if (cs_ptr == NULL) 
    errmess("bin_lines","Ion spectrum for %d, %d not found?",Z,rmJ);

  compress_energy_spectrum(spectrum->Nbin, spectrum->bin_edge, emission, 
			   ERROR_PSEUDO, "Pseudo", &(cs_ptr->Npseudo),
			   &(cs_ptr->E_pseudo),&(cs_ptr->pseudo));
 
  /* And free up the no-longer-needed memory */
  free(emission);
}

void clear_lines(struct EMISSION_LINES **line) {
  
  struct EMISSION_LINES *ptr;

  while (*line != NULL) {
    ptr = *line;
    *line = (*line)->next;
    free(ptr);
  }
}
