#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <fitsio.h>

#include "messages.h"
#include "apec.h"
#include "apec_macros.h"
#include "datatypes.h"

double interpol_huntd(int n, double *x, double *y, double z);
float interpol_huntf(int n, float *x, float *y, float z);
void fits_error(int status);
void init_brems_relativistic(char *filemap_file);
void brems_gaunt_relativistic(double *E,double T,int Z,int N, double *g_ff);
void extract_gauntff(int Z, int NUrow, double gamma2, double **Uvec, 
		     double **GauntFFvec);
void compress_energy_spectrum(int Nbin, double *bin_edge, double *emission, 
			      double error, char *type,
			      int *Nval, double **Eval, double **Fluxval);
void load_filemap(char *filemap_file, struct FILEMAP **filemap);
int get_filename_from_map(struct FILEMAP *filemap, int Z, int J,
			  int datatype, char **filename);

static float *gaunt_U, **gaunt_g2, **gaunt_gf;
static int *gaunt_Z, *gaunt_Ng, Nrow=0;

void brems_gaunt_relativistic(double *E,double T,int Z,int N, double *g_ff) {

  /* We are passed E (keV), T (K), and Z.  We then calculate */
  /*   u = hv/kT, and gamma2 = Z^2 RYDBERG/kT                */
  /* The gaunt factor is calculated from the tables found    */
  /* in Nozawa, Itoh, & Kohyama, 1998, ApJ, 507, 530         */
  /* and updated in astro-ph/9906342                         */
  
  double u, gamma2;
  double *Uvec, *GauntFFvec;
  double *Uvec_l, *GauntFFvec_l;
  double *Uvec_u, *GauntFFvec_u;
  int NUrow=0, NUrow_l=0, NUrow_u=0;
  int iBin, iRow;
  int Zlower, Zupper;

  if (Nrow == 0) {
    errmess("brems_relativistic",
	    "Relativistic bremmstrahlung data not initialized");
  }


  /* message("brems_relativistic","Started: Z=%d, T=%e",Z,T); */
  gamma2 = log10(Z*Z*RYDBERG/(KBOLTZ*T));

  /* So, we're dealing with a particular Z value.  So, time to make up */
  /* some new vectors. */
  for (iRow=0;iRow<Nrow;iRow++) if (gaunt_Z[iRow] == Z) NUrow++;

  if (NUrow > 0) {
    extract_gauntff(Z, NUrow, gamma2, &Uvec, &GauntFFvec);
  } else {  /* No matching Z values.  Have to interpolate */
    Zlower = 0;
    Zupper = 100;

    for (iRow=0;iRow<Nrow;iRow++) {
      if ((gaunt_Z[iRow] < Z)&&( gaunt_Z[iRow] > Zlower)) Zlower=gaunt_Z[iRow];
      if ((gaunt_Z[iRow] > Z)&&( gaunt_Z[iRow] < Zupper)) Zupper=gaunt_Z[iRow];
    }

    for (iRow=0;iRow<Nrow;iRow++) if (gaunt_Z[iRow] == Zlower) NUrow_l++;
    for (iRow=0;iRow<Nrow;iRow++) if (gaunt_Z[iRow] == Zupper) NUrow_u++;

    if ((NUrow_l == 0) && (NUrow_u > 0)) {
      NUrow = NUrow_u;
      extract_gauntff(Zupper, NUrow, gamma2, &Uvec, &GauntFFvec);
    }
    if ((NUrow_l > 0) && (NUrow_u == 0)) {
      NUrow = NUrow_l;
      extract_gauntff(Zlower, NUrow, gamma2, &Uvec, &GauntFFvec);
    }
    if ((NUrow_l == 0) && (NUrow_u == 0)) {
      message("brems_relativistic",
	      "No data found!  No Relativistic bremsstrahlung calculated.");
    }
    if ((NUrow_l > 0) && (NUrow_u > 0)) {
      extract_gauntff(Zlower, NUrow_l, gamma2, &Uvec_l, &GauntFFvec_l);
      extract_gauntff(Zupper, NUrow_u, gamma2, &Uvec_u, &GauntFFvec_u);
      
      /* Now some more scaling work.  Uvec_l and Uvec_u MUST match. */
      
      if (NUrow_l != NUrow_u) errmess("brems_relativistic","Bad U range");
      NUrow = NUrow_l;
      Uvec = (double *) malloc_safe(NUrow*sizeof(double));
      GauntFFvec = (double *) malloc_safe(NUrow*sizeof(double));
      
      for (iRow=0;iRow<NUrow_l;iRow++) {
	/* message("brems_relativistic","Ulow[%d] = %e, Uhigh[%d] = %e",
	   iRow,Uvec_l[iRow],iRow,Uvec_u[iRow]); */
	if (fabs(Uvec_l[iRow] - Uvec_u[iRow]) > 0.001)
	  errmess("brems_relativistic","U vector mismatch at %d in %d",iRow,Z);
	Uvec[iRow] = Uvec_l[iRow];
	GauntFFvec[iRow] = ((Zupper-Z)*GauntFFvec_l[iRow] + 
			    (Z-Zlower)*GauntFFvec_u[iRow])/(Zupper-Zlower);
      }
      free(Uvec_l); free(Uvec_u); free(GauntFFvec_l); free(GauntFFvec_u);
    }
  }

  /* Now, start the calculations */

  for (iBin=0;iBin < N; iBin++) {
    u = log10(E[iBin]/(KBOLTZ*T));

    /* First, some special cases on u */
    /* Long wavelength.  Use approximation from Scheuer 1960, MNRAS, 120,231 */
    /* As noted in Hummer, 1988, ApJ, 325, 477 */

    if (u < Uvec[0]) {
      g_ff[iBin]=-0.55133*(0.5*log(10.)*gamma2+log(10.)*u+0.056745);
    }
    /* Short wavelength.This is approximate, but brems is suppressed by e^-u */
    /* anyway, so getting the gaunt factor wrong by a bit (maybe 10x?) */
    /* isn't a terrific problem. */

    if (u > Uvec[NUrow-1]) g_ff[iBin] = 1.;  

    if ((u >= Uvec[0])&&(u <= Uvec[NUrow-1])) { 
      g_ff[iBin] = interpol_huntd(NUrow, Uvec, GauntFFvec, u);
    }
  }
  free(Uvec); free(GauntFFvec);
}

void init_brems_relativistic(char *filemap_file) { 

  int status=0, hdunum, hdutype, anynul, casesen;
  int col_u, col_z, col_ng, col_g2, col_gf;
  int iRow;
  char column[MAXSTRLEN], comment[MAXSTRLEN];
  fitsfile *fptr;
  long num_rows, firstrow, firstelem;
  void *nulval;
  struct FILEMAP *filemap;
  char *fn_brems_rel;

  if (Nrow != 0) return; /* Already initialized */

  load_filemap(filemap_file, &filemap);
  if (get_filename_from_map(filemap, 0, 0, GAUNT_FF_TYPE, &fn_brems_rel)) {
    errmess("brems_relativistic",
	    "Cannot find Relativistic brems file in filemap %s",filemap_file);
  }
  if (fits_open_file(&fptr, fn_brems_rel, READONLY, &status)) {
    message("brems_hummer","Error opening fits file %s",fn_brems_rel);
    fits_error(status);
  }

  hdunum = 2;  /* Skip the (non-existant) image frame */
  if (fits_movabs_hdu(fptr, hdunum, &hdutype, &status)) {
    message("brems_relativistic","Error opening data array in fits file %s",
	    fn_brems_rel);
    fits_error(status);
  }
  
  casesen = FALSE;
  fits_get_colname(fptr, casesen, GAUNT_FF_U,  column, &col_u,  &status);
  fits_get_colname(fptr, casesen, GAUNT_FF_Z,  column, &col_z,  &status);
  fits_get_colname(fptr, casesen, GAUNT_FF_NG, column, &col_ng, &status);
  fits_get_colname(fptr, casesen, GAUNT_FF_G2, column, &col_g2, &status);
  fits_get_colname(fptr, casesen, GAUNT_FF_GF, column, &col_gf, &status);

  if (status!=0) {
    message("brems_relativistic",
	    "Error getting data column names in fits file %s",fn_brems_rel);
    fits_error(status);
  }

  if (fits_read_key_lng(fptr,"NAXIS2", &num_rows, comment, &status)) {
    message("brems_relativistic","Error getting dimensions in fits file.");
    fits_error(status);
  }

  Nrow = num_rows;
  gaunt_U = (float *)  malloc(num_rows*sizeof(float));
  gaunt_Z = (int *)    malloc(num_rows*sizeof(int));
  gaunt_Ng= (int *)    malloc(num_rows*sizeof(int));
  gaunt_g2= (float **) malloc(num_rows*sizeof(float *));
  gaunt_gf= (float **) malloc(num_rows*sizeof(float *));

  nulval = NULL;
  firstrow = 1;
  firstelem = 1;

  fits_read_col(fptr, TFLOAT, col_u, firstrow, firstelem, num_rows,
		nulval, gaunt_U, &anynul, &status);
  fits_read_col(fptr, TINT, col_z, firstrow, firstelem, num_rows,
		nulval, gaunt_Z, &anynul, &status);
  fits_read_col(fptr, TINT, col_ng, firstrow, firstelem, num_rows,
		nulval, gaunt_Ng, &anynul, &status);

  if (status != 0) {
    message("init_brems_relativistic","Error reading fits file");
    fits_error(status);
  }

  for (iRow=0;iRow<num_rows;iRow++) {
    gaunt_g2[iRow] = (float *) malloc(gaunt_Ng[iRow]*sizeof(float));
    gaunt_gf[iRow] = (float *) malloc(gaunt_Ng[iRow]*sizeof(float));
    if (fits_read_col(fptr, TFLOAT, col_g2, iRow+1, firstelem, gaunt_Ng[iRow],
		      nulval, gaunt_g2[iRow], &anynul, &status)) {
      message("init_brems_relativistic","Error, row %d getting gamma2 in %s.",
	      iRow, fn_brems_rel);
      fits_error(status);
    }
    if (fits_read_col(fptr, TFLOAT, col_gf, iRow+1, firstelem, gaunt_Ng[iRow],
		      nulval, gaunt_gf[iRow], &anynul, &status)) {
      message("init_brems_relativistic","Error, row %d getting gf in %s.",
	      iRow, fn_brems_rel);
      fits_error(status);
    }
  }

  if (fits_close_file(fptr, &status)) {
    message("brems_relativistic","Error closing fits file %s.",fn_brems_rel);
    fits_error(status);
  }

  /* At this point, the data has been loaded into the appropriate variables.*/

  /* message("brems_relativistic",
             "Got the bremsstrahlung gaunt factor data");*/
}

void extract_gauntff(int Z, int NUrow, double gamma2, 
		     double **Uvec, double **GauntFFvec) {
  
  int iPos, iRow;
  float gamma2_f;

  *Uvec       = (double *) malloc_safe(NUrow*sizeof(double));
  *GauntFFvec = (double *) malloc_safe(NUrow*sizeof(double));
  iPos = 0;
  gamma2_f = gamma2;
  for (iRow=0;iRow<Nrow;iRow++) {
    if (gaunt_Z[iRow] == Z) {
      (*Uvec)[iPos] = gaunt_U[iRow];
      if ((gamma2 >= (gaunt_g2[iRow])[0]) && 
	  (gamma2 <= (gaunt_g2[iRow])[gaunt_Ng[iRow]-1])) {
	(*GauntFFvec)[iPos] = interpol_huntf(gaunt_Ng[iRow], gaunt_g2[iRow],
					     gaunt_gf[iRow], gamma2_f);
      } else {
	/* Outside range of loaded values.  Set to default value */
	message("extract_gauntff","Warning, gamma^2 = %e is out of range.",
		gamma2);
	if (gamma2 < gaunt_g2[iRow][0]) 
	  (*GauntFFvec)[iPos] = gaunt_gf[iRow][0];
	if (gamma2 > gaunt_g2[iRow][gaunt_Ng[iRow]-1])
	  (*GauntFFvec)[iPos] = gaunt_gf[iRow][gaunt_Ng[iRow]-1];
      }
      iPos++;
    }
  }
}
