#include <stdlib.h>
#include <stdio.h>

#include "apec.h"
#include "messages.h"
#define ERRLIMIT 1.e-10  /* Rather big, actually, for double precision */

char *roman_element(int Z, int J);
void allocate_level_rates(int nlev, struct LEVEL_RATES *level_rates);
extern int dgesv_(int *ndim, int *nrhs, double *a, int *lda, int *ipiv,
		  double *b, int *ldb, int *info);
void dgemv_(char *trans, int *ndimX, int *ndimY, double *alpha, double *AAch, 
	    int *lda, double *B, int *incx, double *beta, double *Bch,
	    int *incy);
void calc_ionrec_rate(double Te, double Ne, double dt, 
		      struct DEBUG_FILES *debug,
		      struct ELEMENT_ION_STATE *state, double init_lev_pop[], 
		      double ioniz_rate[], double *recomb_rate,
		      struct EMISSION_LINES **ionrec_line);

void calc_ionbal(double Te, double Ne, struct ELEMENT_ION_STATE *state,
		 struct DEBUG_FILES *debug,
		 struct LEVEL_RATES *level_rates, double Ion_pop[],
		 struct EMISSION_LINES **ionrec_line) {

  /* Goal here is to calculate an ionization balance directly, and at the
     same time calculate any emission lines due to ion-to-ion transitions
     (eg, Fe I fluorescence, O VI excitation/ionization).  Note that the 
     general idea is identical to the level population calculation; we have
     one population equation Sum(Ion_States) = 1.0, and a bunch of rate 
     calculations.  */

  double *init_lev_pop;
  double ion_recomb_rate, *ion_ioniz_rate;
  /* ion_recomb_rate = rate of recombinations from rmJ+1 to rmJ */
  /* ion_ioniz_rate  = rate of ionizations from rmJ to rmJ+N */
  int *nlev;
  int Z, rmJ, iCol, iRow, j;
  int rmJ_init, rmJ_final;
  double *AA, *B, *Bdebug, *AAch, *Bch, *Rch;
  char trans;
  double alpha, beta, Err, LevSum; /* Check values */
  int incx, incy;
  int N, nrhs, lda, ldb, *ipiv, info, ndim;  /* Variables needed by dgesv */

  /* Concept here: 
     
     A_ij P_i = B_i 
     
     where A_ij = rate from state i to state j [ A_{col,row} is the format ]
           A_ij = AA[(i-1)*(Z+1) + j-1]
     except A_i0 = 1.0 for all i   
     
     In this case, if i != 0,
           A_ii = - Sum ( Transitions out of i level ) 
  */

  /* First, get all the rates calculated; Z+1 x Z+1 matrix  */

  *ionrec_line = NULL;
  Z      = state[0].ion_inf->Z;
  AA     = (double *) malloc_safe((Z+1)*(Z+1)*sizeof(double));
  B      = (double *) malloc_safe((Z+1)*sizeof(double));
  Bdebug = (double *) malloc_safe((Z+1)*sizeof(double));
  for (iCol=0;iCol<Z+1;iCol++) {
    B[iCol] = 0.0;
    for (iRow=0;iRow<Z+1;iRow++) AA[iCol*(Z+1)+iRow] = 0.0;
  }
    
  /* Set the population conservation equation, sum(p_i) = 1.0 */
  for (j=0;j<Z+1;j++) AA[j*(Z+1) + 0] = 1.0;
  B[0] = 1.0;

  nlev = (int *) malloc_safe((Z+1)*sizeof(int));
  ion_ioniz_rate = (double *) malloc_safe((Z+1)*sizeof(double));

  for (rmJ_init=1;rmJ_init<Z+1;rmJ_init++) { /* Calculating a column */

    /* Do some setup */
    for (j=0;j<Z+1;j++) ion_ioniz_rate[j] = 0.0;
    nlev[rmJ_init-1] = state[rmJ_init-1].ion_inf->nlev;
    if (nlev[rmJ_init-1] != 0) {
      init_lev_pop = (double *) malloc_safe(nlev[rmJ_init-1]*sizeof(double));
    } else {
      init_lev_pop = (double *) malloc_safe(sizeof(double));
    }
    allocate_level_rates(nlev[rmJ_init-1], &(level_rates[rmJ_init-1]));

    /* Calculate all rates into and out of Z, rmJ */

    calc_ionrec_rate(Te, Ne, 0.0, debug, &(state[rmJ_init-1]), init_lev_pop, 
		     ion_ioniz_rate, &ion_recomb_rate, ionrec_line);
    
    free(init_lev_pop);

    /* Now, put the results where they should go: */

    /* Ionization (including diagonal elements) */ 
    for (rmJ_final=rmJ_init+1;rmJ_final<=Z+1;rmJ_final++) {
      AA[(rmJ_init-1)*(Z+1) + rmJ_final-1] +=  ion_ioniz_rate[rmJ_final-1];
      if (rmJ_init != 1) {
	AA[(rmJ_init-1)*(Z+1) + rmJ_init-1]  += -ion_ioniz_rate[rmJ_final-1];
      }
    }
    /* Recombination (include diagonal elements) */
    AA[(rmJ_init)*(Z+1) + rmJ_init-1] += ion_recomb_rate;
    AA[(rmJ_init)*(Z+1) + rmJ_init] += -ion_recomb_rate;
   
  }
  free(ion_ioniz_rate);

  /* OK, now we've got the arrays set up.  Time to solve for the ion balance */
  ndim = Z+1;
  nrhs = 1;
  lda = ndim;
  ldb = ndim;
  N   = ndim;
  ipiv = (int *) malloc_safe(ndim*sizeof(int));
  /* Set up the backup variables for checking. */
  AAch = (double *) malloc_safe(ndim*ndim*sizeof(double));
  Bch  = (double *) malloc_safe(ndim*sizeof(double));
  Rch  = (double *) malloc_safe(ndim*sizeof(double));
  for (iCol=0;iCol<ndim;iCol++) {
    for (iRow=0;iRow<ndim;iRow++) {
      AAch[iCol*ndim + iRow]=AA[iCol*ndim + iRow];
      if (debug->do_ionbal && debug->ascii) {
	fprintf(debug->ionbal.ascii,"AA[%d,%d] = %8g \n",iCol,iRow,
		AA[iCol*ndim + iRow]);
      }
    }
    Bch[iCol] = B[iCol];
    Bdebug[iCol] = B[iCol];
    if (debug->do_ionbal && debug->ascii) {
      fprintf(debug->ionbal.ascii,"B[%d] = %8g \n",iCol,B[iCol]);
      fflush(debug->ionbal.ascii);
    }
    Rch[iCol] = 0.0;
  }

  /* We're all set.  Now just do the inversion. */

  (void) dgesv_(&N, &nrhs, AA, &lda, ipiv, B, &ldb, &info);
  
  if (info != 0) {
    if (info < 0) {
      message("calc_ionbal", "Error solving matrix for ion %s", 
	      roman_element(Z,-info-1));
    } else {
      message("calc_ionbal","Error solving matrix: U(%d,%d) = 0",
	      info, info);
    }
  }
  
  /* Now check the result */
  trans = 'N';
  lda = ndim;
  alpha = 1.0;
  beta =  0.0;
  incx = 1;
  incy = 1;

  dgemv_(&trans,&ndim, &ndim, &alpha, AAch, &lda, B, &incx, &beta, Rch, &incy);

  /* Check the matrix results */
  LevSum = 0.0;

  for (iRow=0;iRow<ndim;iRow++) {
    LevSum += B[(ndim-1)-iRow];
    if (Rch[iRow] == Bch[iRow]) {
      Err = 0.0;
    } else {
      Err = fabs((Rch[iRow] - Bch[iRow])/(Rch[iRow] + Bch[iRow]));
      if (Bch[iRow] == 0.0) Err = fabs(Rch[iRow]);
    }
    if (Err > ERRLIMIT) {
      message("calc_ionbal","Pop[%3d] = %e, Rch=%e, Bch=%e, Error=%e",
	      iRow, B[iRow], Rch[iRow], Bch[iRow], Err);
    }
  }
  if (fabs(LevSum - 1.) > ERRLIMIT) {
    message("calc_ionbal","Error: Total level population is %e.",LevSum);
  }
   
  if (debug->do_ionbal && debug->ascii) {
    fprintf(debug->ionbal.ascii,"%4d: %8g%8g ",Z, log10(Te),log10(Ne));
  }

  /* Set the Lev_pop array equal to solution vector B.  */
  for (rmJ=1; rmJ <= Z+1; rmJ++) {
    if (debug->do_ionbal && debug->ascii) 
      fprintf(debug->ionbal.ascii,"%8g ",B[rmJ-1]);
    if (B[rmJ-1] < 0.0) {
      message("calc_ionbal","Err:Level pop for Z=%d, rmJ=%d is %e",
	      Z, rmJ, B[rmJ-1]);
      message("calc_ionbal","Set to zero.");
      B[rmJ-1] = 0.0;
    }
    Ion_pop[rmJ-1] = B[rmJ-1];
  }

  if (debug->do_ionbal && debug->ascii) {
    fprintf(debug->ionbal.ascii,"\n");
    fflush(debug->ionbal.ascii);
  }

  if (debug->do_ionbal && debug->fits) {
    for (iCol=0;iCol<ndim;iCol++) {
      for (iRow=0;iRow<ndim;iRow++) {
	fd_write_row(&debug->ionbal.fits,
		     Z, 
		     log10(Te),
		     log10(Ne),
		     iCol, iRow,
		     AA[iCol*ndim + iRow],
		     Bdebug[iCol],
		     B[iCol]);
      }
    }
  }

  free(B);
  free(Bdebug);
  free(AA);
  free(AAch);
  free(Bch);
  free(Rch);
  free(ipiv);

}
