#include <stdlib.h>

#include "apec.h"
#include "datatypes.h"
#include "messages.h"
#define TRUE 1
#define FALSE 0

double f1_fcn(double x);
double f2_fcn(double x);
char *roman_element(int Z, int J);
double calc_ionrec_rr(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line);
double calc_ionrec_dr(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line);
double calc_ionrec_ci(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line);
double calc_ionrec_ea(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line);

void calc_ionrec_rate(double Te, double Ne, double dt, 
		      struct DEBUG_FILES *debug,
		      struct ELEMENT_ION_STATE *state, 
		      double init_lev_pop[], 
		      double ioniz_rate[], double *recomb_rate,
		      struct EMISSION_LINES **ionrec_line) {
  
  /* **********************************************************************/
  /* Input parameters:                                                    */
  /* Te : in Kelvin, the electron temperature in the plasma               */
  /* Ne : in cm^-3, the electron density in the plasma                    */
  /* dt : in seconds (0 = equilibrium), the time step to calculate        */
  /* *debug : output file pointers                                        */
  /* *state : the atomic state vector for this particular ion             */
  /* *init_lev_pop : vector for the level population of the ion initially */
  /*                  (not currently used)                                */
  /* Output Parameters:                                                   */
  /* ioniz_rate[] : ionization rate out of this ion to others(in cm^3/s)  */
  /* e.g.  ioniz_rate[1] to singly-ionized atom                           */
  /* e.g.  ioniz_rate[2] to doubly-ionized atom                           */
  /* *recomb_rate : recombination rate of this ion  (in cm^3/s)           */
  /* *ionrec_line : emission line output from these transitions.          */
  /* **********************************************************************/
  
  
  int iTr, i;
  int Z;
  struct IONREC_DATA *ptr;
  double rr=0.0, dr=0.0, ci, ea;
  double tot_rr=0.0, tot_ci=0.0, tot_ea=0.0;

  Z = state->ion_inf->Z;
  *recomb_rate = 0.0; /* not vector; only dealing with e + I, not e+e+I,etc */
  for (i=0;i<Z+1;i++) ioniz_rate[i] = 0.0; 
  
  for (iTr = 0; iTr < state->ion_inf->nionrec; iTr++) {

    ptr = &(state->ionrec_dat[iTr]);
    switch (ptr->tr_type) {
    case (IONREC_TYPE_RR):
      rr = calc_ionrec_rr(Z, Te, Ne, init_lev_pop[ptr->level_init-1], ptr,
			  ionrec_line);
      tot_rr = rr;
      if (rr < 0.0) {
	message("calc_ionrec_rate","RR(%10s,T=%9.3e) = %8g",
		roman_element(Z,state->ion_inf->rmJ-1),Te,rr);
      }
      *recomb_rate += rr; 
      break;
    case (IONREC_TYPE_DR):
      dr = calc_ionrec_dr(Z, Te, Ne, init_lev_pop[ptr->level_init-1], ptr,
			  ionrec_line);
      if (dr < 0.0) {
	message("calc_ionrec_rate","DR(%10s,T=%9.3e) = %8g",
		roman_element(Z,state->ion_inf->rmJ-1),Te,dr);
      }
      *recomb_rate += dr;
      break;
    case (IONREC_TYPE_CI):
      ci = calc_ionrec_ci(Z, Te, Ne, init_lev_pop[ptr->level_init-1], ptr,
			  ionrec_line);
      if (ci < 0.0) {
	message("calc_ionrec_rate","CI(%10s -> %10s,T=%9.3e) = %8g",
		roman_element(Z,state->ion_inf->rmJ-1),
		roman_element(Z,ptr->ion_final-1),
		Te,ci);
      }
      ioniz_rate[ptr->ion_final-1] += ci;
      tot_ci += ci;
      break;
    case (IONREC_TYPE_EA):
      ea = calc_ionrec_ea(Z, Te, Ne, init_lev_pop[ptr->level_init-1], ptr,
			  ionrec_line);
      if (ea < 0.0) {
	message("calc_ionrec_rate","EA(%10s,T=%9.3e) = %8g",
		roman_element(Z,state->ion_inf->rmJ-1),Te,ea);
      }
      ioniz_rate[ptr->ion_final-1] += ea;
      tot_ea += ea;
      break;
    default:
      errmess("get_ioniz_recomb","Illegal collision type");
    }
  }
  printf("%d  %e %e %e\n", Z,tot_rr,*recomb_rate-tot_rr, tot_ci+tot_ea); 
}

double calc_ionrec_rr(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line) {
  
  double rr=0.0;
  double *c;
  double tt0,tt1;
  
  c = ptr->IonRec_Par;
  
  switch (ptr->par_type) {
  case (RR_SHULL): /* Shull & Van Steenberg 1982; 1982ApJS...48...95S  */
    rr = c[0]*pow(Te/1.e4,-c[1]);
    break;
  case (RR_VERNER): /* Verner & Ferland 1996; 1996ApJS..103..467V */
    tt0=sqrt(Te/c[2]);
    tt1=sqrt(Te/c[3]);
    rr = c[0]/( tt0*pow((tt0+1), 1-c[1]) * pow(1+tt1, 1+c[1]) );
    /* printf("%e %e %e %e rr=%e\n",c[0],c[1],c[2],c[3],rr); */
    break;
  case (RR_ARNRAY): /* Arnaud & Raymond 1992; 1992ApJ...398..394A  */
    rr = c[0]*pow(Te/1.e4,-(c[1]+c[2]*log10(Te/1.e4)));
    break;
  default:
    errmess("calc_ionrec_rate","Recombination type not recognized");
    break;
  }
  return rr;
}

double calc_ionrec_dr(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line) {
  
  double T_eV;
  double dr=0.0;
  double *c;
  
  T_eV = 1.e3*KBOLTZ*Te;
  c = ptr->IonRec_Par;

  switch (ptr->par_type) {
  case (DR_MAZZOTTA) :
    dr = (c[0]/(pow(T_eV,1.5)))*(c[5]*exp(-c[1]/T_eV) +
				 c[6]*exp(-c[2]/T_eV) +
				 c[7]*exp(-c[3]/T_eV) +
				 c[8]*exp(-c[4]/T_eV));
    break;
  default:
    errmess("calc_ionrec_rate",
	    "Dielectronic recombination rate not recognized!");
    break;
  }
  return dr;
  
}

double calc_ionrec_ci(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line) {
  
  double x, T_eV, *c;
  double f1_val;
  double ci = 0.0;
  
  c = ptr->IonRec_Par;

  switch (ptr->par_type) {
  case (CI_YOUNGER): /* See Arnaud \& Rothenflug, 1985 A&ASS 60, 425 */
    if ((Te >= ptr->min_temp) && (Te <= ptr->max_temp)) {
      T_eV = 1.e3*KBOLTZ*Te;
      x = c[0]/T_eV;
      if (x <= 30.0) {
	f1_val = f1_fcn(x);
	ci = (exp(-x)/x)*
	  ( c[1]*( 1 - x*f1_val ) +
	    c[2]*( 1 + x - x*(x+2)*f1_val )+
	    c[3]*f1_val + 
	    c[4]*x*f2_fcn(x) );
      }
      ci *= 6.69e-07/(T_eV*sqrt(T_eV));
    }
    break;
  default:
    errmess("calc_ionrec_rate","No method given!");
  }

  return ci;
}

double calc_ionrec_ea(int Z, double Te, double Ne, double levpop, 
		      struct IONREC_DATA *ptr, struct EMISSION_LINES **line) {

  double T_eV;
  double *c;
  double y=0.0, a, b, G, yf1, f1_val=0.0;
  double ea = 0.0;
  double I_ea;
  
  c = ptr->IonRec_Par;
  T_eV = 1.e3*KBOLTZ*Te;

  switch (ptr->par_type) {
  case (EA_ARNROTH_LITHIUM): /* 1985A&AS...60..425A (Arnaud & Rothenflug) */
    /* Note: c[1] = b/Zeff^2 in Arnaud & Rothenflug */
    y = c[0]/T_eV;
    if (y < 50.0) { /* otherwise, rate is effectively zero */
      yf1 = y*f1_fcn(y);
      G = 2.22*f1_fcn(y) + 0.67*(1 - yf1) + 0.49*yf1 + 1.2*y*(1 - yf1);
      ea = c[2] * c[1]  * 1.92e-07 * exp(-y) * G/sqrt(T_eV);
    }
    break;
  case (EA_ARNROTH): /* 1985A&AS...60..425A (Arnaud & Rothenflug) */
    I_ea = c[0]; /* in eV */
    a = c[1];
    b = c[2];
    y =  I_ea/T_eV;
    if (y < 50.0) {
      f1_val = f1_fcn(y);
      ea = 6.69e7 * a * (I_ea/sqrt(T_eV)) * exp(-y) * 
	(1.0 + b*f1_val - c[3]*y*f1_val - 
	 c[4]*0.5*(y - y*y + y*y*y*f1_val));
    }
    break;

  case (EA_MAZZOTTA_IRON): /* '1998A&AS..133..403M (Mazzotta etal 1998) */
    y=c[0]/T_eV;
    if (y < 50.0) {
      f1_val = f1_fcn(y);
      ea = (6.69e7/sqrt(T_eV))*exp(-y)*1.e-16*
	(c[1]+
	 c[2]*(1-y*f1_val)+
	 c[3]*(1-y*(1-y*f1_val))+
	 c[4]*(1-0.5*(y - y*y + y*y*y*f1_val))+
	 c[5]*f1_val);
    }
    break;
  default:
    errmess("calc_ionrec_rate","No excitation-ionization-rate listed");
    break;
  }
  /* printf("f1(%e) = %e, ea = %f\n",y,f1_val,ea); */
  return ea;
}

double f1_fcn(double x) {

  /* Analytic approx.  to f_1(x) = exp(x) \int_1^{\infty} (exp(-xt)/t) dt */
  /* Commented-out version from                                           */
  /*   Arnaud \& Rothenflug, 1985 A&A, Appendix B, p. 436                 */
  /* Version used from Mazzotta's original code.                          */
  
  double result=0.0;

  double a[6] = {-0.57721566,0.99999193,-0.24991055,
		 0.05519968,-0.00976004,0.00107857};
  double b[8] = {8.5733287401,18.0590169730,8.6347608925,0.2677737343,
		 9.5733223454,25.6329561486,21.0996530827,3.9584969228};

  if (x < 0.0 ) {
    errmess("f1_fcn","Negative values of x not allowed");
  }

  /* 
  if (x <= 0.02) {
    result = exp(x)*(log(x) - 0.5772 + x);
  } else {
    if (x < 1.5) {
      result = log((x+1)/x) - (0.36 + 0.03/sqrt(x + 0.01))/((x-1)*(x-1));
    } else {
      if (x < 10) {
	result = log((x+1)/x) - (0.36 + 0.03*sqrt(x + 0.01))/((x-1)*(x-1));
      } else {
	result = (1 - 1/x + 2/(x*x) - 6/(x*x*x) + 24/(x*x*x*x));
      }
    }
  }
  */
  if (x < 1.0) {
    result=(a[0]+a[1]*x+a[2]*x*x + a[3]*pow(x,3)+ a[4]*pow(x,4)+ 
	    a[5]*pow(x,5)-log(x))*exp(x);
  } else {
    if (x < 1.0e9) {
      result=(pow(x,4)+b[0]*pow(x,3)+b[1]*pow(x,2)+b[2]*x+b[3])/
	(pow(x,4)+b[4]*pow(x,3)+b[5]*pow(x,2)+b[6]*x+b[7])/x;
    } else {
      result=1.0/x;
    }
  }
  /* printf("%e %e\n",result, aa); */

  return result;
}
      

double f2_fcn(double x) {
  
  /* Analytic approximation to                                     */
  /* f_2(x) = exp(x) \int_1^{\infty} (log(t) exp(-xt)/t) dt      */
  /* Taken from Arnaud \& Rothenflug, 1985 A&A, Appendix B, p. 436 */
  
  
  double pc[15] = {1.0000e0 ,2.1658e2 ,2.0336e4 ,1.0911e6 ,3.7114e7 ,
		   8.3963e8 ,1.2889e10,1.3449e11,9.4002e11,4.2571e12,
		   1.1743e13,1.7549e13,1.0806e13,4.9776e11,0.0000e0};
  double qc[15] = {1.0000e0 ,2.1958e2 ,2.0984e4 ,1.1517e6 ,4.0349e7 ,
		   9.4900e8 ,1.5345e10,1.7182e11,1.3249e12,6.9071e12,
		   2.3531e13,4.9432e13,5.7760e13,3.0225e13,3.3641e12};
  double p, q, xp, result;
  int i;
  
  if (x > 0.27) {
    p=0.0;
    q=0.0;
    xp = 1;
    for(i=0;i<15;i++) {
      p += pc[i]/xp;
      q += qc[i]/xp;
      xp *= x;
    }
      result=((p/q)/x)/x;
  } else {
    result=(((log(x)*log(x))/2.0)+(0.57722*log(x)))+1.0;
  }
  return result;
}
