#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "messages.h"
#include "apec.h"

void clear_spectrum(int N_Z, int list_Z[], struct BIN_SPEC *spectrum);

void create_spectral_bins(long BinType, long Nbin, 
			  double BinMin, double BinMax, 
			  int N_Elements, int *list_Z, 
			  struct BIN_SPEC *spectrum) {

  long i;
  int iZ, Z;
  double logBmin, logBmax, logdB;
  
  spectrum->spec_type = BinType;
  spectrum->Nbin = Nbin;
  spectrum->bin_min = BinMin;
  spectrum->bin_max = BinMax;
  spectrum->bin_logmin = 0.0;
  spectrum->bin_logmax = 0.0;
  if (spectrum->bin_min > 0.0) spectrum->bin_logmin = log10(BinMin);
  if (spectrum->bin_max > 0.0) spectrum->bin_logmax = log10(BinMax);
  Nbin = spectrum->Nbin;

  spectrum->bin_edge = (double *) malloc((Nbin+1)*sizeof(double));

  for (iZ=0;iZ<N_Elements;iZ++) {
    Z = list_Z[iZ];  
    spectrum->bin_emission[Z] = (double *) malloc(Nbin*sizeof(double));
    spectrum->brems_emiss[Z]  = (double *) malloc(Nbin*sizeof(double));
    spectrum->pseudo_emiss[Z] = (double *) malloc(Nbin*sizeof(double));
    spectrum->rrc_emiss[Z]    = (double *) malloc(Nbin*sizeof(double));
    spectrum->twoph_emiss[Z]  = (double *) malloc(Nbin*sizeof(double));
    spectrum->bin_error[Z]    = (double *) malloc(Nbin*sizeof(double));
    if (spectrum->bin_error[Z]==NULL) 
      errmess("create_spectral_bins","Memory allocation error.");
    spectrum->total_emission[Z] = 0.0;
    spectrum->tot_line_emis[Z] = 0.0;
    spectrum->tot_brems_emis[Z] = 0.0;
    spectrum->tot_rrc_emis[Z] = 0.0;
    spectrum->tot_twoph_emis[Z] = 0.0;
    for (i=0; i < spectrum->Nbin; i++) {  /* Zero the arrays */
      spectrum->bin_emission[Z][i] = 0.0;
      spectrum->brems_emiss[Z][i] = 0.0;
      spectrum->pseudo_emiss[Z][i] = 0.0;
      spectrum->rrc_emiss[Z][i] = 0.0;
      spectrum->twoph_emiss[Z][i] = 0.0;
      spectrum->bin_error[Z][i] = 0.0;
    }
    spectrum->cs = NULL;
  }

  if (spectrum->bin_min >= spectrum->bin_max) {
    quitmess("create_spectral_bins","Bin minimum (%e) >= Bin maximum (%e)",
	     spectrum->bin_min, spectrum->bin_max);
  }

  switch (spectrum->spec_type) {
  case (BIN_LIN_WAVE):
    spectrum->delta_bin = (spectrum->bin_max - spectrum->bin_min)/Nbin;
    for (i=0; i < Nbin; i++) {
      spectrum->bin_edge[i] = spectrum->bin_min + i*spectrum->delta_bin;
    }
    spectrum->bin_edge[Nbin] = spectrum->bin_max;
    break;
  case (BIN_LOG_WAVE):
    if (spectrum->bin_min == 0.0) {
      errmess("create_spectral_bins","Zero minimum bin but log binning.");
    }
    logBmin = spectrum->bin_logmin;
    logBmax = spectrum->bin_logmax;
    logdB = (logBmax - logBmin)/Nbin;
    spectrum->delta_bin = logdB;
    for (i=0; i < Nbin; i++) {
      spectrum->bin_edge[i] = pow(10.0, logBmin + i*logdB);
    }
    spectrum->bin_edge[Nbin] = pow(10.0, logBmax);
    break;
  case (BIN_LIN_ENERGY):
    spectrum->delta_bin = (spectrum->bin_max - spectrum->bin_min)/Nbin;
    for (i=0; i < Nbin; i++) {
      spectrum->bin_edge[i] = spectrum->bin_min + i*spectrum->delta_bin;
    }
    spectrum->bin_edge[Nbin] = spectrum->bin_max;
    break;
  case (BIN_LOG_ENERGY):
    if (spectrum->bin_min == 0.0) {
      errmess("create_spectral_bins","Zero minimum bin but log binning.");
    }
    logBmin = spectrum->bin_logmin;
    logBmax = spectrum->bin_logmax;
    logdB = (logBmax - logBmin)/Nbin;
    spectrum->delta_bin = logdB;
    for (i=0; i < Nbin; i++) {
      spectrum->bin_edge[i] = pow(10.0, logBmin + i*logdB);
    }
    spectrum->bin_edge[Nbin] = pow(10.0, logBmax);
    break;
  case (BIN_ARBITRARY):
    errmess("create_spectral_bins","Arbitrary binning not ready yet.");
    break;
  default:
    errmess("create_spectral_bins","Error, no binning type defined.");
  }
}

void alloc_spectrum(int Z, int rmJ, struct BIN_SPEC *spectrum) {

  /* This routine allocates potentially substantial amounts of memory */
  /* into the bin_real element.  This is then cleared by the */
  /* clear_spectrum routine, called after everything is output to a file. */

  int iBin;
  struct COMPRESS_CONT *csptr;

  csptr = spectrum->cs;
  if (csptr == NULL) {
    spectrum->cs = (struct COMPRESS_CONT *) 
      malloc_safe(sizeof(struct COMPRESS_CONT));
    csptr = spectrum->cs;
  } else {
    while (csptr->next != NULL) csptr = csptr->next;
    csptr->next = (struct COMPRESS_CONT *) 
      malloc_safe(sizeof(struct COMPRESS_CONT));
    csptr = csptr->next;
  }
  csptr->next = NULL;
  csptr->Z   = Z;
  csptr->rmJ = rmJ;
  csptr->Nbrems = 0;
  csptr->E_brems = NULL;
  csptr->brems = NULL;
  csptr->Ntwoph = 0;
  csptr->E_twoph = NULL;
  csptr->twoph = NULL;
  csptr->Nrrc = 0;
  csptr->E_rrc = NULL;
  csptr->rrc = NULL;
  csptr->Nreal = 0;
  csptr->E_real = NULL;
  csptr->real = NULL;
  csptr->Npseudo = 0;
  csptr->E_pseudo = NULL;
  csptr->pseudo = NULL;
  csptr->bin_real = (double *) malloc((spectrum->Nbin)*sizeof(double));
  for (iBin=0;iBin<spectrum->Nbin;iBin++) csptr->bin_real[iBin] = 0.0;
}

void clear_spectrum(int N_Z, int list_Z[], struct BIN_SPEC *spectrum) {

  int iZ, Z, i;
  struct COMPRESS_CONT *csptr, *csptr_next;

  for (iZ=0;iZ<N_Z;iZ++) {
    Z = list_Z[iZ];
    spectrum->total_emission[Z] = 0.0;
    spectrum->tot_line_emis[Z] = 0.0;
    spectrum->tot_brems_emis[Z] = 0.0;
    spectrum->tot_rrc_emis[Z] = 0.0;
    spectrum->tot_twoph_emis[Z] = 0.0;

    csptr = spectrum->cs;
    while (csptr != NULL) {
      if (csptr->E_brems != NULL) free(csptr->E_brems);
      if (csptr->brems != NULL) free(csptr->brems);
      if (csptr->E_twoph != NULL) free(csptr->E_twoph);
      if (csptr->twoph != NULL) free(csptr->twoph);
      if (csptr->E_rrc != NULL) free(csptr->E_rrc);
      if (csptr->rrc != NULL) free(csptr->rrc);
      if (csptr->E_real != NULL) free(csptr->E_real);
      if (csptr->real != NULL) free(csptr->real);
      if (csptr->E_pseudo != NULL) free(csptr->E_pseudo);
      if (csptr->pseudo != NULL) free(csptr->pseudo);
      if (csptr->bin_real != NULL) free(csptr->bin_real);
      csptr_next = csptr->next;
      free(csptr);
      csptr = csptr_next;
    }
    spectrum->cs = NULL;
    
    for (i=0; i < spectrum->Nbin; i++) {  /* Zero the arrays */
      spectrum->bin_emission[Z][i] = 0.0;
      spectrum->brems_emiss[Z][i] = 0.0;
      spectrum->pseudo_emiss[Z][i] = 0.0;
      spectrum->rrc_emiss[Z][i] = 0.0;
      spectrum->twoph_emiss[Z][i] = 0.0;
      spectrum->bin_error[Z][i] = 0.0;
    }
  }
}


