#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "apec.h"
#include "messages.h"

int strip_strcmp(const char *s1, const char *s2);

void setup_debug(char *debugformat, char *debugfitsformat,
		 struct DEBUG_FILES *debug) {

  if (strip_strcmp(debugformat,"fits")==0) {
    debug->ascii = 0;
    debug->fits  = DEBUG_DO_FITS;
  }
  if (strip_strcmp(debugformat,"ascii")==0) {
    debug->ascii = DEBUG_DO_ASCII;
    debug->fits  = 0;
  }
  if (strip_strcmp(debugformat,"both")==0) {
    debug->ascii = DEBUG_DO_ASCII;
    debug->fits  = DEBUG_DO_FITS;
  }
  if (strip_strcmp(debugfitsformat,"binary")==0) {
    debug->table_type  = BINARY_TBL; 
  } else {
    debug->table_type  = ASCII_TBL; 
  }
  /* standard setting */
  debug->ionbal_read_or_calc = IONBAL_CALC;

}

void open_debug_output(struct DEBUG_FILES *debug) {
  
  char filename[MAXSTRLEN];
  char fitsfilename[MAXSTRLEN];
  int status = 0;
  int flush = 1;
  int supersede = 1;
  
  /* Check to see if debug=NULL */
  if (debug == NULL) return;

  /* Now start setting up files for debug output */
  if (debug->do_emiss && debug->ascii) {
    strcpy(filename,"emiss");
    strcat(filename,".DATA");
    debug->emiss.ascii = fopen(filename,"w");
    fprintf(debug->emiss.ascii," Z  J  logNe  logTe  lu  ll   lambda Lph  Lerg\n");
  }
  if (debug->do_emiss && debug->fits) {
    strcpy(fitsfilename,"emiss");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->emiss.fits, fitsfilename, debug->table_type,
		       "Emission", &status, flush, supersede, 9,
		       "I5",     "Atomic number",      "",
		       "I5",     "Ionization Level",   "Roman numerals",
		       "D18.10", "Log10(Density)",     "Log10(cm^-3)",
		       "D18.10", "Log10(Temperature)", "Log10(K)",
		       "I5",     "Lower level",        "",
		       "I5",     "Upper level",        "",
		       "D18.10", "Wavelength",         "Angstrom",
		       "D18.10", "Volume emissivity",  "phot cm^-3",
		       "D18.10", "Volume emissivity",  "erg cm^-3")) {
	quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_ionbal && debug->ascii) {
    strcpy(filename,"ionbal");
    strcat(filename,".DATA");
    debug->ionbal.ascii = fopen(filename,"w"); 
  }
  if (debug->do_ionbal && debug->fits) {
    strcpy(fitsfilename,"ionbal");
    strcat(fitsfilename,".FITS");
    if (debug->ionbal_read_or_calc == IONBAL_READ) {
      if (fd_create_file(&debug->ionbal.fits, fitsfilename, debug->table_type,
			 "Ionbalance", &status, flush, supersede, 5,
			 "D12.5",  "Temperature", "K",
			 "I5",     "Atomic number", "",
			 "I5",     "State", "",
			 "D12.5",  "Ionization Balance", "Unitless",
			 "D12.5",  "Sum Lev. pop.", "")) {
	quitmess("open_debug_output","Error opening %s\n",fitsfilename);
      }
    } else if (debug->ionbal_read_or_calc == IONBAL_CALC) {
      if (fd_create_file(&debug->ionbal.fits, fitsfilename, debug->table_type,
			 "Ionbalance", &status, flush, supersede, 8,
			 "I6",     "Atomic number", "",
			 "D12.5",  "Log10(Temperature)", "log10(K)",
			 "D12.5",  "Log10(Density)", "log10(cm^-3)",
			 "I6",     "iCol", "",
			 "I6",     "iRow", "",
			 "D12.5",  "Transition matrix AA[iCol,iRow]", "",
			 "D12.5",  "B[iCol]", "",
			 "D12.5",  "B[iCol]", "")) {
	quitmess("open_debug_output","Error opening %s\n",fitsfilename);
	}
    } else {
      quitmess("open_debug_output","ionbal_read_or_calc undefined (%d)",
	       debug->ionbal_read_or_calc);
    }
  }

  if (debug->do_table && debug->ascii) {
    strcpy(filename,"table");
    strcat(filename,".DATA");
    debug->table.ascii = fopen(filename,"w");
  }
  if (debug->do_table && debug->fits) {
    strcpy(fitsfilename,"table");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->table.fits, fitsfilename, debug->table_type,
		       "Table", &status, flush, supersede, 42,
		       /* start columns (9 cols) */
		       "A7",    "Ion",              "",
		       "I4",    "Atomic number",    "Roman numerals",
		       "I5",    "Ionization level", "",
		       "I5",    "# ion./recomb. transitions", "",
		       "I5",    "# levels for model ion", "",
		       "I8",    "# radiative trans. calculated", "",
		       "I8",    "# collisional trans. calculated", "",
		       "I5",    "# proton trans. calculated", "",
		       "I5",    "# satelltie lines available", "",
		       /* loop over ion/recomb lines (8 cols) */
		       "I8",      "Final ion. state", "",
		       "I7",      "Final level", "",
		       "I7",      "Initial level", "",
		       "I7",      "Transition type", "",
		       "I7",      "Transition index", "",
		       "D12.5",   "Min. Temperature", "K",
		       "D12.5",   "Max. Temperature", "K",
		       "D12.5",   "Ion./recomb. parameter [0]", "",
		       /* loop over levels  (7 cols)*/
		       "I4",      "ID for level", "",
		       "I5",      "Statistical weight, or degeneracy", "",
		       "D12.5",   "Energy above ground state", "eV",
		       "I5",      "Principle quantum number", "",
		       "I5",      "Orbital quantum number", "",
		       "F6.2",   "Spin quantum number", "",
		       "I5",      "Type of photoionization coeff.", "",
		       /* loop over radiative transitions (4 cols) */
		       "I3",      "Upper Level", "",
		       "I3",      "Lower Level", "",
		       "D12.5",   "Wavelength", "Angstrom",
		       "D12.5",   "Einstein prop. coefficient A", "",
		       /* loop over collisional transitions (8 cols) */
		       "I4",      "Upper level", "",
		       "I3",      "Lower level", "",
		       "I5",      "Collision type", "",
		       "D12.5",   "Fit to coll. strength formula m[0]", "",
		       "D12.5",   "Fit to coll. strength formula m[1]", "",
		       "D12.5",   "Fit to coll. strength formula m[2]", "",
		       "D12.5",   "Fit to coll. strength formula m[3]", "",
		       "D12.5",   "Fit to coll. strength formula m[4]", "",
		       /* loop over satellite lines  (6 cols) */
		       "I4",      "Upper level", "",
		       "I3",      "Lower level", "",
		       "I5",      "Data type", "",
		       "D12.5",   "Wavelength", "Angstrom",
		       "D12.5",   "q_exc", "K^3/2 cm^3 s^-1",
		       "D12.5",   "E_excite", "keV")) {
      quitmess("open_debug_output","Error opening %s, %d\n",fitsfilename, status);
    }
  }

  if (debug->do_qioniz && debug->ascii) {
    strcpy(filename,"qioniz");
    strcat(filename,".DATA");
    debug->qioniz.ascii = fopen(filename,"w");
    fprintf(debug->qioniz.ascii,"chi  alpha  beta  fchi\n");
  }
  if (debug->do_qioniz && debug->fits) {
    strcpy(fitsfilename,"qioniz");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->qioniz.fits, fitsfilename, debug->table_type,
		       "Ionization", &status, flush, supersede, 5,
		       "D12.5", "chi (Reduced temperature)", "",
		       "D12.5", "alpha", "",
		       "D12.5", "beta", "",
		       "D12.5", "fchi (distorted wavefactor)", "",
		       "D12.5", "Ionization rate", "")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_qrec && debug->ascii) {
    strcpy(filename,"qrec");
    strcat(filename,".DATA");
    debug->qrec.ascii = fopen(filename,"w");
  }
  if (debug->do_qrec && debug->fits) {
    strcpy(fitsfilename,"qrec");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->qrec.fits, fitsfilename, debug->table_type,
		       "Recombinatioin", &status, flush, supersede, 7,
		       "I4",    "Atomic Number", "",
		       "I4",    "Ionization level", "Roman numerals",
		       "D12.5", "Effective temperature", "K",
		       "I5",    "Level", "",
		       "D12.5", "Radiative recomb rate", "",
		       "D12.5", "Total recomb rate", "",
		       "D12.5", "Tot. recomb rate (obsolete)", "")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_qtrans && debug->ascii) {
    strcpy(filename,"qtrans");
    strcat(filename,".DATA");
    debug->qtrans.ascii = fopen(filename,"w");
  }
  if (debug->do_qtrans && debug->fits) {
    strcpy(fitsfilename,"qtrans");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->qtrans.fits, fitsfilename, debug->table_type,
		       "Transition", &status, flush, supersede, 14,
		       "I6",    "Atomic Number", "",
		       "I6",    "Ionization level", "Roman numerals",
		       "D12.5", "Delta Energie", "eV",
		       "I6",    "Lower level", "",
		       "I6",    "Upper level", "",
		       "D12.5", "Log10(Density)", "log10(cm^-3)",
		       "D12.5", "Log10(Temperature)", "log10(K)",
		       "I6",    "Ionization level", "Roman numerals",
		       "I6",    "Lower level", "",
		       "I6",    "Upper level", "",
		       "D12.5", "q[ll][lu]", "",
		       "D12.5", "q[lu][ll]", "",
		       "D12.5", "qp[ll][lu]", "",
		       "D12.5", "Arad[lu][ll]", "")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_rates && debug->ascii) {
    strcpy(filename,"rates");
    strcat(filename,".DATA");
    debug->rates.ascii = fopen(filename,"w");
  }
  if (debug->do_rates && debug->fits) {
    strcpy(fitsfilename,"rates");
    strcat(fitsfilename,".FITS");
    /* Currently nothing is being written to file so no need
       to open anything */
  }

  if (debug->do_satel && debug->ascii) {
    strcpy(filename,"satel");
    strcat(filename,".DATA");
    debug->satel.ascii = fopen(filename,"w");
  }
  if (debug->do_satel && debug->fits) {
    strcpy(fitsfilename,"satel");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->satel.fits, fitsfilename, debug->table_type,
		       "Satellite", &status, flush, supersede, 9,
		       "I6",    "Atomic Number", "",
		       "I6",    "Ionization level", "Roman numerals",
		       "I6",    "Upper level", "",
		       "I6",    "Lower level", "",
		       "D12.5", "Temperature", "K",
		       "D12.5", "Wavelength", "Angstrom",
		       "D12.5", "Excitation energy", "keV",
		       "D12.5", "q_exc", "K^3/2 cm^3 s^-1",
		       "D12.5", "Atomic factor", "photons cm^3 s^-1")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_twoph && debug->ascii) {
    strcpy(filename,"twoph");
    strcat(filename,".DATA");
    debug->twoph.ascii = fopen(filename,"w");
    fprintf(debug->twoph.ascii,"  Z   J   A   Emission\n");
  }
  if (debug->do_twoph && debug->fits) {
    strcpy(fitsfilename,"twoph");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->twoph.fits, fitsfilename, debug->table_type,
		       "Twophoton", &status, flush, supersede, 4,
		       "I6",    "Atomic Number", "",
		       "I6",    "Ionization level", "Roman numerals",
		       "D12.5", "Einstein coefficient A", "",
		       "D12.5", "Total emission", "")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_levpop && debug->ascii) {
    strcpy(filename,"levpop");
    strcat(filename,".DATA");
    debug->levpop.ascii = fopen(filename,"w");
  }
  if (debug->do_levpop && debug->fits) {
    strcpy(fitsfilename,"levpop");
    strcat(fitsfilename,".FITS");
    if (fd_create_file(&debug->levpop.fits, fitsfilename, debug->table_type,
		       "Levelpopulation", &status, flush, supersede, 10,
		       "I6",    "Atomic Number", "",
		       "I6",    "Ionization level", "Roman numerals",
		       "A10",   "Ion", "", 
		       "D12.5", "Log10(Temperature)", "log10(K)",
		       "D12.5", "Log10(Density)", "log10(cm^-3)",
		       "I6",    "I", "Counter",
		       "D12.5", "Recombination rate RR[i]", "",
		       "D12.5", "Satelitte recom. rate DR[i]", "",
		       "D12.5", "Level population B[i]", "",
		       "D12.5", "Delta lev pop. B[i - 1]", "")) {
      quitmess("open_debug_output","Error opening %s\n",fitsfilename);
    }
  }

  if (debug->do_compress && debug->ascii) {
    strcpy(filename,"compress");
    strcat(filename,".DATA");
    debug->compress.ascii = fopen(filename,"w");
  }
  if (debug->do_compress && debug->fits) {
    strcpy(fitsfilename,"compress");
    strcat(fitsfilename,".FITS");
    /* Currently nothing is being written to file so no need
       to open anything */
  }

}

void close_debug_output(struct DEBUG_FILES *debug) {

  int status = 0;

  if (debug != NULL) {
    if (debug->do_emiss    && debug->ascii) fclose(debug->emiss.ascii);
    if (debug->do_ionbal   && debug->ascii) fclose(debug->ionbal.ascii);
    if (debug->do_table    && debug->ascii) fclose(debug->table.ascii);
    if (debug->do_qioniz   && debug->ascii) fclose(debug->qioniz.ascii);
    if (debug->do_qrec     && debug->ascii) fclose(debug->qrec.ascii);
    if (debug->do_qtrans   && debug->ascii) fclose(debug->qtrans.ascii);
    if (debug->do_rates    && debug->ascii) fclose(debug->rates.ascii);
    if (debug->do_satel    && debug->ascii) fclose(debug->satel.ascii);
    if (debug->do_twoph    && debug->ascii) fclose(debug->twoph.ascii);
    if (debug->do_levpop   && debug->ascii) fclose(debug->levpop.ascii);
    if (debug->do_compress && debug->ascii) fclose(debug->compress.ascii);
    
    /* FITSDEBUG */
    if (debug->do_emiss    && debug->fits) 
      fd_exit_file(&debug->emiss.fits, &status);
    if (debug->do_ionbal   && debug->fits)
      fd_exit_file(&debug->ionbal.fits, &status);
    if (debug->do_table    && debug->fits) 
      fd_exit_file(&debug->table.fits, &status);
    if (debug->do_qioniz   && debug->fits)
      fd_exit_file(&debug->qioniz.fits, &status);
    if (debug->do_qrec     && debug->fits)  
      fd_exit_file(&debug->qrec.fits, &status);
    if (debug->do_qtrans   && debug->fits)
    fd_exit_file(&debug->qtrans.fits, &status);
    if (debug->do_rates    && debug->fits) 
      fd_exit_file(&debug->rates.fits, &status);
    if (debug->do_satel    && debug->fits) 
      fd_exit_file(&debug->satel.fits, &status);
    if (debug->do_twoph    && debug->fits) 
      fd_exit_file(&debug->twoph.fits, &status);
    if (debug->do_levpop   && debug->fits) 
      fd_exit_file(&debug->levpop.fits, &status);
    if (debug->do_compress && debug->fits) 
      fd_exit_file(&debug->compress.fits, &status);
  }
}

void write_debug(struct DEBUG_GENERIC *debug, int ascii, int fits,
		 char* fmt, ...) {


  va_list args;

  if (fits) {
    va_start(args, fmt);
    fd_vwrite_row(&debug->fits, &args);
    va_end(args);
  }

  if (ascii) {
    va_start(args, fmt);
    vfprintf(debug->ascii, fmt, args);
    va_end(args);
  }
  
}
