#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <fitsio.h>

#include "messages.h"
#include "apec.h"
#include "datatypes.h"

/* Abundance FITS file information */
#define SOURCE "Source"
#define SOURCE_LEN 10
#define ABUND_TYPE 10  /* The file type in the filemap file. */
#define MAX_ROWS 100  /* Maximum number of abundance types */

void fits_error(int status);
char *element_name(int Z);
void load_filemap(char *filemap_file, struct FILEMAP **filemap);
int get_filename_from_map(struct FILEMAP *filemap, int Z, int J,
			  int datatype, char **filename);

void get_abund(char *abund_source, char *filemap_file, double **abund) {

  int Z, nfound;
  long num_rows;
  long naxes[MAX_ROWS];
  long firstrow, firstelem;
  fitsfile *fptr;
  int hdunum, colnum, casesen;
  char column[8];
  char *source_names[MAX_ROWS];
  char *fn_abundance;
  char comment[80];
  int status, hdutype, anynul;
  int iRow, RowMatch;
  int num_el;
  
  struct FILEMAP *filemap;

  void *nulval;
  double *data_ab;

  status = 0;
  
  if (abund_source[0] == '\0') {
    errmess("get_abund","No abundance specified.");
  }

  *abund = (double *) malloc((MAXIMUM_Z+1)*sizeof(double)); 
  if (abund == NULL) {
    errmess("get_abund","Malloc error for abund variable");
  }

  load_filemap(filemap_file, &filemap);
  if (get_filename_from_map(filemap, 0, 0, ABUND_TYPE, &fn_abundance)) {
    errmess("get_abund","Cannot find abundance file in filemap %s",
	    filemap_file);
  }

  if (fits_open_file(&fptr, fn_abundance, READONLY, &status)) {
    message("get_abund","Error opening fits file.");
    fits_error(status);
  }

  hdunum = 2;
  if (fits_movabs_hdu(fptr, hdunum, &hdutype, &status)) {
    message("get_abund","Error opening data array in fits file.");
    fits_error(status);
  }

  /* First, figure out which row should be read. */

  casesen = TRUE;
  if (fits_get_colname(fptr, casesen, SOURCE, column, &colnum, &status)) {
    message("get_abund","Error getting source column name in fits file.");
    fits_error(status);
  }
  
  if (fits_read_keys_lng(fptr,"NAXIS", 1, 2, naxes, &nfound, &status)) {
    message("get_abund","Error getting dimensions in fits file.");
    fits_error(status);
  }

  num_rows = naxes[1];

  /* message("get_abund","naxes = %d, %d",(int) naxes[0],(int) naxes[1]); */
  if (num_rows > MAX_ROWS) {
    errmess("get_abund","Too many rows (%d), max is %d",num_rows,MAX_ROWS);
  }

  for (iRow=0;iRow<num_rows;iRow++) {
    source_names[iRow] = (char *) malloc(SOURCE_LEN*sizeof(char));
  }

  nulval = NULL;
  firstrow = 1;
  firstelem = 1;

  if (fits_read_col(fptr, TSTRING, colnum, firstrow, firstelem, num_rows,
		    nulval, source_names, &anynul, &status)) {
    message("get_abund","Error getting source column in fits file.");
    fits_error(status);
  }

  RowMatch = -1;
  for (iRow=0; iRow < num_rows; iRow++) {
    if ((strcmp(source_names[iRow], abund_source))==0) {
      if (RowMatch != -1) {
	errmess("get_abund","Multiple hits on Abundance source in file %s",
		fn_abundance);
      } else {
	RowMatch = iRow;
      }
    }
  }

  if (RowMatch == -1) {
    errmess("get_abund","Matching abundance type not found.");
  }

  /* Now we've found the correct row.  Time to read in the data. */

  /* if (fits_get_num_cols(fptr, &num_el, &status)) { */
  if (fits_read_key(fptr, TINT, "TFIELDS", &num_el, comment, &status)) {
    message("get_abund","Error getting number of columns in fits file %s.",
	    fn_abundance);
    fits_error(status);
  }
  num_el--;
  /* message("get_abund","Num_el = %d",num_el); */

  data_ab = (double *) malloc((num_el+1)*sizeof(double));
  /* message("get_abund","column number = %d",colnum);*/


  for (Z = 1; Z < num_el+1; Z++) {
    if (fits_get_colname(fptr, casesen, element_name(Z), column, &colnum, 
			 &status)) {
      message("get_abund","Error getting element %i column name in file %s.",
	      Z, fn_abundance);
      fits_error(status);
    }
    /* message("get_abund","Column for %s, Z=%d is %d",
       element_name(Z),Z,colnum); */

    if (fits_read_col(fptr, TDOUBLE, colnum, RowMatch+1, firstelem, 1,
		      nulval, &(data_ab[Z]), &anynul, &status)) { 
      message("get_abund","Error getting column data in fits file.");
      fits_error(status);
    }
    

    (*abund)[Z] = pow(10.,(data_ab[Z]-data_ab[1]));
    /* message("get_abund","abundance[%d] = %e",Z,abund[Z]); */
  }

  if (fits_close_file(fptr, &status)) {
    message("get_abund","Error closing fits file.");
    fits_error(status);
  }

  /* message("get_abund","Got the abundance data");  */
  free(data_ab);
  for (iRow=0;iRow<num_rows;iRow++) free(source_names[iRow]);
}


