PRO write_ioniz_gr, filename, Z, rmJ, data, version=version, COMMENTS=comments
;
;  Writes a Ioniz_gr FITS file using the data in the structure data,
;  for element Z and ion stage rmJ.  The code first checks to see if
;  filename exists.  If it doesn't, it creates filename and then
;  writes the primary header.  If it does, the code checks to make
;  sure this particular ion stage hasn't already been written to the
;  file.  Assuming all is OK, we are positioned at the end of the
;  file.  Then the header is created, and the data is written to the
;  file.
;

;
;  Check to see if file exists
;
result = findfile(filename,count=count)
if (count eq 0) then begin      ; must create file and primary HDU
    fxhmake,  primary_header, /extend, /initialize, /date
    fxaddpar, primary_header, 'IONIZ_GR','Ground state ionization data'
    fxaddpar, primary_header, 'FILENAME','IDL_routine','Source of data'
    fxaddpar, primary_header, 'ORIGIN','SAO/ASC','Origin of FITS file'
    fxaddpar, primary_header, 'HDUCLASS','ATOMIC','Atomic data'
    fxaddpar, primary_header, 'HDUCLAS1','IONIZ_GR',$
      'Ground state ionization data'
    if (n_elements(version) eq 0) then version = '1.0.0'
    fxaddpar, primary_header, 'HDUVERS1', version, 'Version of datafile'
    fxwrite, filename, primary_header
endif

if (count gt 1) then begin
    print,'More than 1 file exists with this name: ',result
    stop
endif

if (count eq 1) then begin
    Ext = match_header(filename, Z, rmJ, 'IONIZ_GR')
    if (Ext gt 0) then begin
        print,'Extension ',Ext,' already has data for Z, rmJ = ',Z,rmJ,$
          ' in file ',filename
        return
    endif
endif
;
;  Now we're all set, ready to create the new header.
;
integer = 1L
realvar = 1.0
array = fltarr(20)
smarr = fltarr(10)
strings = '012345678901234567890123456789012345678901234567890123456789'
nrows = n_elements(data)
extname = ionname(Z,rmJ)
fxbhmake, header, nrows, extname, 'Ion Name'
fxaddpar, header, 'HDUCLASS', 'ATOMIC', 'Atomic Data'
fxaddpar, header, 'HDUCLAS1','IONIZ_GR','Ground state ionization data'
fxaddpar, header, 'ELEMENT', Z, 'Number of protons in element'
fxaddpar, header, 'ION_STAT',rmJ-1,'Ion state (0 = neutral)'
fxaddpar, header, 'ION_NAME',extname, 'Ion Name'
fxaddpar, header, 'N_SHELL',nrows,'Number of shells'
if (n_elements(version) eq 0) then version = '1.0.0'
fxaddpar, header, 'HDUVERS1', version, 'Version of datafile'
fxbaddcol, 1, header, integer, 'N_E_SH'
fxbaddcol, 2, header, '0123456789', 'Sh_Label'
fxbaddcol, 3, header, integer, 'N_q'
fxbaddcol, 4, header, integer, 'L_q'
fxbaddcol, 5, header, realvar, 'Energy',TUNIT='keV'
fxbaddcol, 6, header, realvar, 'E_Error',TUNIT='keV'
fxbaddcol, 7, header, integer, 'CollType'
fxbaddcol, 8, header, smarr  , 'CollCoef'
fxbaddcol, 9, header, integer, 'PhotType'
fxbaddcol,10, header, smarr  , 'PhotCoef'
fxbaddcol,11, header, integer, 'RR_Type'
fxbaddcol,12, header, smarr  , 'RR_Coef'
fxbaddcol,13, header, integer, 'DR_Type'
fxbaddcol,14, header, array  , 'DR_Coef'
fxbaddcol,15, header, strings, 'Reference'
;
; Now add any additional comments
;
    for i=0,n_elements(comments)-1 do begin
        if (strpos(comments(i),'COMMENT') eq 0) then begin
            commstr = strmid(comments(i),10)
            fxaddpar, header, 'COMMENT', commstr
        endif
    endfor
;
;  Now open the file and write the new header
;
fxbcreate, unit, filename, header
;
;  Write the data
;
for i=0L,nrows-1 do begin
    fxbwrite, unit, data(i).N_E_SH,    1, i+1
    fxbwrite, unit, data(i).Sh_Label,  2, i+1
    fxbwrite, unit, data(i).N_q,       3, i+1
    fxbwrite, unit, data(i).L_q,       4, i+1
    fxbwrite, unit, data(i).Energy,    5, i+1
    fxbwrite, unit, data(i).E_Error,   6, i+1
    fxbwrite, unit, data(i).CollType,  7, i+1
    fxbwrite, unit, data(i).CollCoef,  8, i+1
    fxbwrite, unit, data(i).PhotType,  9, i+1
    fxbwrite, unit, data(i).PhotCoef, 10, i+1
    fxbwrite, unit, data(i).RR_Type,  11, i+1
    fxbwrite, unit, data(i).RR_Coef,  12, i+1
    fxbwrite, unit, data(i).DR_Type,  13, i+1
    fxbwrite, unit, data(i).DR_Coef,  14, i+1
    fxbwrite, unit, data(i).Reference,15, i+1
endfor
;
; Close the file
;
fxbfinish, unit

return
end
