PRO write_ionrec, filename, Z, rmJ, data, COMMENTS=comments, NANVALUE=nanvalue
;
;  Writes a IONREC FITS file using the data in the structure data, for
;  element Z and ion stage rmJ.  The code first checks to see if
;  filename exists.  If it doesn't, it creates filename and then
;  writes the primary header.  If it does, the code checks to make
;  sure this particular ion stage hasn't already been written to the
;  file.  Assuming all is OK, we are positioned at the end of the file.
;  Then the header is created, and the data is written to the file.
;

;
;  Check to see if file exists
;
result = findfile(filename,count=count)
if (count eq 0) then begin      ; must create file and primary HDU
    fxhmake, primary_header, /extend, /initialize, /date
    fxaddpar, primary_header, 'IONREC','Ionization/Recombination Rates'
    fxaddpar, primary_header, 'FILENAME','IDL_routine','Source of data'
    fxaddpar, primary_header, 'ORIGIN','CXC','Chandra X-ray Center'
    fxaddpar, primary_header, 'HDUCLASS','ATOMIC','Atomic data'
    fxaddpar, primary_header, 'HDUCLAS1','IONREC',$
      'Ionization/Recombination rates'
    version = '1.1.0'
    fxaddpar, primary_header, 'HDUVERS1', version, 'Version of datafile'
    fxwrite, filename, primary_header
endif

if (count gt 1) then begin
    print,'More than 1 file exists with this name: ',result
    stop
endif

if (count eq 1) then begin
    Ext = match_header(filename, Z, rmJ, 'IONREC')
    if (Ext gt 0) then begin
        print,'Extension ',Ext,' already has data for Z, rmJ = ',Z,rmJ,$
          ' in file ',filename
        return
    endif
endif
;
;  Now we're all set, ready to create the new header.
;
integer = 1L
float = 1.0
array = fltarr(20)
strings  = '01234567890123456789'
smstring = '01'
nrows = n_elements(data)
extname = ionname(Z,rmJ)
fxbhmake, header, nrows, extname, 'Ion Name'
fxaddpar, header, 'HDUCLASS', 'ATOMIC', 'Atomic Data'
fxaddpar, header, 'HDUCLAS1','IONREC','Ionization/recombination rates'
fxaddpar, header, 'HDUCLAS2','ION_STATE','Individual Ion State'
fxaddpar, header, 'ELEMENT', Z, 'Number of protons in element'
fxaddpar, header, 'ION_STAT',rmJ-1,'Ion state (0 = neutral)'
fxaddpar, header, 'ION_NAME',extname, 'Ion Name'
fxaddpar, header, 'N_ION',nrows,'Number of rates'
version = '1.1.0'
fxaddpar, header, 'HDUVERS1', version, 'Version of datafile'
fxbaddcol, 1, header, integer, 'Ion_Final'
fxbaddcol, 2, header, integer, 'Level_Final'
fxbaddcol, 3, header, integer, 'Level_Init'
fxbaddcol, 4, header, smstring,'Tr_type'
fxbaddcol, 5, header, integer, 'Tr_index'
fxbaddcol, 6, header, integer, 'Par_Type'
fxbaddcol, 7, header, float,   'Min_Temp'
fxbaddcol, 8, header, float,   'Max_Temp'
fxbaddcol, 9, header, array,   'Temperature'
fxbaddcol,10, header, array,   'IonRec_Par'
fxbaddcol,11, header, float,   'Wavelen'
fxbaddcol,12, header, float,   'Wave_Obs'
fxbaddcol,13, header, float,   'Wave_Err'
fxbaddcol,14, header, float,   'Br_Ratio'
fxbaddcol,15, header, float,   'Br_Rat_Err'
fxbaddcol,16, header, strings, 'Label'
fxbaddcol,17, header, strings, 'Rate_Ref'
fxbaddcol,18, header, strings, 'Wave_Ref'
fxbaddcol,19, header, strings, 'Wv_Obs_Ref'
fxbaddcol,20, header, strings, 'Br_Rat_Ref'
;
; Now add any additional comments
;
for i=0,n_elements(comments)-1 do begin
    if (strpos(comments(i),'COMMENT') eq 0) then begin
        commstr = strmid(comments(i),8)
        add_reference, header, commstr, /header
    endif else begin
        add_reference, header, comments(i), /header
    endelse
endfor
;
;  Now open the file and write the new header
;
fxbcreate, unit, filename, header
;
;  Write the data
;
structnames = tag_names(data)

for i=0L,nrows-1 do begin
    fxbwrite, unit, data(i).Ion_Final  , 1, i+1
    fxbwrite, unit, data(i).Level_Final, 2, i+1
    fxbwrite, unit, data(i).Level_Init , 3, i+1
    fxbwrite, unit, data(i).Tr_type    , 4, i+1
    fxbwrite, unit, data(i).Tr_index   , 5, i+1
    fxbwrite, unit, data(i).Par_Type   , 6, i+1
    fxbwrite, unit, data(i).Min_Temp   , 7, i+1
    fxbwrite, unit, data(i).Max_Temp   , 8, i+1
    fxbwrite, unit, data(i).Temperature, 9, i+1
    fxbwrite, unit, data(i).IonRec_Par ,10, i+1
    fxbwrite, unit, data(i).Wavelen    ,11, i+1, NANVALUE=nanvalue
    fxbwrite, unit, data(i).Wave_Obs   ,12, i+1, NANVALUE=nanvalue
    fxbwrite, unit, data(i).Wave_Err   ,13, i+1, NANVALUE=nanvalue
    fxbwrite, unit, data(i).Br_Ratio   ,14, i+1, NANVALUE=nanvalue
    fxbwrite, unit, data(i).Br_Rat_Err ,15, i+1, NANVALUE=nanvalue
    fxbwrite, unit, data(i).Label      ,16, i+1
    fxbwrite, unit, data(i).Rate_Ref   ,17, i+1
    fxbwrite, unit, data(i).Wave_Ref   ,18, i+1
    fxbwrite, unit, data(i).Wv_Obs_Ref ,19, i+1
    fxbwrite, unit, data(i).Br_Rat_Ref ,20, i+1
endfor
;
; Close the file
;
fxbfinish, unit

return
end
