%
% Routine to save (to file) and load (from file) apec models
% within sherpa.
%
% Ronnie Hoogerwerf, SAO, 06/19/2002
%

uses("SL-str","SL-ll","SL-io");

variable model_separator = "--------------------------------------";
variable standard_model_file = "Apec_model_file.txt";
variable standard_continuum_file = "Apec_continuum_file.txt";
variable _save_continuum = NULL;
variable _save_message = 1;

define AM_save_model() {

   % 
   % Shortcut to spead up calculation. This flag prevents the 
   % apec function from evaluating and just returns 0 for the
   % model flux. Since we are only interested in the modelnames,
   % parameters, and dataset numbers this is a valid shortcut.
   % Furthermore at this moment (06/19/2002) there is still a 
   % bug in the get_mdl_models() function with respect to 
   % get_dataset_number().
   %
   AM->getmdlmodels = 1;

   % get input parameter
   variable usage = "Usage: AM_save_model([filename [, modelname]]);";
   variable file, thismodel = NULL;
   
   switch (_NARGS) 
   { case 0 : file = standard_model_file; }
   { case 1 : file = (); }
   { case 2 : (file, thismodel) = (); }
   { message(usage); _pop_n(_NARGS); return; }

   % set verbosity to 0
   variable verbose = AM->verbose;
   AM->verbose = 0;

   variable mdl = get_mdl_models();
   variable n = length(mdl.src), i, start, end;
   variable models = Assoc_Type[LList_Type], ll;
   variable p1, p2, modelname, dataset, tmp, nchoice, answer;

   % loop over mdl structure to find apec models and store them
   % in an associated array of linked lists.
   %
   for (i = 0; i < n; i++) {
      if (0 != string_match(mdl.model[i],"apec",1)) {
	 % Hurray: Found apec model
	 % Now get its name
	 p1 = Str->position('[',mdl.model[i]);
	 p2 = Str->position(']',mdl.model[i]);
	 modelname = substr(mdl.model[i],p1+2,p2-p1-1);
	 dataset = mdl.src[i];
	 if (assoc_key_exists(models,modelname) == 1) {
	    !if (LL->member(dataset,models[modelname])) { 
	       models[modelname] = LL->append(dataset,models[modelname]);
	    }
	 } else {
	    ll = @LList_Type;
	    ll.content = dataset;
	    models[modelname] = ll;
	 }
      }
   }
   message("\n\n\nIGNORE above messages. These are needed to determine");
   message("the apec-model characteristics currently in use by sherpa.\n\n");


   if (_save_continuum == 1) {
      % if we are saving a continuum model using AM_save_continuum
      % we will ask some questions to find out if the model(s) of
      % the continuum is valid for all datasets or only for the datasets
      % associated with the apec models. 
      
      variable k, vs, v, mdls, extra;
      message("\n\n");
      if (_save_message != NULL) {
	 message("You are about to save the following model(s) as a continuum:");
      } else {
	 message("You are about to set the following model(s) as a continuum:");
      }
      mdls = assoc_get_keys(models);
      foreach (models) using ("keys","values") {
	 (k,vs) = ();
	 tmp = "";
	 foreach (vs) {
	    v = ();
	    tmp = tmp + string(v.content) + " ";
	 }
	 vmessage("* apec model \"%s\" for dataset(s) %s",k,tmp);
      }
      if (_save_message != NULL) {
	 message("\nDo you want to save the continuum :");
      } else {
	 message("\nDo you want to set the continuum :");
      }
      message("- using the model(s) and dataset(s) specified above    [1]");
      extra = 24;
      % Handle the different options
      if (Utils->assoc_length(models) > 1) {
	 % there is more than one model defined.

	 tmp = "";
	 for (i = 0; i < length(mdls); i++) {
	    if (tmp == "") {
	       tmp = tmp + mdls[i];
	    } else {
	       tmp = tmp + " + " + mdls[i];
	    }
	    vmessage("- using model %s for all datasets%s[%d]",
		     mdls[i],Str->empty_string(extra-strlen(mdls[i])),i+2);
	 }
	 vmessage("- using model %s for all datasets%s[%d]",
		  tmp,Str->empty_string(extra-strlen(tmp)),i+2);
	 vmessage("- do nothing%s[%d]",Str->empty_string(extra+19),i+3);
	 nchoice = i+3;
	 answer = -1;
	 while (answer < 1 or answer > nchoice) {
	     IO->get_user_reply(&answer,"  Your choice [1] : ",
				Integer_Type);
	     if (answer == NULL) answer = 1;
	 }
	 % if (answer == 1) do nothing and procede
	 if (answer == nchoice) {
	    return;
	 } else if (answer == nchoice-1) {
	    % save all models but for all datasets
	    foreach(models) using ("keys") {
	       k = (); 	    
	       ll = @LList_Type;
	       ll.content = 0;
	       models[k] = ll;
	    }
	 } else if (answer > 1) {
	    % remove all models except chosen one and set datasets to all
	    foreach(mdls) {
	       k = (); 	    
	       if (k == mdls[answer-2]) {
		  ll = @LList_Type;
		  ll.content = 0;
		  models[k] = ll;
	       } else {
		  assoc_delete_key(models,k);
	       }
	    }
	 }

      } else {
	 % there is only one model
	    
	 vmessage("- using model %s for all datasets%s[%d]",
		  mdls[0],Str->empty_string(extra-strlen(mdls[0])),2);
	 vmessage("- do nothing%s[%d]",Str->empty_string(extra+19),3);
	 nchoice = 3;
	 answer = -1;
	 while (answer < 1 or answer > nchoice) {
	    IO->get_user_reply(&answer,"  Your choice [1] : ",
			       Integer_Type);
	    if (answer == NULL) answer = 1;
	 }
	 % if (answer == 1) do nothing and procede
	 if (answer == 2) {
	    % save all models but for all datasets
	    ll = @LList_Type;
	    ll.content = 0;
	    models[mdls[0]] = ll;
	 } else if (answer == 3) {
	    return;
	 }
      }
   } % End of code to handle saving the continuum
   
   % Now that the modelnames are know we need to extract the parameter
   % values from the mdl structure and put those in an isis-model structure
   % so that we can save it to file.

   % 1: open file for output
   variable j, fp = fopen(file,"w");
   if (NULL == fp) verror("AM_save_model: unable to open file %s.",file);

   % Write caution message about editing the file.
   () = fprintf(fp,"#\n# WARNING: Caution... Although it is possible to edit this file by hand\n");
   () = fprintf(fp,"#          one should be careful and only edit the values on the right-hand-side\n");
   () = fprintf(fp,"#          of the colon. DO NOT add to or remove whole lines from this file. \n");
   () = fprintf(fp,"#          This will corrupt the file and make it incompatible with the\n");
   () = fprintf(fp,"#          apec software.\n#\n");

   % Begin marker
   () = fprintf(fp,"BEGIN\n");
   n = length(mdl.model);
   % loop over the models
   foreach (models) using ("keys","values") {
      (modelname,dataset) = ();
 
      if (orelse {thismodel == NULL} {thismodel == modelname}) {
	 extra = strlen(modelname) + 8;
	 extra = "%" + string(extra) + "s";
	 () = fprintf(fp,"%s\n",model_separator);
	 () = fprintf(fp,extra+" : %s\n","Modelname",modelname);
	 () = fprintf(fp,extra+" : ","Datasets");
	 foreach (dataset) {
	    tmp = ();
	    if (tmp.content == 0) {
	       () = fprintf(fp,"%d ",tmp.content);
	       () = fprintf(fp,"%s: Datasets = 0 means all datasets",
			    Str->empty_string(20));
	    } else {
	       () = fprintf(fp,"%d ",tmp.content);
	    }
	 }
	 () = fprintf(fp,"\n");
	 
	 % Find correct starting position in mdl structure for this model
	 for (start = 0; start < n; start++) {
	    tmp = Str->remove_excess_whitespace("apec["+modelname+"]");
	    if (mdl.model[start] == tmp) break;
	 }
	 start++;
	 % Find correct end position in mdl structure for this model
	 for (end = start; end < n; end++) {
	    tmp = Str->remove_excess_whitespace("apec["+modelname+"]");
	    if (mdl.model[end] != tmp) break;
	 }
	 % output individual parameter names and their values
	 for (j = start; j < end; j++) {
	    () = fprintf(fp,extra+" : %10.5g\n",mdl.parname[j],mdl.parvalue[j]);
	 }
      }
   }

   % End marker
   () = fprintf(fp,"END\n");

   % return to normal evaluation of apec function
   AM->getmdlmodels = NULL;
   
   % set verbosity to its original value.
   AM->verbose = verbose;

   if (_save_message != NULL) {
      if (file == standard_model_file) {
	 vmessage("\n\nNo output file has been specified. Using default file name \"%s\".\n", file);
      } else {
	 vmessage("\n\nSaving apec model to file \"%s\".\n",file);
      }
   }

}

define AM_load_model() {
   
   %
   % Loads a model from file and gives the necessary sherpa
   % commands to define the models and assign them to the correct 
   % dataset.
   %
   variable filename, usage = "Usage: AM_load_model([filename]);";

   switch (_NARGS)
   { case 0 : filename = standard_model_file; }
   { case 1 : filename = (); }
   { message(usage); _pop_n(_NARGS); return; }

   variable fp, line, words, modelname, datasets, tmp, sherpa_cmnd;
   variable i, pdot;
   variable srcs = Assoc_Type[String_Type,""];

   fp = fopen(filename,"r");
   if (fp == NULL) verror("File \"%s\" could not be opened",filename);

   vmessage("\nLoading model from file \"%s\".\n",filename);

   while (-1 != fgets(&line,fp)) {

      line = strtrim_end(line);       % remove ' \t\r\n'
      if (line[0] == '#') continue;
      if (line == "BEGIN") continue;
      if (line == "END") continue;
      if (line == model_separator) {


	 % determine modelname
	 () = fgets(&line,fp);
	 words = strchop(line,':',0);
	 modelname = Str->remove_excess_whitespace(words[1]);
	 
	 % determine datasets
	 () = fgets(&line,fp); line = strtrim_end(line); % remove ' \t\r\n'
	 words = strchop(line,':',0);
	 datasets = strchop(Str->remove_excess_whitespace(words[1]),' ',0);
	 foreach (datasets) {
	    tmp = ();
	    if (tmp == "0") tmp = "allsets";
	    if (assoc_key_exists(srcs,tmp) == 0) {
	       srcs[tmp] = modelname;
	    } else {
	       srcs[tmp] = srcs[tmp] + " + " + modelname;
	    }
	 }

	 % Give apec command to sherpa
	 sherpa_cmnd = sprintf("apec[%s]",modelname);
	 AM_sherpa_eval(sherpa_cmnd);
      } else {
	 % Read individual parameters

	 words = Str->split_and_trim(line,':');
	 pdot = string_match(words[0],".T",1);
	 if (pdot > 0) {
	    pdot = integer(substr(words[0],pdot+2,-1));
	    if (pdot > AM->N_Temp_comp_max) {
		  vmessage("The saved model seems to have more temperature components (%d)\nthan the current model (%d). Please make the appropriate changes in \"%s\".\nFor now temperature components larger than %d are ignored.\n",pdot,AM->N_Temp_comp_max,AM->model_file,AM->N_Temp_comp_max);
	    } 

	    % the extra commands are used to prevent a stupid
	    % warning message.

	    sherpa_cmnd = sprintf("%s.MIN = 0.0",words[0]);
	    AM_sherpa_eval(sherpa_cmnd);
	    %
	    sherpa_cmnd = sprintf("%s = %s", words[0], words[1]);
	    AM_sherpa_eval(sherpa_cmnd);
	    %
	    sherpa_cmnd = sprintf("%s.MIN = %g",words[0],AM->Tmin);
	    AM_sherpa_eval(sherpa_cmnd);
	 } else {
	    sherpa_cmnd = sprintf("%s = %s", words[0], words[1]);
	    AM_sherpa_eval(sherpa_cmnd);
	 }
      }
   }

   message("\nIssuing sherpa commands...\n");
   variable k, v, vs;
   foreach (srcs) using ("keys", "values") {
      (k,v) = ();
      sherpa_cmnd = sprintf("source %s = %s", k, v);
      %message(sherpa_cmnd);
      AM_sherpa_eval(sherpa_cmnd);
      vs = strchop(v,' ',0); v = "";
      for (i = 0; i < length(vs); i++) {
	 if (vs[i] == "+") {
	    v = v + vs[i] + " ";
	 } else {
	    v = v + "apec[" + vs[i] + "] ";
	 }
      }
      vmessage("sherpa> source %s = %s",k,v);
   }
   message("\n");

}

define AM_save_continuum() {

   variable file, model = NULL, tmp;
   variable usage = "Usage: AM_save_continuum([file [, model]]);";

   switch (_NARGS)
   { case 0 : tmp = standard_model_file; 
              standard_model_file = standard_continuum_file;
              file = standard_model_file;}
   { case 1 : file = (); }
   { case 2 : (file,model) = (); }
   { message(usage); _pop_n(_NARGS); return; }

   _save_continuum = 1;
   AM_save_model(file, model);
   _save_continuum = NULL;

   if (standard_model_file == standard_continuum_file) {
      standard_model_file = tmp;
   }
   
}

define AM_load_continuum() {
    
   variable filename, usage = "Usage: AM_load_continuum([filename]);";

   switch (_NARGS)
   { case 0 : filename = standard_continuum_file; }
   { case 1 : filename = (); }
   { message(usage); _pop_n(_NARGS); return; }
    
   variable fp, line, words, modelname, datasets, tmp, sherpa_cmnd;
   variable i, components, comp, comps = NULL, elem = NULL, abund = NULL; 
   %comps = Struct_Type[AM->N_Temp_comp_max];
   variable ntemp = 0, pdot, k, newmodel = 0;
   variable srcs = Assoc_Type[LList_Type];

   fp = fopen(filename,"r");
   if (fp == NULL) verror("File \"%s\" could not be opened",filename);

   if (_save_message != NULL) {
      vmessage("\nLoading continuum from file \"%s\".\n",filename);
   }

   while (-1 != fgets(&line,fp)) {

      line = strtrim_end(line);       % remove ' \t\r\n'
      if (line[0] == '#') continue;
      if (line == "BEGIN") continue;
      if (line == "END") continue;
      if (line == model_separator or newmodel == 1) {
	 
	 % determine modelname
	 if (newmodel == 0) () = fgets(&line,fp);
	 newmodel = 0;
	 words = strchop(line,':',0);
	 modelname = Str->remove_excess_whitespace(words[1]);
	 
	 % determine datasets
	 () = fgets(&line,fp); line = strtrim_end(line); % remove ' \t\r\n'
	 words = strchop(line,':',0);
	 datasets = strchop(Str->remove_excess_whitespace(words[1]),' ',0);
	 
	 % Read individual parameter values and include them in an
	 % isis model structure. Initialize the linked lists used to 
	 % collect the model information.
	 comp = isis->default_plasma_state();
	 comp.metal_abund = 0.0;
	 comps = NULL;
	 elem = NULL;
	 abund = NULL;

	 while ((-1 != fgets(&line,fp)) and 
		((line != "END\n") and (line != model_separator+"\n"))) {
	    line = strtrim_end(line);       % remove ' \t\r\n'

	    words = Str->split_and_trim(line,':');
	    % manipulate words[0]
	    pdot = Str->position('.',words[0]);
	    words[0] = substr(words[0],pdot+2,-1);
	    %
	    % handle temps:
	    if (words[0][0] == 'T') {
	       comps = LL->append(@comp,comps);
	       pdot = integer(substr(words[0],2,-1));
	       LL->nth(pdot-1,comps).temperature = atof(words[1])/AM->K_to_keV;
	    }
	    % handle norms:
	    if (words[0][0] == 'n') {
	       pdot = integer(substr(words[0],5,-1));
	       LL->nth(pdot-1,comps).norm = atof(words[1]);
	    }
	    % handle abundances
	    if (words[0][0] == 'a') {
	       pdot = Str->position('.',words[0]);
	       elem = LL->append(AM->Elem_to_Z[strlow(substr(words[0],6,-1))],elem);
	       abund = LL->append(atof(words[1]),abund);
	    }
	 }
	 elem = LL->ll_to_array(elem);
	 abund = LL->ll_to_array(abund);

	 foreach (comps) { 
	    comp = (); 
	    comp.content.elem = elem;
	    comp.content.elem_abund = abund;
	 }
	 comps = LL->ll_to_array(comps);

	 % add model to linked list of components
	 foreach (datasets) {
	    tmp = ();
	    foreach (comps) {
	       comp = ();
	       if (comp.temperature != 0.0) {
	 	  if (assoc_key_exists(srcs,tmp) == 0) {
	 	     components = @LList_Type;
	 	     components.content = comp;
	 	     srcs[tmp] = components;
	 	  } else {
	 	     srcs[tmp] = LL->append(comp,srcs[tmp]);
	 	  }
	       }
	    }
	 }
	 if (line == model_separator+"\n") newmodel = 1;
      }
   }

   % convert the linked list into an array
   AM->continuum = Assoc_Type[Array_Type];
   foreach (srcs) using ("keys") {
      k = ();
      AM->continuum[k] = LL->ll_to_array(srcs[k]);
   }

}

define AM_set_continuum() {

   %
   % Set the continuum variable to the current model or to the model
   % belonging an specific apec_model
   %

   variable usage = "Usage: AM_set_continuum([model])";
   variable file, model = NULL;
   variable dflt = ".apec.tmp", answer = NULL;

   _save_message = NULL;

   switch (_NARGS) 
   { case 0 : file = dflt; }
   { case 1 : file = dflt;
              model = (); }
   { message(usage); _pop_n(_NARGS); return; }

   if (file == dflt) {
      message("\n\nYou are about to set a new continuum");
      while (orelse 
	     {answer == NULL}
	     {Utils->member(strlow(answer),["y","yes","n","no"]) == NULL}) {
	  
	  IO->get_user_reply(&answer,
			     "Do you want to save it to file [y/n] : ");
      }
      if (Utils->member(strlow(answer),["y","yes"]) != NULL) {
	 _save_message = 1;
	 while (orelse {file == NULL} {file == dflt}) {
	    IO->get_user_reply(&file,"==> Enter file name : ");
	 }
      }

   }

   AM_save_continuum(file,model);
   AM_load_continuum(file);
   if (file == dflt) {
      % delete temporary file
      () = remove(file);
   }   

   _save_message = 1;

}

define AM_reset_continuum() {

   return AM->continuum = NULL;

}

define AM_get_continuum() {

   return AM->continuum;

}

