%
% Collection of routines to set elements of the Apec Models definitions,
% e.g., the number of temperature components.
%
% R. Hoogerwerf, SAO, 04/09/2002
%


define AM_set_NTemp () {

   % Function takes 1 argument and sets the number of temperature
   % components in the model. 
   % Valid values are 1 <= n <= AM->N_Temp_comp_max.

   variable i, n, mdls = NULL;

   switch (_NARGS)
   { case 0 : 
	 message("Usage: AM_set_NTemp(n, [model1, model2, ...])");
         message("  Expect at least one argument"); }
   { case 1 : n = (); mdls = AM_get_apec_modelnames_only(); }
   { n = _NARGS -1; 
     mdls = String_Type[n];
     for (i = 0; i < _NARGS - 1; i++) mdls[i] = (); }
   
   variable old = AM->N_Temp_comp;

   if (n >= 1 and n <= AM->N_Temp_comp_max) {
      AM->N_Temp_comp = n;
   } else {
      message("The number of Temperature component must be larger than 1");
      verror("and smaller than %d. You have enter %d", AM->N_Temp_comp_max,
	     n);
   }  

   variable k,v,tmp,sherpa_cmnd;
   foreach (mdls) {
	 
      k = ();
	 
      if (old < AM->N_Temp_comp) {
	 for (i = old+1; i <= AM->N_Temp_comp; i++) {
	    tmp = sprintf("T%d",i);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".VALUE = 1.0";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".MIN = 0.1";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "thaw " + k + "." + tmp;
	    AM_sherpa_eval(sherpa_cmnd);
	    tmp = sprintf("norm%d",i);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".VALUE = 1.0";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "thaw " + k + "." + tmp;
	    AM_sherpa_eval(sherpa_cmnd);
	 }
      } else if (old > AM->N_Temp_comp) {
	 for (i = old; i >= AM->N_Temp_comp + 1; i--) {
	    tmp = sprintf("T%d",i);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".MIN = 0.0";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".VALUE = 0.0";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "freeze " + k + "." + tmp;
	    AM_sherpa_eval(sherpa_cmnd);
	    tmp = sprintf("norm%d",i);
	    sherpa_cmnd = "CREATE " + k + "." + tmp + ".VALUE = 0.0";
	    AM_sherpa_eval(sherpa_cmnd);
	    sherpa_cmnd = "freeze " + k + "." + tmp;
	    AM_sherpa_eval(sherpa_cmnd);
	 }
      }
   }
}

define AM_get_NTemp () {

   % Function takes 0 argument and returns the number of temperature
   % components in the model. 

   return AM->N_Temp_comp;
}

define AM_show_NTemp () {

   % Function takes 0 argument and shows the number of temperature
   % components in the model. 

   vmessage("# of temperature components : %d", AM->N_Temp_comp);

}

define AM_fit_continuum_only () {
   
   % Modify 'Fit' so that the model only includes the continuum

   AM->Fit = AM->Fit_Continuum_only;

}

define AM_fit_lines_only () {
   
   % Modify 'Fit' so that the model only includes the continuum

   AM->Fit = AM->Fit_Lines_only;

}

define AM_fit_continuum_and_lines () {
   
   % Modify 'Fit' so that the model only includes the continuum

   AM->Fit = AM->Fit_Continuum_and_lines;

}

define AM_continuum_only () {
   
   % Modify 'Fit' so that the model only includes the continuum

   AM->Fit = AM->Continuum_only;

}

define AM_set_fit (fit) {

   % Set 'AM->Fit' variable to Continuum_only, Lines_only,  
   %                           Continuum_and_lines, or No_fit.
   % If 'Fit' has a different value throw and error.

   if (orelse {fit == AM->Fit_Continuum_only}
              {fit == AM->Fit_Lines_only}
              {fit == AM->Fit_Continuum_and_lines}
              {fit == AM->Continuum_only}) {
      AM->Fit = fit;
   } else {
      message("AM_set_Fit takes one of the following arguments : ");
      vmessage("  AM->Fit_Continuum_only      [%d]", AM->Fit_Continuum_only);
      vmessage("  AM->Fit_Lines_only          [%d]", AM->Fit_Lines_only);
      vmessage("  AM->Fit_Continuum_and_lines [%d]", AM->Fit_Continuum_and_lines);
      vmessage("  AM->Continuum_only          [%d]", AM->Continuum_only);
      verror("You have entered %d", fit);
   }
}

define AM_get_fit () {

   % Returns value of 'AM->Fit' variable which determines what part of the
   % spectrum is modeled (Cont., or Lines, or both).

   return AM->Fit;

}


define AM_show_fit () {

   switch(AM->Fit) 
   { case AM->Fit_Continuum_only :      message("Fitting continuum only"); }
   { case AM->Fit_Lines_only :          message("Fitting lines only"); }
   { case AM->Fit_Continuum_and_lines : message("Fitting continuum and lines"); }
   { case AM->Continuum_only : message("No fitting; continuum only"); }

}


define AM_set_verbose(verb) {
 
   AM->verbose = verb;

}

define AM_get_data(dataset) {
   
   variable data = get_data(dataset);
 
% Based on old (more sofisticated version) of get_data
%
%   if (data == NULL) {
%      return NULL;
%   } else if (length(data.x0) == 0) {
%      return NULL;
%   } else {
%      return data.x0;
%   }

   return data;

}

define AM_get_data_binsize(dataset) {
   
   variable data = get_data(dataset);
 
   if (data == NULL) {
      return NULL;
   } else if (length(data.x0) == 0) {
      return NULL;
   } else {
      return Math->avg(data.dx0);
   }

}
