/** \file SpecFile.cxx
    \brief Implementation of SpecFile class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/

#include "profit/SpecFile.h"
#include "profit/Spectrum.h"
#include "profit/Table.h"

#include <iomanip>
#include <memory>
#include <sstream>
#include <stdexcept>

namespace profit {

  SpecFile::SpecFile(const std::string & spec_file_name): m_spec_cont(), m_spec_file_name(spec_file_name), m_spec_table(0) {
    m_spec_table = new InputTable(spec_file_name, "SPECTRUM");
    if (0 == m_spec_table) throw std::runtime_error("SpecFile: Unable to open table " + spec_file_name);
    if (1 != m_spec_table->getFieldDim("counts")) {
      m_spec_cont.resize(m_spec_table->getNumRecords(), 0);
    } else {
      m_spec_cont.resize(1, 0);
    }
  }

  SpecFile::~SpecFile() {
    for (SpecCont::reverse_iterator itor = m_spec_cont.rbegin(); itor != m_spec_cont.rend(); ++itor) {
      delete *itor;
    }
    delete m_spec_table;
  }

  SpecFile::size_type SpecFile::getNumSpectra() const { return m_spec_cont.size(); }

  Spectrum * SpecFile::getSpectrum(int spec_index) {
    if (0 > spec_index) throw std::logic_error("SpecFile::getSpectrum(): Invalid (negative) spec_index");
    else if (m_spec_cont.size() <= (unsigned long) spec_index)
      throw std::logic_error("SpecFile::getSpectrum(): Invalid (too large) spec_index");
    Spectrum * spec = m_spec_cont[spec_index];
    if (0 == spec) {
      if (m_spec_cont.size() == 1) {
        spec = new Spectrum(*this);
      } else {
        spec = new Spectrum(*this, spec_index);
      }
      if (0 == spec)
        throw std::runtime_error("SpecFile::getSpectrum(): Unable to create spectrum for " + m_spec_table->getFileName());
      m_spec_cont[spec_index] = spec;
    }
    return spec;
  }

  Spectrum * SpecFile::getSpectrum(const std::string & ebounds_file_name, int spec_index) {
    if (0 > spec_index) throw std::logic_error("SpecFile::getSpectrum(int): Invalid (negative) spec_index");
    else if (m_spec_cont.size() <= (unsigned long) spec_index)
      throw std::logic_error("SpecFile::getSpectrum(int): Invalid (too large) spec_index");
    std::auto_ptr<InputTable> ebounds_table(new InputTable(ebounds_file_name, "EBOUNDS"));
    Spectrum * spec = m_spec_cont[spec_index];
    if (0 == spec) {
      if (m_spec_cont.size() == 1) {
        spec = new Spectrum(*this, *ebounds_table);
      } else {
        spec = new Spectrum(*this, *ebounds_table, spec_index);
      }
      if (0 == spec)
        throw std::runtime_error("SpecFile::getSpectrum(int): Unable to create spectrum for " + m_spec_table->getFileName());
      m_spec_cont[spec_index] = spec;
    }
    return spec;
  }

  std::string SpecFile::getFileName() const { return m_spec_table->getFileName(); }

  void SpecFile::getSummary(std::vector<std::string> & summary) const {
    // Read various keywords and put them into the spectrum summary.
    std::string object;
    try { object = m_spec_table->getKeyword("OBJECT"); } catch (...) { /* Ignore */ }
    if (object.empty()) object = "Unknown";

    std::string obs_date;
    try { obs_date = m_spec_table->getKeyword("DATE-OBS"); } catch (...) { /* Ignore */ }
    if (obs_date.empty()) obs_date = "Unknown";

    std::string exposure;
    try { exposure = formatNumericString(m_spec_table->getKeyword("EXPOSURE"), 7); } catch (...) { /* Ignore */ }
    if (exposure.empty()) exposure = "INDEF";

    std::string ra;
    try { ra = formatNumericString(m_spec_table->getKeyword("RA_NOM"), 7); } catch (...) { /* Ignore */ }
    if (ra.empty())
      try { ra = formatNumericString(m_spec_table->getKeyword("RA_OBJ"), 7); } catch (...) { /* Ignore */ }
    if (ra.empty())
      try { ra = formatNumericString(m_spec_table->getKeyword("RA_PNT"), 7); } catch (...) { /* Ignore */ }
    if (ra.empty()) ra = "?";

    std::string dec;
    try { dec = formatNumericString(m_spec_table->getKeyword("DEC_NOM"), 7); } catch (...) { /* Ignore */ }
    if (dec.empty())
      try { dec = formatNumericString(m_spec_table->getKeyword("DEC_OBJ"), 7); } catch (...) { /* Ignore */ }
    if (dec.empty())
      try { dec = formatNumericString(m_spec_table->getKeyword("DEC_PNT"), 7); } catch (...) { /* Ignore */ }
    if (dec.empty()) dec = "?";

    std::string telescope;
    try { telescope = m_spec_table->getKeyword("TELESCOP"); } catch (...) { /* Ignore */ }
    // Note: don't test for empty condition.

    std::string instrument;
    try { instrument = m_spec_table->getKeyword("INSTRUME"); } catch (...) { /* Ignore */ }
    // Note: don't test for empty condition.

    std::string grating;
    try { grating = m_spec_table->getKeyword("GRATING"); } catch (...) { /* Ignore */ }
    for (std::string::iterator itor = grating.begin(); itor != grating.end(); ++itor) *itor = toupper(*itor);
    // Note: don't test for empty condition.

    // Report telescope and instrument together as "detector".
    std::string detector;
    if (!telescope.empty() && !instrument.empty()) detector = telescope + ": " + instrument;
    else if (!telescope.empty()) detector = telescope;
    else if (!instrument.empty()) detector = instrument;
    else detector = "Unknown";

    // Display the file part only of the file name.
    std::string spec_file_name = m_spec_file_name;
    std::string::size_type slash_pos = spec_file_name.find_last_of("/\\");
    if (std::string::npos != slash_pos) {
      spec_file_name = spec_file_name.substr(slash_pos + 1);
    }

    // Truncate file name to the available size in display only.
    int first_char = spec_file_name.size() > 18 ? spec_file_name.size() - 18 : 0;

    // Format the output string to go in the spectrum summary.
    std::ostringstream os;
    os << std::fixed;
    os << std::left;
    os << std::setw(30) << "File name: " + spec_file_name.substr(first_char) << "Source name: " << object;

    summary.clear();

    summary.push_back(os.str());
    os.str("");
    os << std::setprecision(7) << std::setw(30) << "Exposure: " + exposure << "Obs Date: " + obs_date;
    summary.push_back(os.str());
    os.str("");
    os << std::setprecision(7) << std::setw(30) << "Position: " + ra + ", " + dec << "Detector: " + detector;
    summary.push_back(os.str());
    os.str("");
    if (!grating.empty() && "NONE" != grating) os << "Grating: " + grating;
    summary.push_back(os.str());
  }

  const InputTable * SpecFile::getInputTable() const { return m_spec_table; }

  std::string SpecFile::formatNumericString(const std::string & number, int prec) const {
    std::stringstream ss1;
    ss1 << number;
    double number_d;
    ss1 >> number_d;
    std::stringstream ss2;
    ss2.precision(prec);
    ss2 << number_d;
    return ss2.str();
  }

}
