/** \file SpectrumView.cxx
    \brief Implementation of SpectrumView class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <algorithm>
#include <cmath>
#include <cstddef>
#include <iomanip>
#include <memory>
#include <sstream>
#include <stdexcept>

#include "profit/Gui.h"
#include "profit/Spectrum.h"
#include "profit/SpectrumView.h"
#include "profit/Stream.h"

#include "TCanvas.h"
#include "TError.h"
#include "TF1.h"
#include "TGaxis.h"
#include "TGButton.h"
#include "TGFrame.h"
#include "TH1.h"
#include "TLatex.h"
#include "TPad.h"
#include "TRootEmbeddedCanvas.h"
#include "TSpectrum.h"
#include "TVirtualFitter.h"

#include <iostream>

ClassImp(profit::SpectrumView)

namespace {

  // Maximum number of peaks, heuristically chosen to make fits converge in decent time interval.
  const Int_t s_max_num_peaks = 20;

  // Each peak is represented by a Gaussian, having 3 free parameters.
  const Int_t s_par_per_peak = 3;

  // Symbolic identities of the parameters.
  enum CoeffId { kNumPeaks, kBack0, kBack1, kNorm, kMean, kSigma };

  // Function to compute value of fit is a linear term plus a sum of Gaussians.
  Double_t FitValue(Double_t * x, Double_t * par) {
    Int_t num_peaks = Int_t(par[kNumPeaks]);
    Double_t result = par[kBack0] + par[kBack1] * x[0];
    for (Int_t ii = 0; ii < num_peaks; ++ii) {
      Double_t norm  = par[s_par_per_peak * ii + kNorm];
      Double_t mean  = par[s_par_per_peak * ii + kMean];
      Double_t sigma = par[s_par_per_peak * ii + kSigma];
      result += norm * TMath::Gaus(x[0], mean, sigma);
    }
    return result;
  }
}

namespace profit {

  // Speed of light in km/s.
  const double SpectrumView::s_c = 299792.458; // km/s
  const double SpectrumView::s_fwhm_factor = 2. * sqrt(2. * log(2.)); // Convert from sigma to FWHM.

  SpectrumView::Label::Label(const IEnergy & ienergy, const std::string & text): m_energy(ienergy), m_text(text), m_latex(0),
    m_line(0) {}

  SpectrumView::Label::Label(const Label & label): m_energy(label.m_energy), m_text(label.m_text), m_latex(0), m_line(0) {}

  SpectrumView::Label::~Label() { delete m_line; delete m_latex; }

  void SpectrumView::Label::display(double x, double y) {
    delete m_line;
    delete m_latex;

    m_latex = new TLatex(x, y, m_text.c_str());
    if (0 == m_latex) throw std::runtime_error("Label::display: dynamic allocation failed for TLatex label");
    m_line = new TLine(x, y, x, y * (1 - .1));
    if (0 == m_line) throw std::runtime_error("Label::display: dynamic allocation failed for marker line");

    m_latex->SetTextAngle(90);
    m_latex->SetTextColor(kBlue);
    m_latex->Draw();

    m_line->SetLineColor(kBlue);
    m_line->Draw();
  }

  bool SpectrumView::isDisplayed() const { return m_displayed; }

  void SpectrumView::updateDisplay() {
    if (0 != m_canvas) {
      if (0 != m_hist || 0 != m_fit_func) {

        // Save current pad.
        TVirtualPad * save = gPad;

        // Get TCanvas from TRootEmbeddedCanvas.
        TCanvas * canvas = m_canvas->GetCanvas();

        // Select TCanvas.
        gPad = canvas;

        // Handle highlighting.
        if (m_highlight) canvas->SetHighLightColor(kRed);
        else canvas->SetHighLightColor(kWhite);

        Bool_t editable = gPad->IsEditable();

        gPad->SetEditable(kTRUE);

        gPad->cd(m_pad_id + 1);

        // Draw any drawable objects.
        if (0 != m_hist) {
          m_hist->Draw();
        }
        if (0 != m_top_axis) m_top_axis->Draw();
        if (0 != m_fit_func) m_fit_func->Draw("same");
        displayLabels();

        // Request update.
        gPad->Update();

        gPad->SetEditable(editable);

        // Restore original pad.
        gPad = save;
      }
    }
  }

  const IEnergy & SpectrumView::Label::getEnergy() const { return m_energy; }

  SpectrumView::~SpectrumView() { clearDisplay(); delete m_fit_func; delete m_top_axis; delete m_hist; }

  // TODO: test this: is there OBIWAN error in the number of bins?
  SpectrumView::size_type SpectrumView::numSelectedBins() const {
    size_type num_bins = 0;
    if (0 != m_hist) {
      TAxis * x_axis = m_hist->GetXaxis();
      if (0 != x_axis)
        num_bins = x_axis->GetLast() - x_axis->GetFirst();
    }
    return num_bins;
  }

  SpectrumView * SpectrumView::slice(const IEnergy & min_energy, const IEnergy & max_energy) const {
    // Create new blank view to return.
    // TODO: cloning here creates an m_hist which is deleted below by a call to createHistogram. Is there
    // a way around this inefficiency?
    SpectrumView * new_view = clone();
    if (0 == new_view) throw std::runtime_error("SpectrumView::slice: failed to clone");

    // Select the domain of the new view to match the criteria.
    new_view->select(min_energy, max_energy);

    // Extract the domain and range of the view.
    std::vector<double> domain;
    std::vector<double> range;
    new_view->getDomain(domain);
    new_view->getRange(range);

    // The range used to create a Root histogram needs a leading 0 for its underflow bin.
    std::vector<double> root_range(range.size() + 1, 0.);
    std::copy(range.begin(), range.end(), root_range.begin() + 1);

    // Create a new histogram for the view.
    new_view->createHistogram(domain, root_range);

    return new_view;
  }

  SpectrumView * SpectrumView::slice(double min_domain, double max_domain) const {
    // Create new blank view to return.
    // TODO: cloning here creates an m_hist which is deleted below by a call to createHistogram. Is there
    // a way around this inefficiency?
    SpectrumView * new_view = clone();
    if (0 == new_view) throw std::runtime_error("SpectrumView::slice: failed to clone");

    // Extract the domain and range of the view.
    std::vector<double> domain;
    std::vector<double> range;
    new_view->getDomain(domain);
    new_view->getRange(range);

    std::vector<double>::iterator begin;
    for (begin = domain.begin(); begin != domain.end() && *begin < min_domain; ++begin) {}
    std::vector<double>::iterator end;
    for (end = begin; end != domain.end() && *end < max_domain; ++end) {}

    domain = std::vector<double>(begin, end);
    range = std::vector<double>(range.begin() + (begin - domain.begin()), range.begin() + (end - domain.begin()));

    // The range used to create a Root histogram needs a leading 0 for its underflow bin.
    std::vector<double> root_range(range.size() + 1, 0.);
    std::copy(range.begin(), range.end(), root_range.begin() + 1);

    // Create a new histogram for the view.
    new_view->createHistogram(domain, root_range);

    return new_view;
  }

  void SpectrumView::getDomain(std::vector<double> & domain) const {
    // Start by emptying container supplied by caller.
    domain.clear();
    if (0 != m_hist) {
      // Get axis.
      TAxis * x_axis = m_hist->GetXaxis();
      // Get current range from axis.
      Int_t first = x_axis->GetFirst();
      Int_t last = x_axis->GetLast() + 1;
      // Resize the container.
      Int_t num_bins = last - first + 1;
      domain.resize(num_bins);
      // Fill the container.
      for (Int_t ii = first; ii != last; ++ii) {
        domain[ii - first] = x_axis->GetBinLowEdge(ii);
      }
      domain[num_bins - 1] = x_axis->GetBinUpEdge(last - 1);
    }
  }

  void SpectrumView::getRange(std::vector<double> & range) const {
    range.clear();
    if (0 != m_hist) {
      TAxis * x_axis = m_hist->GetXaxis();
      Int_t first = x_axis->GetFirst();
      Int_t last = x_axis->GetLast() + 1;
      Int_t num_bins = last - first;
      range.resize(num_bins);
      for (Int_t ii = first; ii != last; ++ii) {
        range[ii - first] = m_hist->GetBinContent(ii);
      }
    }
  }

  const Spectrum & SpectrumView::getSpectrum() const { return *m_spectrum; }

  void SpectrumView::display(TGFrame * parent, int hints, long width, long height, unsigned int pad_id) {
    clearDisplay();

    // See if caller directly passed an embeddeding canvas.
    m_canvas = dynamic_cast<TRootEmbeddedCanvas *>(parent);

    if (0 == m_canvas) {
      // Other possibility is that caller passed a normal composite frame.
      // If so, create new embedded canvas in the composite frame.
      TGCompositeFrame * cf = dynamic_cast<TGCompositeFrame *>(parent);
      if (0 != cf) {
        m_layout_hints = new TGLayoutHints(hints);

        m_main_frame = new TGVerticalFrame(cf, width, height);

        if (0 != m_main_frame) {
          m_button_frame = new TGHorizontalFrame(m_main_frame, cf->GetWidth(), 22);
          m_close_button = new TGCheckButton(m_button_frame, "Close");
          m_select_button = new TGCheckButton(m_button_frame, "Select");
          m_canvas = new StoredCanvas(m_canvas_name.c_str(), m_main_frame, width, height - 22);
        }

        // Check whether allocations were successful.
        if (0 == m_layout_hints || 0 == m_main_frame || 0 == m_button_frame || 0 == m_close_button ||
          0 == m_select_button || 0 == m_canvas) {
          delete m_canvas; m_canvas = 0;
          delete m_select_button; m_select_button = 0;
          delete m_close_button; m_close_button = 0;
          delete m_button_frame; m_button_frame = 0;
          delete m_main_frame; m_main_frame = 0;
          delete m_layout_hints; m_layout_hints = 0;
          throw std::runtime_error("SpectrumView::display: dynamic allocation failure");
        }

        // Add new frames to parent.
        cf->AddFrame(m_main_frame, m_layout_hints);
        m_main_frame->AddFrame(m_button_frame, m_layout_hints);
        m_button_frame->AddFrame(m_close_button, m_layout_hints);
        m_button_frame->AddFrame(m_select_button, m_layout_hints);
        m_main_frame->AddFrame(m_canvas, m_layout_hints);

        // Bind actions to button.
        m_close_button->Connect("Clicked()", "profit::SpectrumView", this, "handleClose()");
        m_select_button->Connect("Clicked()", "profit::SpectrumView", this, "handleSelect()");
        m_close_button->SetToolTipText("Delete stored spectrum.");
        m_select_button->SetToolTipText("Check (select) stored spectrum.");

        m_main_frame->MapSubwindows();
        m_main_frame->Layout();
        m_main_frame->MapWindow();

        m_own_canvas = true;
      } else {
        throw std::logic_error("SpectrumView::display method was not passed a TGCompositeFrame *");
      }
    }

    // If canvas is "stored", it needs to refer to this view.
    StoredCanvas * stored_canvas = dynamic_cast<StoredCanvas *>(m_canvas);
    if (0 != stored_canvas) {
      // Let canvas know that this view is displayed in it.
      stored_canvas->setView(this);
    }

// TODO: INIT these members.
    m_hints = hints;
    m_height = height;
    m_width = width;
    m_pad_id = pad_id;

    m_displayed = true;

    // Display the histogram in the canvas.
    updateDisplay();
  }

  bool SpectrumView::fit() {
    bool success = false;
    std::vector<double>::size_type num_par = 0;
    if (0 != m_hist) {
      int max_num_fit = 3;

      std::vector<double> candidate;
      findPeakCandidates(candidate);
      Int_t num_found = candidate.size();

      // Do not proceed with fits if too many peaks were found.
      if (num_found > s_max_num_peaks) {
        std::ostringstream os;
        os << "More than " << s_max_num_peaks << " peak candidates found; please choose a simpler region";;
        throw std::runtime_error(os.str());
      }

      // Compute total number of parameters needed to express this model.
      num_par = s_par_per_peak * (num_found + 1);

      // Estimate linear background.
      TF1 background("fline", "pol1", 0, m_hist->GetNbinsX());

      // Attempt to fit until successful or number of fits exceeded.
      success = false;
      for (int ii = 0; !success && ii < max_num_fit; ++ii) {
        success = 0 == m_hist->Fit("fline", "qn");
        m_hist->Fit("fline", "qn");
      }

      if (!success) {
        m_fit_par.clear();
        m_fit_par_error.clear();
        delete m_fit_func;
        m_fit_func = 0;
        throw std::runtime_error("Linear background fit did not converge");
      }

      m_fit_par.resize(num_par, 0.);
      m_fit_par_error.resize(num_par, 0.);

      // Compute constraints for fits:
      TAxis * x_axis = m_hist->GetXaxis();
      int num_bins = x_axis->GetLast() - x_axis->GetFirst() + 1;

      double min_domain = x_axis->GetBinLowEdge(x_axis->GetFirst());
      double max_domain = x_axis->GetBinUpEdge(x_axis->GetLast());
      double min_sigma = (max_domain - min_domain) / num_bins / s_fwhm_factor;
      double max_sigma = (max_domain - min_domain) / 2.;

      // Loop over all found peaks. Eliminate peaks below the background level.
      m_fit_par[kBack0] = background.GetParameter(0);
      m_fit_par[kBack1] = background.GetParameter(1);
      Int_t num_peaks = 0;
      for (Int_t ii = 0; ii < num_found; ++ii) {
        Float_t xp = candidate[ii];
        Int_t bin = m_hist->GetXaxis()->FindBin(xp);
        Float_t yp = m_hist->GetBinContent(bin);
        if (yp - TMath::Sqrt(yp) < background.Eval(xp)) continue;
        // Keep this peak, and set its initial guesses.
        m_fit_par[s_par_per_peak * num_peaks + kNorm] = yp;
        m_fit_par[s_par_per_peak * num_peaks + kMean] = xp;
        m_fit_par[s_par_per_peak * num_peaks + kSigma] = 3. * min_sigma;
        ++num_peaks;
      }
      delete m_fit_func;

      // Prepare a function object to fit the data.
      m_fit_func = new TF1(m_func_name.c_str(), FitValue, 0, num_bins, num_par);

      TVirtualFitter::Fitter(m_hist, num_par); //There may be more than the default 25 parameters.
      for (int ii = 0; ii < num_peaks; ++ii) {
        m_fit_func->SetParLimits(s_par_per_peak * ii + kNorm, 1., 3 * m_fit_par[s_par_per_peak * ii + kNorm]);
        m_fit_func->SetParLimits(s_par_per_peak * ii + kMean, min_domain, max_domain);
        m_fit_func->SetParLimits(s_par_per_peak * ii + kSigma, min_sigma, max_sigma);
      }
      m_fit_func->SetParameters(&m_fit_par[0]);

      // Set and freeze the coefficient of the fit function giving the number of peaks.
      m_fit_func->FixParameter(kNumPeaks, num_peaks);

      m_fit_func->SetNpx(num_bins);

      // Attempt to fit until successful or number of fits exceeded.
      success = false;
      for (int ii = 0; !success && ii < max_num_fit; ++ii) {
        //success = 0 == m_hist->Fit(m_func_name.c_str(), "QN", 0, min_domain, max_domain);
        success = 0 == m_hist->Fit(m_func_name.c_str(), "QN");
      }

    }

    // If successful, display result of fit.
    if (success) {
      m_fit_par.resize(num_par);
      m_fit_par_error.resize(num_par);
      for (std::vector<double>::size_type ii = 0; ii < num_par; ++ii) {
        m_fit_par[ii] = m_fit_func->GetParameter(ii);
        m_fit_par_error[ii] = m_fit_func->GetParError(ii);
      }
      updateDisplay();
    } else {
      m_fit_par.clear();
      m_fit_par_error.clear();
      delete m_fit_func;
      m_fit_func = 0;
      throw std::runtime_error("Fit did not converge");
    }

    return success;
  }

  void SpectrumView::findPeakCandidates(std::vector<double> & peak_candidate) const {
    peak_candidate.clear();
    if (0 != m_hist) {
      // Save current error chattiness level:
      long root_err_level = gErrorIgnoreLevel;

      // Tell root to ignore problems in searches instead of issuing annoying warnings.
      gErrorIgnoreLevel = 100000;

      // Use TSpectrum to find the peak candidates.
      std::auto_ptr<TSpectrum> spec(new TSpectrum(2 * s_max_num_peaks));
      std::auto_ptr<SpectrumView> tmp_view(slice(getMinEnergy(), getMaxEnergy()));
      Int_t num_found = spec->Search(tmp_view->m_hist, 1, "goff", .01);
      unsigned long ul_num_found = num_found < 0 ? 0 : num_found;

      // Restore current error chattiness level.
      gErrorIgnoreLevel = root_err_level;

      Float_t * position = spec->GetPositionX();
      peak_candidate.resize(ul_num_found);
      for (std::vector<double>::size_type index = 0; index != ul_num_found; ++index) {
        peak_candidate[index] = position[index];
      }

      if (peak_candidate.empty()) throw std::runtime_error("No candidate peaks (or too many) found.");
    }
  }

  void SpectrumView::fitSummary(OStream & os) {
    if (m_fit_par.empty()) {
      os << "Fit did not converge. Change selected range and try again." << std::endl;
    } else {
      os << "Position            FWHM              Peak Counts" << std::endl;
      os << std::left;
      os.precision(6);
      for (size_t ii = 0; ii < m_fit_par[kNumPeaks]; ++ii) {
        std::stringstream ss;
        // Position, with error.
        ss << m_fit_par[s_par_per_peak * ii + kMean] <<
          "(" << std::setprecision(2) << m_fit_par_error[s_par_per_peak * ii + kMean] << ")";
        os << std::setw(20) << ss.str();
        ss.str("");
        // FWHM, with error.
        ss << s_fwhm_factor * m_fit_par[s_par_per_peak * ii + kSigma] <<
          "(" << s_fwhm_factor * m_fit_par_error[s_par_per_peak * ii + kSigma] << ")";
        os << std::setw(18) << ss.str();
        ss.str("");
        // Counts, with error.
        os << std::setprecision(6) << m_fit_par[s_par_per_peak * ii + kNorm] <<
          "(" << std::setprecision(2) << m_fit_par_error[s_par_per_peak * ii + kNorm] << ")";
        os << std::endl;
      }
    }
  }

  void SpectrumView::addLabel(const IEnergy & ienergy, const std::string & text) {
    m_label.push_back(Label(ienergy, text));
  }

  void SpectrumView::clearLabels() {
    m_label.clear();
  }

  bool SpectrumView::getHighlight() const { return m_highlight; }

  void SpectrumView::setHighlight(bool highlight) {
    if (m_highlight != highlight) {
      m_highlight = highlight;
      updateDisplay();
    }
    if (0 != m_select_button) m_select_button->SetState(m_highlight ? kButtonDown : kButtonUp, kFALSE);
  }

  void SpectrumView::scroll(double delta_x) {
    if (0 != m_hist && delta_x != 0.) {
      TAxis * x_axis = m_hist->GetXaxis();
      Int_t begin = x_axis->GetFirst();
      Int_t end = x_axis->GetLast();
      Int_t width = end - begin;

      // Find number of bins by which to scroll.
      Int_t delta = Int_t(delta_x * width);

      // Heuristic measure to increase sensitivity: force range to change for even the smallest mouse movements.
      if (delta == 0) {
        if (delta_x > 0) delta = 1;
        else delta = -1;
      }

      // Determine the new limits.
      begin += delta;
      end += delta;
      if (begin >= 1 && end <= x_axis->GetNbins()) {
        x_axis->SetRange(begin, end);
        updateDisplay();
      }
    }

  }

  void SpectrumView::applySrcRedshift(double src_redshift) {
    m_src_redshift = src_redshift;

    Int_t first = 0;
    Int_t last = 0;

    if (0 != m_hist) {
      // Get current visible range of view.
      TAxis * x_axis = m_hist->GetXaxis();
      first = x_axis->GetFirst();
      last = x_axis->GetLast();
    }

    // Recreate histogram using the current redshift; this will just change the scale of the domain.
    init();

    if (0 != first && 0 != m_hist) {
      // Get current visible range of view.
      TAxis * x_axis = m_hist->GetXaxis();
      x_axis->SetRange(first, last);
    }

    // Display the redshifted result.
    updateDisplay();
  }

  void SpectrumView::handleClose() {
    // Just make view disappear for now. Garbage collection will take place somewhere else.
    if (0 != m_main_frame) {
      m_main_frame->UnmapWindow();

      // It is necessary to layout the parent frame to reflect this change.
      TGCompositeFrame * parent = const_cast<TGCompositeFrame *>(
        dynamic_cast<const TGCompositeFrame *>(m_main_frame->GetParent())
      );
      if (0 != parent) parent->Layout();
    }
    m_displayed = false;
  }

  void SpectrumView::handleSelect() {
    if (0 != m_select_button) setHighlight(kTRUE == m_select_button->IsOn());
  }

  void SpectrumView::smooth() {
    if (0 != m_hist) {
      TAxis * x_axis = m_hist->GetXaxis();
      m_hist->Smooth(1, x_axis->GetFirst(), x_axis->GetLast());
      updateDisplay();
    }
  }

  SpectrumView::SpectrumView(const Spectrum & spectrum, double src_redshift): m_label(), m_fit_par(), m_fit_par_error(),
    m_canvas_name(), m_hist_name(), m_func_name(), m_src_redshift(src_redshift), m_spectrum(&spectrum), m_hist(0),
    m_top_axis(0), m_fit_func(0), m_layout_hints(0), m_main_frame(0), m_button_frame(0), m_close_button(0), m_select_button(0),
    m_canvas(0), m_own_canvas(false), m_highlight(false), m_displayed(false) {
    // Use this pointer to construct names (hopefully unique) for Root objects.
    std::ostringstream os;
    os << this;

    // Create the canvas name.
    m_canvas_name = "canvas " + os.str();

    // Create the histogram name.
    m_hist_name = "histogram " + os.str();

    // Create the function name.
    m_func_name = "function " + os.str();
  }

  SpectrumView::SpectrumView(const SpectrumView & view): m_label(view.m_label), m_fit_par(view.m_fit_par),
    m_fit_par_error(view.m_fit_par_error), m_canvas_name(), m_hist_name(), m_func_name(),
    m_src_redshift(view.getSrcRedshift()), m_spectrum(&view.getSpectrum()), m_hist(0), m_top_axis(0), m_fit_func(0),
    m_layout_hints(0), m_main_frame(0), m_button_frame(0), m_close_button(0), m_select_button(0), m_canvas(0),
    m_own_canvas(false), m_highlight(false), m_displayed(false) {
    // Use this pointer to construct names (hopefully unique) for Root objects.
    std::ostringstream os;
    os << this;

    // Create the canvas name.
    m_canvas_name = "canvas " + os.str();

    // Create the histogram name.
    m_hist_name = "histogram " + os.str();

    // Create the function name.
    m_func_name = "function " + os.str();

    if (0 != view.m_hist) {
      // Copy the source histogram.
      m_hist = new TH1D(*view.m_hist);

      // Prevent copy from having same name as original.
      m_hist->SetName(m_hist_name.c_str());

      if (0 == m_hist) throw std::runtime_error("SpectrumView copy constructor: failed to allocate histogram");
    }

#if 0
// TODO: Make extra axes work.
    if (0 != view.m_top_axis) {
      m_top_axis = new TGaxis(*view.m_top_axis);
      if (0 == m_top_axis) throw std::runtime_error("SpectrumView copy constructor: failed to allocate top axis");
    }
#endif

    if (0 != view.m_fit_func) {
      // Copy the source histogram.
      m_fit_func = new TF1(*view.m_fit_func);

      // Prevent copy from having same name as original.
      m_fit_func->SetName(m_func_name.c_str());

      if (0 == m_fit_func) throw std::runtime_error("SpectrumView copy constructor: failed to allocate fitting function");
    }
  }

  void SpectrumView::createHistogram(const std::vector<double> & domain, const std::vector<double> & range) {
    delete m_fit_func; m_fit_func = 0;
    delete m_top_axis; m_top_axis = 0;
    delete m_hist; m_hist = 0;

    // Make the histogram.
    m_hist = new TH1D(m_hist_name.c_str(), "", domain.size() - 1, &domain[0]);
    if (0 == m_hist) throw std::runtime_error("SpectrumView::createHistogram: failed to allocate histogram");

    // Populate histogram from supplied data.
    for (std::vector<double>::size_type idx = 0; idx != range.size(); ++idx)
      m_hist->SetBinContent(idx, range[idx]);

    m_hist->SetStats(kFALSE);
    m_hist->GetXaxis()->CenterTitle(kTRUE);
    m_hist->GetYaxis()->CenterTitle(kTRUE);
  }

  void SpectrumView::clearDisplay() {
    if (0 != m_canvas) {
      StoredCanvas * stored_canvas = dynamic_cast<StoredCanvas *>(m_canvas);
      if (0 != stored_canvas) stored_canvas->setView(0);

      if (m_own_canvas) {
        // It is necessary to detach it from its parent prior to deleting it.
        TGCompositeFrame * parent = const_cast<TGCompositeFrame *>(
          dynamic_cast<const TGCompositeFrame *>(m_main_frame->GetParent())
        );
        if (0 != parent) parent->RemoveFrame(m_main_frame);
        m_main_frame->UnmapWindow();
        delete m_canvas;
        delete m_select_button;
        delete m_close_button;
        delete m_button_frame;
        delete m_main_frame;
        if (0 != parent) parent->Layout();
      }
      m_canvas = 0;
      m_select_button = 0;
      m_close_button = 0;
      m_button_frame = 0;
      m_main_frame = 0;
    }
    delete m_layout_hints; m_layout_hints = 0;
    m_own_canvas = false;
    m_highlight = false;
    m_displayed = false;
  }

  void SpectrumView::selectDomain(double min, double max) {
    if (0 != m_hist) {
      TAxis * x_axis = m_hist->GetXaxis();
      Int_t begin = 1;
      Int_t end = x_axis->GetNbins() + 1;
      for (; begin != end; ++begin) {
        if (x_axis->GetBinUpEdge(begin) > min) break;
      }
      for (; begin != end; --end) {
        if (x_axis->GetBinLowEdge(end - 1) < max) break;
      }
      x_axis->SetRange(begin, end - 1);
    }
    // Update display so that new selection is visible.
    updateDisplay();
  }

  void SpectrumView::displayLabels() {
    if (0 != m_hist && !m_label.empty()) {
      double y_min = m_hist->GetMinimum();
      double y_max = m_hist->GetMaximum();
      double ordinate = y_min + .7 * (y_max - y_min);

      for (label_cont_type::iterator itor = m_label.begin(); itor != m_label.end(); ++itor) {
        double energy = itor->getEnergy().energy();
        // Only display the label if its energy is in this view's current range.
        if (getMinEnergy().energy() <= energy && energy <= getMaxEnergy().energy()) {
          double abscissa = getAbscissa(itor->getEnergy());
          itor->display(abscissa, ordinate);
        }
      }
    }
  }

  EnergyView::EnergyView(const Spectrum & spectrum, double src_redshift): SpectrumView(spectrum, src_redshift) {
    // Create a view of the spectrum.
    init();
  }

  EnergyView::EnergyView(const SpectrumView & view): SpectrumView(view.getSpectrum(), view.getSrcRedshift()) {
    // Create a view of the same spectrum referenced by the source view.
    init();

    // Select the same visible domain as the source view.
    select(view.getMinEnergy(), view.getMaxEnergy());
  }

  EnergyView * EnergyView::clone() const { return new EnergyView(*this); }

  Energy EnergyView::getEnergy(size_type idx) const {
    Energy energy(0.);
    if (0 != m_hist && idx < size_type(m_hist->GetNbinsX()))
      energy = Energy(m_hist->GetXaxis()->GetBinLowEdge(idx));
    return energy;
  }

  Energy EnergyView::getMinEnergy() const {
    TAxis * x_axis = m_hist->GetXaxis();
    return Energy(x_axis->GetBinLowEdge(x_axis->GetFirst()));
  }

  Energy EnergyView::getMaxEnergy() const {
    TAxis * x_axis = m_hist->GetXaxis();
    return Energy(x_axis->GetBinUpEdge(x_axis->GetLast()));
  }

  double EnergyView::getAbscissa(const IEnergy & ienergy) const {
    return ienergy.energy();
  }

  void EnergyView::findPeakCandidates(std::vector<Energy> & peak_candidate) const {
    std::vector<double> dbl_peak;
    peak_candidate.resize(0, Energy(0.));
    SpectrumView::findPeakCandidates(dbl_peak);
    std::vector<double>::size_type num_found = dbl_peak.size();
    peak_candidate.resize(num_found, Energy(0.));
    for (std::vector<double>::size_type index = 0; index != num_found; ++index) {
      peak_candidate[index] = dbl_peak[index];
    }
  }

  bool EnergyView::isVelocity() const { return false; }

  void EnergyView::select(const IEnergy & min_energy, const IEnergy & max_energy) {
    // The domain of this view is in energy space.
    selectDomain(min_energy.energy(), max_energy.energy());
  }

  void EnergyView::findPeakCandidates(std::vector<double> & peak_candidate) const {
    SpectrumView::findPeakCandidates(peak_candidate);
  }

  void EnergyView::init() {
    // Make local aliases to spectrum's energy arrays; these will be used to form the domain vector.
    const std::vector<double> & min_energy(m_spectrum->minEnergy());
    const std::vector<double> & max_energy(m_spectrum->maxEnergy());

    // Make local alias to spectrum's count array, which is used to compute the range of the graph.
    const std::vector<double> & counts(m_spectrum->counts());

    // Create the domain and range.
    std::vector<double> domain(min_energy.size() + 1);
    std::vector<double> range(min_energy.size() + 1);

    // Copy the min_energy container to the domain.
    std::copy(min_energy.begin(), min_energy.end(), domain.begin());

    // Use the final value from the max_energy container for the last point.
    domain[min_energy.size()] = max_energy.back();

    // Apply source redshifts.
    if (m_src_redshift != 0.) {
      for (std::vector<double>::iterator itor = domain.begin(); itor != domain.end(); ++itor) {
        *itor *= (1. + m_src_redshift);
      }
    }

    // Copy counts into range after underflow.
    std::copy(counts.begin(), counts.end(), range.begin() + 1);

    // Underflow bin has 0 counts in every case.
    range[0] = 0.;

    // Make the histogram.
    createHistogram(domain, range);
  }

  void EnergyView::createHistogram(const std::vector<double> & domain, const std::vector<double> & range) {
    SpectrumView::createHistogram(domain, range);

    m_hist->SetXTitle("Energy (keV)");
    m_hist->SetYTitle("Counts");

#if 0
// TODO: Make extra axes work.
    double x_min = domain.front();
    double x_max = domain.back();
    double y_min = *std::min_element(range.begin(), range.end());
    double y_max = *std::max_element(range.begin(), range.end());
    double top = y_min + 1.05 * (y_max - y_min);

    m_top_axis = new TGaxis(x_min, top, x_max, top, x_min, x_max, 510, "-");
    if (0 == m_top_axis) throw std::runtime_error("EnergyView::createHistogram: failed to allocate top axis");
    m_top_axis->ImportAxisAttributes(m_hist->GetXaxis());
#endif
  }

  WavelengthView::WavelengthView(const Spectrum & spectrum, double src_redshift): SpectrumView(spectrum, src_redshift) {
    // Create a view of the spectrum.
    init();
  }

  WavelengthView::WavelengthView(const SpectrumView & view): SpectrumView(view.getSpectrum(), view.getSrcRedshift()) {
    // Create a view of the same spectrum referenced by the source view.
    init();

    // Select the same visible domain as the source view.
    select(view.getMinEnergy(), view.getMaxEnergy());
  }

  WavelengthView * WavelengthView::clone() const { return new WavelengthView(*this); }

  Energy WavelengthView::getEnergy(size_type idx) const {
    Energy energy(0.);
    if (0 != m_hist && idx < size_type(m_hist->GetNbinsX()))
      energy = Wavelength(m_hist->GetXaxis()->GetBinLowEdge(idx));
    return energy;
  }

  Energy WavelengthView::getMinEnergy() const {
    TAxis * x_axis = m_hist->GetXaxis();
    return Wavelength(x_axis->GetBinUpEdge(x_axis->GetLast()));
  }

  Energy WavelengthView::getMaxEnergy() const {
    TAxis * x_axis = m_hist->GetXaxis();
    return Wavelength(x_axis->GetBinLowEdge(x_axis->GetFirst()));
  }

  double WavelengthView::getAbscissa(const IEnergy & ienergy) const {
    return ienergy.wavelength();
  }

  void WavelengthView::findPeakCandidates(std::vector<Energy> & peak_candidate) const {
    std::vector<double> dbl_peak;
    peak_candidate.resize(0, Energy(0.));
    SpectrumView::findPeakCandidates(dbl_peak);
    std::vector<double>::size_type num_found = dbl_peak.size();
    peak_candidate.resize(num_found, Energy(0.));
    for (std::vector<double>::size_type index = 0; index != num_found; ++index) {
      peak_candidate[index] = Wavelength(dbl_peak[index]).energy();
    }
  }

  bool WavelengthView::isVelocity() const { return false; }

  void WavelengthView::select(const IEnergy & min_energy, const IEnergy & max_energy) {
    // The domain of this view is in energy space.
    selectDomain(max_energy.wavelength(), min_energy.wavelength());
  }

  void WavelengthView::findPeakCandidates(std::vector<double> & peak_candidate) const {
    SpectrumView::findPeakCandidates(peak_candidate);
  }

  void WavelengthView::init() {
    // Make local aliases to spectrum's energy arrays; these will be used to form the domain vector.
    const std::vector<double> & min_energy(m_spectrum->minEnergy());
    const std::vector<double> & max_energy(m_spectrum->maxEnergy());

    // Make local alias to spectrum's count array, which is used to compute the range of the graph.
    const std::vector<double> & counts(m_spectrum->counts());

    // Create the domain and range.
    std::vector<double> domain(min_energy.size() + 1);
    std::vector<double> range(min_energy.size() + 1);

    // Copy the max_energy container to the domain.
    std::copy(max_energy.rbegin(), max_energy.rend(), domain.begin());

    // Use the final value from the min_energy container for the last point.
    domain[max_energy.size()] = min_energy.front();

    // Apply source redshifts, and transform to wavelength space.
    for (std::vector<double>::iterator itor = domain.begin(); itor != domain.end(); ++itor) {
      if (m_src_redshift != 0.)
        *itor *= (1. + m_src_redshift);
      *itor = Energy(*itor).wavelength();
    }

    // Copy counts into range after underflow.
    std::copy(counts.rbegin(), counts.rend(), range.begin() + 1);

    // Underflow bin has 0 counts in every case.
    range[0] = 0.;

    // Make the histogram.
    createHistogram(domain, range);
  }

  void WavelengthView::createHistogram(const std::vector<double> & domain, const std::vector<double> & range) {
    SpectrumView::createHistogram(domain, range);

    m_hist->SetXTitle("Wavelength (A)");
    m_hist->SetYTitle("Counts");
  }

  VelocityView::VelocityView(const Spectrum & spectrum, const IEnergy & origin, double src_redshift):
    SpectrumView(spectrum, src_redshift), m_origin(origin) {
    // Create a view of the spectrum.
    init();
  }

  VelocityView::VelocityView(const SpectrumView & view, const IEnergy & origin):
    SpectrumView(view.getSpectrum(), view.getSrcRedshift()),
    m_origin(origin) {
    // Create a view of the same spectrum referenced by the source view.
    init();

    // Select the same visible domain as the source view.
    select(view.getMinEnergy(), view.getMaxEnergy());
  }

  void VelocityView::init() {
    // Make local aliases to spectrum's energy arrays; these will be used to form the domain vector.
    const std::vector<double> min_energy(m_spectrum->minEnergy());
    const std::vector<double> & max_energy(m_spectrum->maxEnergy());

    // Create the domain from these energies. Include an extra value for the low edge of first bin.
    std::vector<double> energy_domain(min_energy.size() + 1);

    // Copy the min_energy container to the domain.
    std::copy(min_energy.begin(), min_energy.end(), energy_domain.begin());

    // Use the last value in max_energy for the final edge.
    energy_domain.back() = max_energy.back();

    // Apply source redshifts.
    if (m_src_redshift != 0.) {
      for (std::vector<double>::iterator itor = energy_domain.begin(); itor != energy_domain.end(); ++itor) {
        *itor *= (1. + m_src_redshift);
      }
    }

    std::vector<double> domain(energy_domain.size());

    EnergyToVelocity to_vel(m_origin);

    std::transform(energy_domain.rbegin(), energy_domain.rend(), domain.begin(), to_vel);

    // Make local alias to spectrum's count array, which is used to compute the range of the graph.
    const std::vector<double> & counts(m_spectrum->counts());

    // Range has N + 1 values; value 0 is the underflow.
    std::vector<double> range(counts.size() + 1);

    // Copy counts into range after underflow.
    std::copy(counts.rbegin(), counts.rend(), range.begin() + 1);

    // Underflow bin has 0 counts in every case.
    range[0] = 0.;

    // Make the histogram.
    createHistogram(domain, range);
  }

  VelocityView * VelocityView::clone() const { return new VelocityView(*this); }

  Energy VelocityView::getEnergy(size_type idx) const {
    double energy = 0.;
    if (0 != m_hist && idx < size_type(m_hist->GetNbinsX())) {
      VelocityToEnergy to_energy(m_origin);
      energy = to_energy(m_hist->GetXaxis()->GetBinLowEdge(idx));
    }
    return Energy(energy);
  }

  Energy VelocityView::getMinEnergy() const {
    double energy = 0.;
    if (0 != m_hist) {
      VelocityToEnergy to_energy(m_origin);
      TAxis * x_axis = m_hist->GetXaxis();
      energy = to_energy(x_axis->GetBinUpEdge(x_axis->GetLast()));
    }
    return Energy(energy);
  }

  Energy VelocityView::getMaxEnergy() const {
    double energy = 0.;
    if (0 != m_hist) {
      VelocityToEnergy to_energy(m_origin);
      TAxis * x_axis = m_hist->GetXaxis();
      energy = to_energy(x_axis->GetBinLowEdge(x_axis->GetFirst()));
    }
    return Energy(energy);
  }

  double VelocityView::getAbscissa(const IEnergy & ienergy) const {
    EnergyToVelocity to_velocity(m_origin);
    return to_velocity(ienergy.energy());
  }

  void VelocityView::findPeakCandidates(std::vector<Energy> & peak_candidate) const {
    std::vector<double> dbl_peak;
    SpectrumView::findPeakCandidates(dbl_peak);
    peak_candidate.resize(0, Energy(0.));
    std::vector<double>::size_type num_found = dbl_peak.size();
    peak_candidate.resize(num_found, 0.);
    VelocityToEnergy to_energy(m_origin);
    for (size_t index = 0; index != num_found; ++index) {
      peak_candidate[index] = to_energy(dbl_peak[index]);
    }
  }

  bool VelocityView::isVelocity() const { return true; }

  const Wavelength & VelocityView::getOrigin() const { return m_origin; }

  void VelocityView::select(const IEnergy & min_energy, const IEnergy & max_energy) {
    EnergyToVelocity to_vel(m_origin);
    selectDomain(to_vel(max_energy.energy()), to_vel(min_energy.energy()));
  }

  void VelocityView::findPeakCandidates(std::vector<double> & peak_candidate) const {
    SpectrumView::findPeakCandidates(peak_candidate);
  }

  void VelocityView::createHistogram(const std::vector<double> & domain, const std::vector<double> & range) {
    SpectrumView::createHistogram(domain, range);

    m_hist->SetXTitle("Velocity(km/s)");
    m_hist->SetYTitle("Counts");
  }

}
