/** \file Table.cxx
    \brief Implementation of Table class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <cctype>
#include <cstring>
#include <fstream>
#include <limits>
#include <map>
#include <sstream>
#include <stdexcept>
#include <utility>
#include <vector>

#include "fitsio.h"

#include "TableFilePtr.h"
#include "TextTableFilePtr.h"
#include "profit/Table.h"

namespace profit {

  /** \class FitsException
      \brief Encapsulation of errors from cfitsio.
  */
  class FitsException : public std::exception {
    public:
      /* \brief Create an exception object.
         \param status The status returned by the cfitsio call.
         \param text A context-specific message. The cfitsio short error string will be appended to this message.
      */
      FitsException(int status, const std::string & text): std::exception(), m_what(text), m_status(status) {
        char msg[31] = "";
        fits_get_errstatus(status, msg);
        m_what += std::string(" (CFITSIO: ") + msg + ")";
      }

      virtual ~FitsException() throw() {}

      virtual const char * what() const throw() { return m_what.c_str(); }

      /// \brief Return the cfitsio status code associated with this exception.
      int status() const { return m_status; }

    private:
      std::string m_what;
      int m_status;
  };

  /** \class FitsTableFilePtr
      \brief Internal class containing cfitsio primitive.
  */
  class FitsTableFilePtr : public TableFilePtr {
    public:
      /** \brief Create an file accessor object.
          \param file_name The name of the file.
          \param table_name The name of the table in the file.
      */
      FitsTableFilePtr(const std::string & file_name, const std::string & table_name);

      /// \brief Close the file being accessed, and free all associated resources.
      virtual ~FitsTableFilePtr();

      /** \brief Get the contents of a table column. Column values must all be convertible to double.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<double> & column) const;

      /** \brief Get the contents of a table column as strings.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<std::string> & column) const;

      /** \brief Return the content of a keyword, as a string.
          \param key_name The name of the keyword to get.
      */
      virtual std::string getKeyword(const std::string & key_name) const;

      /** \brief Get the identifier for the given column.
          \param field_name The name of the field/column.
      */
      virtual size_type getColumnId(const std::string & field_name) const;

      /** \brief Get the field encapsulation for the given column.
          \param field_name The name of the field/column.
      */
      virtual Field getField(const std::string & field_name) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
      */
      virtual double getCell(const std::string & field_name, size_type record_index) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::string & field_value) const;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::vector<double> & field_value) const;

      /** \brief Filter the current table contents using the given expression.
          \param expression The expression (Cfitsio row filtering syntax.)
      */
      virtual void filter(const std::string & expression);

      /** \brief Filter the current table contents using the given expression.
          \param field_name The name of the field on which to filter.
          \param min_value The minimum value of the filter range.
          \param max_value The maximum value of the filter range.
      */
      virtual void filter(const std::string & field_name, double min_value, double max_value);

      /** \brief Get the dimensionality of the given field.
          \param field_name The name of the field/column.
      */
      virtual size_type getFieldDim(const std::string & field_name) const;

    private:
      size_type getNumRowsInFile(fitsfile * fp) const;

      fitsfile * m_fp;
  };

  Record::Record(const TableFilePtr * fp, Record::size_type record_index): m_fp(fp), m_index(record_index) {}

  double Record::getField(const std::string & field_name) const {
    return m_fp->getCell(field_name, m_index);
  }

  void Record::getField(const std::string & field_name, std::string & field_value) const {
    m_fp->getCell(field_name, m_index, field_value);
  }

  void Record::getField(const std::string & field_name, std::vector<double> & field_value) const {
    m_fp->getCell(field_name, m_index, field_value);
  }

  InputTable::InputTable(const std::string & file_name, const std::string & table_name): m_fp(0) {
    bool is_fits = true;
    {
      std::ifstream fs(file_name.c_str());
      if (fs.good()) {
        char buf[80];
        fs.getline(buf, 80);
        char * cp;
        for (cp = buf; '\0' != *cp; ++cp) *cp = toupper(*cp);
        cp = buf;
        if (0 != strncmp(cp, "SIMPLE", 6)) {
          is_fits = false;
        } else {
          cp += 6;
          while (0 != isspace(*cp)) ++cp;
          if ('=' != *cp) {
            is_fits = false;
          } else {
            ++cp;
            while (0 != isspace(*cp)) ++cp;
            if ('T' != *cp) {
              is_fits = false;
            }
          }
        }
      }
    }
    if (is_fits) m_fp = new FitsTableFilePtr(file_name, table_name);
    else m_fp = new TextTableFilePtr(file_name);
  }

  InputTable::~InputTable() { delete m_fp; }

  InputTable::size_type InputTable::getNumRecords() const { return m_fp->getNumRecords(); }

  void InputTable::getColumn(const std::string & field_name, std::vector<double> & column) const
    { m_fp->getColumn(field_name, column); }

  void InputTable::getColumn(const std::string & field_name, std::vector<std::string> & column) const
    { m_fp->getColumn(field_name, column); }

  InputTable::size_type InputTable::getFieldDim(const std::string & field_name) const
    { return m_fp->getFieldDim(field_name); }

  std::string InputTable::getKeyword(const std::string & key_name) const { return m_fp->getKeyword(key_name); }

  const std::string & InputTable::getFileName() const { return m_fp->getFileName(); }

  std::string InputTable::getDirectoryName() const {
    std::string dir(getFileName());
    std::string::size_type last_slash = dir.find_last_of("/\\");
    if (std::string::npos != last_slash) dir = dir.substr(0, last_slash);
    else dir = "";
    return dir;
  }

  Record InputTable::getRecord(Record::size_type record_index) const { return m_fp->getRecord(record_index); }

  void InputTable::filter(const std::string & expression) { m_fp->filter(expression); }

  void InputTable::filter(const std::string & field_name, double min_value, double max_value) {
    m_fp->filter(field_name, min_value, max_value);
  }

  FitsTableFilePtr::FitsTableFilePtr(const std::string & file_name, const std::string & table_name):
    TableFilePtr(file_name, table_name), m_fp(0) {
    int status = 0;
    std::string full_name = file_name;

    // Concatenate the table name with the file name.
    if (!table_name.empty()) full_name += "[" + table_name + "]";

    // Add a selection which does not exclude any data, but causes the file to be opened as a copy in memory.
    full_name += "[#row>0]";

    // Open table.
    fitsfile * fp = 0;
    fits_open_table(&fp, const_cast<char *>(full_name.c_str()), READONLY, &status);
    if (0 != status) throw FitsException(status, "Cannot open table " + full_name);

    try {
      // Get number of rows in the file.
      m_num_rows = getNumRowsInFile(fp);

      // Iterate over all columns in file, putting the name of each in the column container.
      char col_name[128];
      char * match_all = "*";
      do {
        *col_name = '\0';
        int col_num = 0;
        // Get each column's name.
        fits_get_colname(fp, CASESEN, match_all, col_name, &col_num, &status);
        if (0 == status || COL_NOT_UNIQUE == status) {
          // Standardize all names to lowercase.
          for (char * cp = col_name; 0 != *cp; ++cp) *cp = tolower(*cp);

          int col_type_status = 0;
          long repeat = 0;
          fits_get_coltype(fp, col_num, 0, &repeat, 0, &col_type_status);
          if (0 != col_type_status)
            throw FitsException(col_type_status, std::string("Cannot get type of column ") + col_name + " in file " + full_name);

          m_col_lookup.insert(std::make_pair(std::string(col_name), Field(col_num, repeat)));
        }
      } while (COL_NOT_FOUND != status && 0 != status);

    } catch (const FitsException & x) {
      m_col_lookup.clear();
      m_num_rows = 0;
      int status = x.status();
      fits_close_file(fp, &status);
      throw;
    }

    // Success: keep fitsfile pointer.
    m_fp = fp;
  }

  FitsTableFilePtr::~FitsTableFilePtr() {
    if (0 != m_fp) {
      int status = 0;
      fits_close_file(m_fp, &status);
    }
  }

  void FitsTableFilePtr::getColumn(const std::string & field_name, std::vector<double> & column) const {
    int status = 0;
    // Get rid of current contents in the output column, if any.
    column.clear();

    // Find the column in the map of column names.
    size_type field_id = getColumnId(field_name);

    // Reserve enough space for the number of rows in the column.
    column.resize(m_num_rows);

    // Read the column from the file.
    fits_read_col(m_fp, TDOUBLE, field_id, 1, 1, m_num_rows, 0, &(column[0]), 0, &status);
    if (0 != status)
      throw FitsException(status, "Cannot read double values from field named \"" + field_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);
  }

  void FitsTableFilePtr::getColumn(const std::string & field_name, std::vector<std::string> & column) const {
    // Get rid of current contents in the output column, if any.
    column.clear();
    throw std::logic_error("FitsTableFilePtr::getColumn(field_name, vector<string> &) not yet supported");
  }

  std::string FitsTableFilePtr::getKeyword(const std::string & key_name) const {
    int status = 0;
    char key_value[FLEN_VALUE] = "";
    fits_read_key(m_fp, TSTRING, const_cast<char *>(key_name.c_str()), key_value, 0, &status);
    if (0 != status)
      throw FitsException(status, "Cannot read keyword \"" + key_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);
    return key_value;
  }

  TableFilePtr::size_type FitsTableFilePtr::getColumnId(const std::string & field_name) const {
    // Perform case-insensitive lookup.
    std::string lc_field_name = field_name;
    for (std::string::iterator itor = lc_field_name.begin(); itor != lc_field_name.end(); ++itor) *itor = tolower(*itor);

    // Find the column in the map of column names.
    ColCont::const_iterator itor = m_col_lookup.find(lc_field_name);
    if (m_col_lookup.end() == itor)
      throw std::runtime_error("Cannot find field named \"" + field_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);

    return itor->second.m_field_index;
  }

  Field FitsTableFilePtr::getField(const std::string & field_name) const {
    // Perform case-insensitive lookup.
    std::string lc_field_name = field_name;
    for (std::string::iterator itor = lc_field_name.begin(); itor != lc_field_name.end(); ++itor) *itor = tolower(*itor);

    // Find the column in the map of column names.
    ColCont::const_iterator itor = m_col_lookup.find(lc_field_name);
    if (m_col_lookup.end() == itor)
      throw std::runtime_error("Cannot find field named \"" + field_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);

    return itor->second;
  }

  double FitsTableFilePtr::getCell(const std::string & field_name, TableFilePtr::size_type record_index) const {
    int status = 0;
    double cell_content = 0.;

    // Look up the column id.
    size_type column_id = getColumnId(field_name);

    // Read the column from the file.
    fits_read_col(m_fp, TDOUBLE, column_id, record_index + 1, 1, 1, 0, &cell_content, 0, &status);
    if (0 != status)
      throw FitsException(status, "Cannot read values from column named \"" + field_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);

    return cell_content;
  }

  void FitsTableFilePtr::getCell(const std::string & field_name, TableFilePtr::size_type record_index, std::string & field_value)
    const {
    throw std::logic_error("FitsTableFilePtr::getCell not yet implemented for strings.");
  }

  void FitsTableFilePtr::getCell(const std::string & field_name, TableFilePtr::size_type record_index,
    std::vector<double> & field_value) const {
    int status = 0;

    // Look up the column id.
    Field field = getField(field_name);

    // Allocate enough space for vector value.
    field_value.resize(field.m_field_dim);

    // Read the column from the file.
    fits_read_col(m_fp, TDOUBLE, field.m_field_index, record_index + 1, 1, field.m_field_dim, 0, &(field_value[0]), 0, &status);
    if (0 != status)
      throw FitsException(status, "Cannot read values from column named \"" + field_name +
        "\" in table " + m_table_name + " in FITS file " + m_file_name);

  }

  void FitsTableFilePtr::filter(const std::string & expression) {
    int status = 0;
    fits_select_rows(m_fp, m_fp, const_cast<char *>(expression.c_str()), &status);
    if (0 != status)
      throw FitsException(status, "filter(\"" + expression + "\") failed in table " + m_table_name + " in FITS file "
        + m_file_name);

    m_num_rows = getNumRowsInFile(m_fp);
  }

  void FitsTableFilePtr::filter(const std::string & field_name, double min_value, double max_value) {
    std::ostringstream os;
    os.precision(std::numeric_limits<double>::digits10);
    os << "(" << field_name << ">=" << min_value << "&&" << field_name << "<=" << max_value << ")";
    filter(os.str());
  }

  InputTable::size_type FitsTableFilePtr::getFieldDim(const std::string & field_name) const {
    return getField(field_name).m_field_dim;
  }

  TableFilePtr::size_type FitsTableFilePtr::getNumRowsInFile(fitsfile * fp) const {
    int status = 0;
    long num_rows = 0;
    fits_get_num_rows(fp, &num_rows, &status);
    if (0 != status) throw FitsException(status, "Cannot get number of rows in table");
    else if (num_rows < 0) num_rows = 0;
    return num_rows;
  }
}
