/** \file TableFilePtr.h
    \brief Declaration of TableFilePtr class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#ifndef profit_TableFilePtr_h
#define profit_TableFilePtr_h

#include <cstddef>
#include <map>
#include <string>
#include <vector>

namespace profit {

  class Record;

  struct Field {
    typedef size_t size_type;
  
    Field(size_type field_index, size_type field_dim): m_field_index(field_index), m_field_dim(field_dim) {}

    size_type m_field_index;
    size_type m_field_dim;
  };

  /** \class TableFilePtr
      \brief Internal base class for concrete implementations of Table interface for different file formats.
  */
  class TableFilePtr {
    public:
      typedef size_t size_type;

      /** \brief Create an file accessor object.
          \param file_name The name of the file.
          \param table_name The name of the table in the file.
      */
      TableFilePtr(const std::string & file_name, const std::string & table_name);

      /// \brief Close the file being accessed, and free all associated resources.
      virtual ~TableFilePtr();

      /** \brief Get the contents of a table column. Column values must all be convertible to double.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<double> & column) const = 0;

      /** \brief Get the contents of a table column as strings.
          \param field_name The name of the field/column.
          \param column The destination array to hold the column contents.
      */
      virtual void getColumn(const std::string & field_name, std::vector<std::string> & column) const = 0;

      /** \brief Get the identifier for the given column.
          \param field_name The name of the field/column.
      */
      virtual size_type getColumnId(const std::string & field_name) const = 0;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
      */
      virtual double getCell(const std::string & field_name, size_type record_index) const = 0;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::string & field_value) const = 0;

      /** \brief Get the cell identified by the field/column and record/row number.
          \param field_name The name of the field/column.
          \param record_index The index (0-based row number) in the file.
          \param field_value The value of the field/column.
      */
      virtual void getCell(const std::string & field_name, size_type record_index, std::vector<double> & field_value) const = 0;

      /** \brief Filter the current table contents using the given expression.
          \param expression The expression (Cfitsio row filtering syntax.)
      */
      virtual void filter(const std::string & expression) = 0;

      /** \brief Filter the current table contents using the given expression.
          \param field_name The name of the field on which to filter.
          \param min_value The minimum value of the filter range.
          \param max_value The maximum value of the filter range.
      */
      virtual void filter(const std::string & field_name, double min_value, double max_value) = 0;

      /** \brief Get the dimensionality of the given field.
          \param field_name The name of the field/column.
      */
      virtual size_type getFieldDim(const std::string & field_name) const = 0;

      /// \brief Return the name of the file containing this table.
      virtual const std::string & getFileName() const;

      /// \brief Return the number of records in the table.
      virtual size_type getNumRecords() const;

      /** \brief Return the content of a keyword, as a string.
          \param key_name The name of the keyword to get.
      */
      virtual std::string getKeyword(const std::string & key_name) const;

      /** \brief Return a single record (row) of the file.
          \param record_index The index (0-based row number) in the file.
      */
      virtual Record getRecord(size_type record_index) const;
     
    protected:
      typedef std::map<std::string, Field> ColCont;
      ColCont m_col_lookup;
      std::string m_file_name;
      std::string m_table_name;
      size_type m_num_rows;
  };

}

#endif
