/** \file TestSpectrumView.cxx
    \brief Implementation of SpectrumView class-specific parts of the unit test.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include <functional>
#include <memory>
#include <unistd.h>

#include "profit/Energy.h"
#include "profit/SpecFile.h"
#include "profit/Spectrum.h"
#include "profit/SpectrumView.h"
#include "profit/Table.h"
#include "profit/UnitTestGui.h"

#include "TGFrame.h"
#include "TH1.h"
#include "TPad.h"

namespace profit {

  void UnitTestGui::testSpectrumView() {
    std::string method("testSpectrumView()");

    // Start by creating a spectrum object. Note that this parallels the test in
    // the testSpectrum method, but this is not testing the spectrum object.
    info(method) << "Opening table \"SPECTRUM\" in file \"xrs.fak\"." << std::endl;
    SpecFile spec_file("xrs.fak");
    info(method) << std::endl;

    info(method) << "Opening table \"EBOUNDS\" in file \"xrs.rmf\"." << std::endl;
    InputTable ebounds_table("xrs.rmf", "EBOUNDS");
    info(method) << std::endl;

    info(method) << "Creating spectrum object from input data tables." << std::endl;
    Spectrum spectrum(spec_file, ebounds_table);
    info(method) << std::endl;

    info(method) << "Selecting on spectrum to produce a small test data set." << std::endl;
    spectrum.select(.98995, 1.0105);
    info(method) << std::endl;

    // Create a view of this spectrum.
    info(method) << "Creating EnergyView object from the spectrum object." << std::endl;
    EnergyView energy_view(spectrum);
    info(method) << std::endl;

    // Construct correct domain.
    info(method) << "Extracting energy/counts containers from spectrum object." << std::endl;
    std::vector<double> domain = spectrum.minEnergy();
    domain.push_back(spectrum.maxEnergy().back());
    std::vector<double> range = spectrum.counts();
    info(method) << std::endl;

    testSpectrumView("After creating EnergyView", energy_view, domain, range);

    // Clean up.
    info(method) << "(Cleaning up from any prior runs of this test.)" << std::endl;
    delete m_spectrum_view; m_spectrum_view = 0;
    clearFrames(m_temp_frame_cont);
    info(method) << std::endl;

    info(method) << "Cloning EnergyView object." << std::endl;
    m_spectrum_view = energy_view.clone();
    info(method) << std::endl;

    // Confirm that cloned object agrees with original.
    testSpectrumView("After cloning EnergyView", *m_spectrum_view, domain, range);

    info(method) << "Running fit method and displaying fit result." << std::endl;
    m_spectrum_view->fit();
    info(method) << std::endl;
    
    info(method) << "Adding a label at the central energy of the display." << std::endl;
    m_spectrum_view->addLabel(Energy(.998), ".998 keV.");
    info(method) << std::endl;

    info(method) << "Testing display of clone of EnergyView" << std::endl;
    m_spectrum_view->display(m_main_frame, kLHintsNormal, 440, 270);
    info(method) << std::endl;

    // Confirm that cloned object agrees with original.
    testSpectrumView("After displaying clone of EnergyView", *m_spectrum_view, domain, range);

    double min_energy = .9925;
    double max_energy = 1.0035;

    info(method) << "Performing energy selection." << std::endl;
    m_spectrum_view->select(Energy(min_energy), Energy(max_energy));
    info(method) << std::endl;

    // Perform the cut on the local copy of the domain, to make sure expected answer is obtained.
    std::vector<double>::size_type begin_idx = 0;
    std::vector<double>::size_type end_idx = domain.size();

    if (begin_idx != end_idx) {
      for (; begin_idx != end_idx - 1; ++begin_idx) {
        // Find bin containing the minimum.
        if (domain[begin_idx] <= min_energy && min_energy < domain[begin_idx + 1]) break;
      }

      for (; 0 != end_idx - 1; --end_idx) {
        // Find bin containing the maximum.
        if (domain[end_idx - 2] <= max_energy && max_energy < domain[end_idx - 1]) break;
      }
    }

    domain = std::vector<double>(domain.begin() + begin_idx, domain.begin() + end_idx);
    // Range array has one fewer element.
    range = std::vector<double>(range.begin() + begin_idx, range.begin() + end_idx - 1);
    
    testSpectrumView("After performing energy selection on clone of EnergyView", *m_spectrum_view, domain, range);

    info(method) << "Slicing a spectrum using whole energy domain." << std::endl;
    // Note: this test is correctly defined to maximize its usefulness. The object m_spectrum_view has been
    // selected upon, so it will not display the whole spectrum, but it should still contain the whole spectrum.
    // Thus we should be able to call m_spectrum_view->slice, but use min/max energy from the original energy_view.
    std::auto_ptr<SpectrumView> new_view(m_spectrum_view->slice(energy_view.getMinEnergy(), energy_view.getMaxEnergy()));
    info(method) << std::endl;

    // Reset correct domain and range to energy_view's original domain/range.
    energy_view.getDomain(domain);
    energy_view.getRange(range);

    testSpectrumView("After slicing spectrum using whole energy domain", *new_view, domain, range);

    info(method) << "Slicing a spectrum using energy subrange." << std::endl;
    new_view.reset(energy_view.slice(Energy(min_energy), Energy(max_energy)));
    info(method) << std::endl;

    // Reset correct domain and range to m_spectrum_view's selected range.
    m_spectrum_view->getDomain(domain);
    m_spectrum_view->getRange(range);

    testSpectrumView("After slicing spectrum using energy subrange", *new_view, domain, range);

    info(method) << "Slicing a sliced spectrum to include larger range, which should not change the range." << std::endl;
    new_view.reset(new_view->slice(Energy(0.), Energy(10.)));
    info(method) << std::endl;

    testSpectrumView("After slicing spectrum again using larger range", *new_view, domain, range);

    info(method) << "Slicing domain of spectrum to include the a smaller range." << std::endl;
    new_view.reset(new_view->slice(min_energy + .002, max_energy - .002));
    info(method) << std::endl;

    std::vector<double> sliced_domain(domain.begin() + 3, domain.end() - 3);
    std::vector<double> sliced_range(range.begin() + 3, range.end() - 3);
    testSpectrumView("After slicing domain of spectrum", *new_view, sliced_domain, sliced_range);

    info(method) << "Test transforming from energy to velocity space." << std::endl;
    VelocityView::EnergyToVelocity to_vel(Energy(1.));
    double input_energy = 1.;
    double correct_velocity = 0.;
    double velocity = to_vel(input_energy);
    if (velocity != correct_velocity)
      error(method) << "EnergyToVelocity::operator() returned velocity " << velocity << " for energy = " << input_energy <<
        ", not " << correct_velocity << ", as expected." << std::endl;

    input_energy = .5;
    static double s_c = 299792.458; // km/s
    correct_velocity = .5 * s_c;
    velocity = to_vel(input_energy);
    if (velocity != correct_velocity)
      error(method) << "EnergyToVelocity::operator() returned velocity " << velocity << " for energy = " << input_energy <<
        ", not " << correct_velocity << ", as expected." << std::endl;

    input_energy = 2.;
    correct_velocity = -s_c;
    velocity = to_vel(input_energy);
    if (velocity != correct_velocity)
      error(method) << "EnergyToVelocity::operator() returned velocity " << velocity << " for energy = " << input_energy <<
        ", not " << correct_velocity << ", as expected." << std::endl;
    info(method) << std::endl;

    info(method) << "Test transforming from velocity to energy space." << std::endl;
    VelocityView::VelocityToEnergy to_energy(Energy(1.));
    double input_velocity = s_c;
    double correct_energy = 0.;
    double energy = to_energy(input_velocity);
    if (energy != correct_energy)
      error(method) << "VelocityToEnergy::operator() returned energy " << energy << " for velocity = " << input_velocity <<
        ", not " << correct_energy << ", as expected." << std::endl;

    input_velocity = .5 * s_c;
    correct_energy = .5;
    energy = to_energy(input_velocity);
    if (energy != correct_energy)
      error(method) << "VelocityToEnergy::operator() returned energy " << energy << " for velocity = " << input_velocity <<
        ", not " << correct_energy << ", as expected." << std::endl;

    input_velocity = -s_c;
    correct_energy = 2.;
    energy = to_energy(input_velocity);
    if (energy != correct_energy)
      error(method) << "VelocityToEnergy::operator() returned energy " << energy << " for velocity = " << input_velocity <<
        ", not " << correct_velocity << ", as expected." << std::endl;
    info(method) << std::endl;

    info(method) << "Test creating velocity view." << std::endl;
    double energy_0 = 1.;
    double lambda_0 = 12.3984;
    VelocityView vel_view(*m_spectrum_view, Wavelength(lambda_0));
    info(method) << std::endl;

    std::reverse(domain.begin(), domain.end());
    std::reverse(range.begin(), range.end());
    for (std::vector<double>::iterator itor = domain.begin(); itor != domain.end(); ++itor) {
      *itor = s_c * (energy_0 - Energy(*itor).energy()) / energy_0;
    }

    testSpectrumView("After creating velocity view", vel_view, domain, range);

    // Test peak candidate finder in energy space.
    info(method) << "Test findPeakCandidates on energy view" << std::endl;
    std::vector<Energy> peak_candidate;
    energy_view.findPeakCandidates(peak_candidate);
    for (std::vector<Energy>::size_type ii = 0; ii != peak_candidate.size(); ++ii) {
      info(method) << peak_candidate[ii].energy() << std::endl;
    }
    info(method) << "Test findPeakCandidates on energy view in natural units" << std::endl;
    std::vector<double> dbl_peak_candidate;
    energy_view.findPeakCandidates(dbl_peak_candidate);
    for (std::vector<double>::size_type ii = 0; ii != dbl_peak_candidate.size(); ++ii) {
      info(method) << dbl_peak_candidate[ii] << std::endl;
    }
    info(method) << std::endl;

    // TODO: confirm numbers are exactly correct.
    // Test peak candidate finder in velocity space.
    info(method) << "Test findPeakCandidates on velocity view" << std::endl;
    VelocityView vel_view2(spectrum, Wavelength(lambda_0));
    vel_view2.findPeakCandidates(peak_candidate);
    for (std::vector<Energy>::size_type ii = 0; ii != peak_candidate.size(); ++ii) {
      info(method) << to_vel(peak_candidate[ii].energy()) << std::endl;
    }
    info(method) << "Test findPeakCandidates on velocity view in natural units" << std::endl;
    vel_view2.findPeakCandidates(dbl_peak_candidate);
    for (std::vector<double>::size_type ii = 0; ii != dbl_peak_candidate.size(); ++ii) {
      info(method) << dbl_peak_candidate[ii] << std::endl;
    }
    info(method) << std::endl;

    m_main_frame->MapSubwindows();
    m_main_frame->Layout();
    m_main_frame->MapWindow();
  }

  void UnitTestGui::testSpectrumView(const std::string & hint, const SpectrumView & spectrum_view,
    const std::vector<double> & domain, const std::vector<double> & range) const {
    std::string method("testSpectrumView");
    std::vector<double> actual;
    std::vector<double>::size_type end;

    // Confirm contents of view's domain.
    info(method) << hint << ", confirming SpectrumView's domain agrees with expected domain." << std::endl;
    spectrum_view.getDomain(actual);
    if (domain.size() != actual.size())
      error(method) << hint << ", domain has size " << actual.size() << ", not " << domain.size() <<
      ", as expected." << std::endl;
    
    // Report errors in overlapping arrays.
    end = domain.size() < actual.size() ? domain.size() : actual.size();
    for (std::vector<double>::size_type idx = 0; idx != end; ++idx)
      if (!isEqual(domain[idx], actual[idx], 0., s_float_epsilon))
        error(method) << hint << ", domain[" << idx << "] = " << actual[idx] << ", not " << domain[idx] <<
          ", as expected" << std::endl;

    // Report errors in portions of arrays which do not overlap.
    for (std::vector<double>::size_type idx = end; idx < domain.size(); ++idx) {
      error(method) << hint << ", expected[" << idx << "] = " << domain[idx] << " (no actual value)." << std::endl;
    }
    for (std::vector<double>::size_type idx = end; idx < actual.size(); ++idx) {
      error(method) << hint << ", domain[" << idx << "] = " << actual[idx] << " (no expected value)." << std::endl;
    }
    info(method) << std::endl;

    // Confirm contents of view's range.
    info(method) << hint << ", confirming SpectrumView's range agrees with expected range." << std::endl;
    spectrum_view.getRange(actual);
    if (range.size() != actual.size())
      error(method) << hint << ", range has size " << actual.size() << ", not " << range.size() <<
      ", as expected." << std::endl;
    
    end = range.size() < actual.size() ? range.size() : actual.size();
    for (std::vector<double>::size_type idx = 0; idx != end; ++idx)
      if (!isEqual(range[idx], actual[idx], 0., s_float_epsilon))
        error(method) << hint << ", range[" << idx << "] = " << actual[idx] << ", not " << range[idx] <<
          ", as expected" << std::endl;

    // Report errors in portions of arrays which do not overlap.
    for (std::vector<double>::size_type idx = end; idx < range.size(); ++idx) {
      error(method) << hint << ", expected range[" << idx << "] = " << range[idx] << " (no actual value)." << std::endl;
    }
    for (std::vector<double>::size_type idx = end; idx < actual.size(); ++idx) {
      error(method) << hint << ", range[" << idx << "] = " << actual[idx] << " (no expected value)." << std::endl;
    }
    info(method) << std::endl;

    // Confirm min/max energy.
    info(method) << hint << ", confirming SpectrumView::getMinEnergy/getMaxEnergy return correct values." << std::endl;
    const EnergyView * energy_view = dynamic_cast<const EnergyView *>(&spectrum_view);
    if (0 != energy_view) {
      if (energy_view->getMinEnergy().energy() != domain.front())
        error(method) << hint << ", EnergyView::getMinEnergy() returned " << energy_view->getMinEnergy().energy() <<
          " keV, not " << domain.front() << " keV, as expected." << std::endl;
      if (energy_view->getMaxEnergy().energy() != domain.back())
        error(method) << hint << ", EnergyView::getMaxEnergy() returned " << energy_view->getMaxEnergy().energy() <<
          " keV, not " << domain.back() << " keV, as expected." << std::endl;
    }

    const VelocityView * velocity_view = dynamic_cast<const VelocityView *>(&spectrum_view);
    if (0 != velocity_view) {
      VelocityView::VelocityToEnergy to_energy(velocity_view->getOrigin());
      if (!isEqual(velocity_view->getMinEnergy().energy(), to_energy(domain.back()), s_float_epsilon))
        error(method) << hint << ", VelocityView::getMinEnergy() returned " << velocity_view->getMinEnergy().energy() <<
          " keV, not " << to_energy(domain.back()) << " keV, as expected." << std::endl;
      if (!isEqual(velocity_view->getMaxEnergy().energy(), to_energy(domain.front()), s_float_epsilon))
        error(method) << hint << ", VelocityView::getMaxEnergy() returned " << velocity_view->getMaxEnergy().energy() <<
          " keV, not " << to_energy(domain.front()) << " keV, as expected." << std::endl;
    }
    info(method) << std::endl;

    // Confirm expected computation of abscissa.
    info(method) << hint << ", confirming getAbscissa() returned correct value." << std::endl;
    if (0 != energy_view) {
      double abscissa = energy_view->getAbscissa(Energy(7.));
      if (7. != abscissa)
        error(method) << hint << ", EnergyView::getAbscissa() returned " << abscissa << ", not 7. as expected." << std::endl;
    }

    if (0 != velocity_view) {
      VelocityView::EnergyToVelocity to_velocity(velocity_view->getOrigin());
      double abscissa = velocity_view->getAbscissa(Energy(7.));
      if (!isEqual(abscissa, to_velocity(7.), s_float_epsilon))
        error(method) << hint << ", VelocityView::getAbscissa() returned " << abscissa << ", not " << to_velocity(7.) <<
          ", as expected." << std::endl;
    }
    info(method) << std::endl;
  }

}
