/** \file XstarDatabase.cxx
    \brief Implementation of XstarDatabase class.
    \author James Peachey

    The Profit software package was developed by Dr. Randall Smith and
    Mr. James Peachey at the High Energy Astrophysics Science Archive
    Research Center (HEASARC) at the Goddard Space Flight Center (GSFC)
    of the National Aeronautics and Space Administration (NASA).
*/
#include "XstarDatabase.h"
#include "profit/Energy.h"

#include <limits>

namespace profit {

  XstarDatabase::XstarDatabase(const std::string & file_name): m_table(file_name, ""), m_matches() {}

  void XstarDatabase::filter(const IEnergy & min_energy, const IEnergy & max_energy) {
    m_table.filter("wave", max_energy.wavelength(), min_energy.wavelength());
  }

  void XstarDatabase::filter(double min_emissivity) {
    m_table.filter("MaxEmis", min_emissivity, std::numeric_limits<double>::max());
  }

  const AtomDbImp::AtomicLineCont & XstarDatabase::getMatches() const {
    // Erase the current set of matches.
    m_matches.clear();

    InputTable::size_type num_rec = m_table.getNumRecords();

    // Make sure container has enough room for the matches.
    m_matches.resize(num_rec);

    // Read all data from input table.
    for (InputTable::size_type index = 0; index != num_rec; ++index) {
      Record rec = m_table.getRecord(index);

      int element = int(rec.getField("Z"));
      int ion = int(rec.getField("rmJ"));
      double max_emis = rec.getField("MaxEmis");

      // Assign peak emissivity == 1. for now.
      m_matches[index] = AtomicLine("XSTAR", rec.getField("wave"), element, ion,
        int(rec.getField("index_upper")), int(rec.getField("index_low")), max_emis);
    }

    return m_matches;
  }

}
